/******************************************************************************
 * File: Property.impl.h
 * Date: 12.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "Property.h"
#include "PropertyExceptions.h"
#include "util/StringUtils.h"
#include "util/StringConv.h"

template<typename ValType>
Property<ValType>::Property(PropertyID id, QString name, value_type defValue, Flags flags) : PropertyBase(id, name, flags),
	_value{defValue}
{

}

template<typename ValType>
Property<ValType>::Property(const value_type& val) : PropertyBase(val),
	_value{val}
{

}

template<typename ValType>
Property<ValType>::Property(value_type&& val) : PropertyBase(val),
	_value{std::move(val)}
{

}

template<typename ValType>
Property<ValType>& Property<ValType>::operator =(const value_type& val)
{
	setValue(val);
	return *this;
}

template<typename ValType>
Property<ValType>& Property<ValType>::operator =(value_type&& val)
{
	setValue(val);
	return *this;
}

template<typename ValType>
bool Property<ValType>::operator ==(const property_type& other) const
{
	return _value == other._value;
}

template<typename ValType>
bool Property<ValType>::operator ==(const value_type& val) const
{
	return _value == val;
}

template<typename ValType>
void Property<ValType>::setValue(const value_type& val)
{
	_setValue(val, [&val, this]() { _value = val; });
}

template<typename ValType>
void Property<ValType>::setValue(value_type&& val)
{
	_setValue(val, [&val, this]() { _value = std::move(val); });
}

template<typename ValType>
template<template<typename> class ConstraintType, typename... Args>
void Property<ValType>::createConstraint(Args... args)
{
	_constraints.emplace_back(new ConstraintType<value_type>(this, std::forward<Args>(args)...));
}

template<typename ValType>
void Property<ValType>::removeConstraint(const PropertyConstraints::value_type& constraint)
{
	std::remove(_constraints.cbegin(), _constraints.cend(), constraint);
}

template<typename ValType>
void Property<ValType>::copyValue(const PropertyBase* property)
{
	if (auto typedProperty = dynamic_cast<const property_type*>(property))
		setValue(typedProperty->_value);
}

template<typename ValType>
QString Property<ValType>::toString() const
{
	return StringUtils::encodeEscapeCharacters(StringConv::convertValue(getValue()));
}

template<typename ValType>
void Property<ValType>::fromString(const QString& data)
{
	bool ok = false;
	setValue(StringConv::convertString<value_type>(StringUtils::decodeEscapeCharacters(data), &ok));

	if (!ok)
		throw PropertyException{this, _EXCPT(QString{"Invalid string data passed ('%1')"}.arg(data))};
}

template<typename ValType>
void Property<ValType>::serialize(SerializationContext& ctx) const
{
	PropertyBase::serialize(ctx);

	// Serialize values
	if (!hasFlag(Flag::ReadOnly))
		ctx.settings()(Serialization_Value_Value) = toString();
}

template<typename ValType>
void Property<ValType>::deserialize(DeserializationContext& ctx)
{
	PropertyBase::deserialize(ctx);

	// Deserialize values
	if (!hasFlag(Flag::ReadOnly))
		fromString(ctx.settings()(Serialization_Value_Value).toString());
}

template<typename ValType>
template<typename V, typename Setter>
void Property<ValType>::_setValue(V val, Setter setter)
{
	if (hasFlag(Flag::ReadOnly))
		throw PropertyException{this, _EXCPT("Trying to modify a read-only property")};

	if (_value != val)
	{
		auto oldValue = _value;

		setter();
		applyConstraints();

		if (_value != oldValue)
			emit valueChanged();
	}
}

template<typename ValType>
void Property<ValType>::applyConstraints()
{
	for (const auto& constraint : _constraints)
	{
		// Cast the generic constraint to a constraint of the same value type as this property
		if (auto valuedConstraint = dynamic_cast<PropertyConstraint<value_type>*>(constraint.get()))
			valuedConstraint->applyConstraint(_value);	// Constraints will modify the passed value if needed
	}
}
