/******************************************************************************
 * File: ImageEditorController.cpp
 * Date: 15.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageEditorController.h"
#include "core/GrinderApplication.h"
#include "image/ImageTags.h"
#include "image/ImageExceptions.h"
#include "image/draftitems/PixelsDraftItem.h"
#include "ui/image/ImageEditor.h"
#include "ui/image/LayersListWidget.h"
#include "ui/LongOperationIndicator.h"
#include "util/MathUtils.h"

ImageEditorController::ImageEditorController(ImageEditor* imageEditor) : GenericController("Image editor"), ImageEditorComponent(imageEditor)
{
	if (!imageEditor)
		throw std::invalid_argument{_EXCPT("imageEditor may not be null")};
}

void ImageEditorController::assignUiComponents(ImageEditorView* view, LayersListWidget* layersList)
{
	if (!view)
		throw std::invalid_argument{_EXCPT("view may not be null")};

	if (!layersList)
		throw std::invalid_argument{_EXCPT("layersList may not be null")};

	_view = view;
	_layersList = layersList;
}

void ImageEditorController::showImageBuild(ImageBuild* build)
{
	if (build == _activeImageBuild)
		return;

	if (!_view)
		throw std::runtime_error{_EXCPT("No view has been set")};

	if (_activeImageBuild)	// No longer listen for events from the previous image build
	{
		disconnect(_activeImageBuild, nullptr, this, nullptr);
		connectLayerSignals(_activeImageBuild, false);
	}

	if (_activeImageBuild)
		emit imageBuildSwitching(_activeImageBuild);

	_activeImageBuild = build;

	// Unset any currently active scene
	_view->setEditorScene(nullptr);
	_activeScene.reset(nullptr);

	// Unset any currently active layer
	switchLayer(nullptr);

	if (build)	// Create a new scene for the given image build and show it
	{
		_activeScene = std::make_unique<ImageEditorScene>(_imageEditor, _activeImageBuild, _view);
		_activeScene->buildScene();

		_view->setEditorScene(_activeScene.get());

		// Listen for image build events
		connect(_activeImageBuild, &ImageBuild::imageDataChanged, this, &ImageEditorController::refreshEditorImage);
		connect(_activeImageBuild, &ImageBuild::layerCreated, this, &ImageEditorController::layerCreated);
		connect(_activeImageBuild, &ImageBuild::layerRemoved, this, &ImageEditorController::layerRemoved);
		connect(_activeImageBuild, &ImageBuild::layerMoved, this, &ImageEditorController::layerMoved);

		// Listen for signals from all existing layers of the image build
		connectLayerSignals(_activeImageBuild);

		// Show the new editor image and fit it to its window if the corresponding option is turned on
		refreshEditorImage();

		if (grinder()->settings().imageEditorOptions().autoFitToWindow)
			_activeScene->fitViewToImage();
	}

	emit imageBuildSwitched(_activeImageBuild);
}

void ImageEditorController::switchLayer(Layer* layer)
{
	if (layer == _activeLayer)
		return;

	if (_activeLayer)
		emit layerSwitching(_activeLayer);

	_activeLayer = layer;
	emit layerSwitched(_activeLayer);
}

void ImageEditorController::copyImageBuild(const ImageBuild* imageBuild) const
{
	// Use the currently active image build if non was provided
	if (!imageBuild)
		imageBuild = _activeImageBuild;

	if (imageBuild)
	{
		callControllerFunction("Copying image build", [imageBuild]() {
			grinder()->clipboardManager().serialize<ImageBuild>(ImageBuildVector::Serialization_Element, {imageBuild});
			return true;
		});
	}
}

void ImageEditorController::pasteImageBuild(ImageBuild* imageBuild)
{
	// Use the currently active image build if non was provided
	if (!imageBuild)
		imageBuild = _activeImageBuild;

	if (imageBuild)
	{
		callControllerFunction("Pasting an image build", [this, imageBuild]() {
			clearImageBuild(imageBuild);

			grinder()->clipboardManager().deserialize<ImageBuild>(ImageBuildVector::Serialization_Element, [imageBuild](const SettingsContainer& settings) {
				Q_UNUSED(settings);
				return imageBuild;
			});

			return true;
		});
	}
}

void ImageEditorController::duplicateImageBuild(const ImageBuild* sourceImageBuild, ImageBuild* targetImageBuild, bool switchToTargetImageBuild)
{
	// If no image source image build was provided, use the currently active one
	if (!sourceImageBuild)
		sourceImageBuild = _activeImageBuild;

	// Get the next image reference
	auto currentImageRef = grinder()->projectController().activeImageReference();
	auto nextImageRef = grinder()->project().imageReferences().getNeighbor(currentImageRef, ImageReferenceVector::Neighbor::Next);

	// // Use the image build for the next image if non was provided
	if (!targetImageBuild)
	{
		// Get the correct image build for the next image reference
		if (const auto activeLabel = grinder()->projectController().activeLabel())
			targetImageBuild = activeLabel->imageBuildPool().imageBuild(sourceImageBuild->block(), nextImageRef.get()).get();
	}

	if (targetImageBuild)
	{
		duplicateImageBuild(sourceImageBuild, std::vector<ImageBuild*>{targetImageBuild});

		if (switchToTargetImageBuild)
		{
			// Switch to the next image reference and execute the current label
			grinder()->projectController().switchImageReference(nextImageRef.get());
			grinder()->engineController().executeLabel(nullptr, nullptr, Engine::ExecutionMode::View);
		}
	}
}

void ImageEditorController::duplicateImageBuild(const ImageBuild* sourceImageBuild, std::vector<ImageReference*> targetImageReferences)
{
	// If no image source image build was provided, use the currently active one
	if (!sourceImageBuild)
		sourceImageBuild = _activeImageBuild;

	if (sourceImageBuild && !targetImageReferences.empty())
	{
		std::vector<ImageBuild*> targetImageBuilds;

		// Get the correct image builds for the selected image references
		if (const auto activeLabel = grinder()->projectController().activeLabel())
		{
			for (auto imageRef : targetImageReferences)
			{
				if (auto targetImageBuild = activeLabel->imageBuildPool().imageBuild(sourceImageBuild->block(), imageRef).get())
					targetImageBuilds.push_back(targetImageBuild);
			}
		}

		duplicateImageBuild(sourceImageBuild, targetImageBuilds);
	}
}

std::shared_ptr<Layer> ImageEditorController::createLayer(QString name) const
{
	return callControllerFunction("Creating a new layer", [this](QString name) {
		return _activeImageBuild->createLayer(name);
	}, name);
}

void ImageEditorController::copyLayer(const Layer* layer) const
{
	callControllerFunction("Copying a layer", [this](const Layer* layer) {
		grinder()->clipboardManager().serialize<Layer>(LayerVector::Serialization_Element, {layer});
		return true;
	}, layer);
}

void ImageEditorController::pasteLayer()
{
	callControllerFunction("Pasting a layer", [this]() {
		auto layers = grinder()->clipboardManager().deserialize<Layer>(LayerVector::Serialization_Element, [this](const SettingsContainer& settings) {
			QString name = settings(Layer::Serialization_Value_Name).toString();

			while (_activeImageBuild->layers().selectByName(name))
				name += " - Copy";

			return createLayer(name).get();
		});

		// Switch to the last pasted layer
		if (!layers.empty())
			switchLayer(layers.back());

		return true;
	});
}

void ImageEditorController::cutLayer(const Layer* layer)
{
	copyLayer(layer);
	removeLayer(layer);
}

void ImageEditorController::removeLayer(const Layer* layer, bool enforceRemoval)
{
	if (layer && (layer->isRemovable() || enforceRemoval))
	{
		// If removing the active layer, unset it first
		if (_activeLayer == layer)
			switchLayer(nullptr);

		callControllerFunction("Removing a layer", [this](const Layer* layer) {
			_activeImageBuild->removeLayer(layer);
			return true;
		}, layer);
	}
}

void ImageEditorController::removeAllLayers(bool enforceRemoval)
{
	if (_activeImageBuild)		
	{
		if (!enforceRemoval)
		{
			// Only remove all removable layers
			std::vector<Layer*> layersToRemove;

			for (auto layer : _activeImageBuild->layers())
			{
				if (layer->isRemovable())
					layersToRemove.push_back(layer.get());
			}

			for (auto layer : layersToRemove)
				removeLayer(layer, true);
		}
		else
			_activeImageBuild->removeAllLayers();
	}
}

std::shared_ptr<DraftItem> ImageEditorController::createDraftItem(DraftItemType type, Layer* layer, bool autoCreateLayer)
{
	if (!layer)
		layer = _activeLayer;

	// If there are no layers, create a default one and activate it
	if (!layer && autoCreateLayer)
		layer = createDefaultLayer();

	if (layer)
	{
		return callControllerFunction("Creating a draft item", [this](DraftItemType type, Layer* layer) {
			return layer->createDraftItem(type);
		}, type, layer);
	}
	else
		return nullptr;
}

void ImageEditorController::removeDraftItem(const DraftItem* item) const
{
	auto layer = const_cast<Layer*>(item->layer());	// We don't modify the item itself but its parent layer, so remove constness

	if (item)
	{
		callControllerFunction("Removing a draft item", [this](const DraftItem* item, Layer* layer) {
			layer->removeDraftItem(item);
			return true;
		}, item, layer);
	}
}

void ImageEditorController::setLayerVisibility(Layer* layer, bool visible) const
{
	layer->setVisible(visible);
}

void ImageEditorController::setLayerAlpha(Layer* layer, float alpha) const
{
	layer->setAlpha(alpha);
}

void ImageEditorController::moveLayer(Layer* layer, bool up) const
{
	callControllerFunction("Moving a layer", [this](Layer* layer, bool up) {
		_activeImageBuild->moveLayer(layer, up);
		return true;
	}, layer, up);
}

void ImageEditorController::renameLayer(Layer* layer, QString newName) const
{
	callControllerFunction("Renaming a layer", [this](Layer* layer, QString newName) {
		validateLayerName(newName, layer);
		renameImageBuildItem(layer, newName);
		return true;
	}, layer, newName);
}

void ImageEditorController::renameImageBuildItem(ImageBuildItem* item, QString newName) const
{
	item->setName(newName);
}

void ImageEditorController::setNodesPrimaryColor(QColor color, const std::vector<DraftItemNode*>& nodes) const
{
	if (_activeScene)
	{
		for (const auto& node : nodes)
		{
			if (auto item = node->draftItem().lock())
				item->primaryColor()->setValue(color);
		}
	}
}

void ImageEditorController::convertSelectedNodesToPixels() const
{
	callControllerFunction("Converting items to pixels", [this]() {				
		// We need to group all selected nodes into their respective layers
		std::map<Layer*, std::vector<DraftItem*>> itemNodes;

		for (auto node : _activeScene->getNodes<DraftItemNode>(true))
		{
			if (auto draftItem = node->draftItem().lock())
				itemNodes[draftItem->layer()].push_back(draftItem.get());
		}

		LongOperationIndicator longOperation{"Converting items to pixels...", itemNodes.size()};

		// Render all draft items on their layers
		for (auto it : itemNodes)
		{
			it.first->layerPixels().renderDraftItems(it.second, _activeScene->view()->sceneStyle().getItemNodeStyle());
			++longOperation;
		}

		// Remove all selected draft items after converting them to pixels
		removeSelectedNodes();

		return true;
	});
}

void ImageEditorController::convertLayerPixelsToItems(Layer* layer)
{
	if (!layer)
		layer = _activeLayer;

	if (layer)
	{
		callControllerFunction("Converting pixels to items", [this](Layer* layer) {
			LongOperationIndicator longOperation{"Converting pixels to items..."};

			_activeScene->clearSelection();

			// Create draft items for each region of each color
			for (auto color : layer->layerPixels().data().getColors())
			{
				for (auto& region : layer->layerPixels().data().at(color).regions())
				{
					auto draftItem = createDraftItem(DraftItemType::Pixels, layer);

					if (auto pixelsItem = dynamic_cast<PixelsDraftItem*>(draftItem.get()))
					{
						// Locate the new item at the pixels' top-left corner
						auto bounds = region.bounds();
						pixelsItem->position()->setValue(QPoint{bounds.first.first, bounds.second.first});

						// Move the region pixels to (0,0) so they are placed at the item's origin
						region.moveTo(0, 0);
						pixelsItem->pixelsData().set(region, color);
					}

					// Select the created draft items
					if (auto draftItemNode = _activeScene->findDraftItemNode(draftItem.get()))
						draftItemNode->setSelected(true);
				}
			}

			// Clear all pixels after converting them to items
			layer->layerPixels().data().clear();

			return true;
		}, layer);
	}
}

void ImageEditorController::paintPixels(Layer* layer, QColor color, const std::vector<QPoint>& pixels, bool autoCreateLayer)
{
	if (!layer)
		layer = _activeLayer;

	// If there are no layers, create a default one and activate it
	if (!layer && autoCreateLayer)
		layer = createDefaultLayer();

	if (layer)
	{
		callControllerFunction("Painting pixels", [this](Layer* layer, QColor color, const std::vector<QPoint>& pixels) {
			auto imgSize = QSize{_activeImageBuild->imageData().cols, _activeImageBuild->imageData().rows};

			layer->layerPixels().data().beginPainting();

			for (auto pos : pixels)
			{
				if (pos.x() < 0 || pos.x() >= imgSize.width() || pos.y() < 0 || pos.y() >= imgSize.height())	// Pixel is out-of-bounds
					continue;

				if (color.isValid())
				{
					// Draw on the layer's pixels
					if (!layer->layerPixels().data().isSet(pos) || layer->layerPixels().data().get(pos) != color)	// Don't perform unnecessary pixel operations
						layer->layerPixels().data().set(pos, color);
				}
				else
				{
					// Erase the layer's pixels
					if (layer->layerPixels().data().isSet(pos))	// Don't perform unnecessary pixel operations
						layer->layerPixels().data().clear(pos);
				}
			}

			layer->layerPixels().data().endPainting();
			return true;
		}, layer, color, pixels);
	}
}

void ImageEditorController::copySelectedNodes() const
{
	callControllerFunction("Copying draft items", [this]() {
		// Copy all selected draft items
		auto selItems = _activeScene->getNodes<DraftItemNode>(true);
		std::vector<const DraftItem*> draftItems;

		for (const auto& node : selItems)
		{
			if (const auto& item = node->draftItem().lock())
				draftItems.push_back(item.get());
		}

		grinder()->clipboardManager().serialize(DraftItemVector::Serialization_Element, draftItems);
		return true;
	});
}

void ImageEditorController::pasteSelectedNodes()
{
	callControllerFunction("Pasting draft items", [this]() {
		// Create draft items for each object in the clipboard
		auto draftItems = grinder()->clipboardManager().deserialize<DraftItem>(DraftItemVector::Serialization_Element, [this](const SettingsContainer& settings) {
			DraftItemType type = settings(DraftItem::Serialization_Value_Type, DraftItemType::Undefined).toString();

			return createDraftItem(type).get();
		});

		// Select the pasted items
		if (!draftItems.empty())
		{
			_activeScene->clearSelection();

			for (const auto& draftItem : draftItems)
			{
				if (auto draftItemNode = _activeScene->findDraftItemNode(draftItem))
					draftItemNode->setSelected(true);
			}
		}

		return true;
	});
}

void ImageEditorController::cutSelectedNodes() const
{
	copySelectedNodes();
	removeSelectedNodes();
}

void ImageEditorController::removeSelectedNodes() const
{
	if (_activeScene)
	{
		// Remove all selected draft items
		for (auto node : _activeScene->getNodes<DraftItemNode>(true))
		{
			if (auto item = node->draftItem().lock())
				removeDraftItem(item.get());
		}
	}
}

void ImageEditorController::assignImageTags(ImageTagsAllotmentProperty* imageTagsAllotment, const QStringList& imageTagNames, bool clearFirst) const
{
	callControllerFunction("Assigning tags", [this](ImageTagsAllotmentProperty* imageTagsAllotment, const QStringList& imageTagNames, bool clearFirst) {
		if (auto inputImageTags = _activeImageBuild->inputImageTags())
		{
			std::set<ImageTag*> imageTags;

			// Search for the given image tags by their names
			for (auto imageTagName : imageTagNames)
			{
				if (auto imageTag = inputImageTags->tags().selectByName(imageTagName))
					imageTags.emplace(imageTag.get());
			}

			if (!imageTags.empty())
				assignImageTags(imageTagsAllotment, imageTags, clearFirst);
		}

		return true;
	}, imageTagsAllotment, imageTagNames, clearFirst);
}

void ImageEditorController::assignImageTags(ImageTagsAllotmentProperty* imageTagsAllotment, const std::set<ImageTag*>& imageTags, bool clearFirst) const
{
	if (clearFirst)
		imageTagsAllotment->object().unallotAllTags();

	for (auto imageTag : imageTags)
		imageTagsAllotment->object().allotTag(imageTag);
}

void ImageEditorController::controllerFunctionCalled() const
{
	// Redraw the scene after calling a controller function
	if (_activeScene)
		_activeScene->update();
}

void ImageEditorController::duplicateImageBuild(const ImageBuild* sourceImageBuild, std::vector<ImageBuild*> targetImageBuilds)
{
	if (!sourceImageBuild)
		throw std::invalid_argument{_EXCPT("sourceImageBuild may not be null")};

	if (!targetImageBuilds.empty())
	{
		auto clipboardContent = grinder()->clipboardManager().saveClipboard();	// Save the current clipboard contents to restore them afterwards
		copyImageBuild(sourceImageBuild);

		LongOperationIndicator longOperation{"Duplicating image builds...", targetImageBuilds.size(), true};

		for (unsigned int i = 0; i < targetImageBuilds.size(); ++i)
		{
			auto imageBuild = targetImageBuilds[i];

			if (imageBuild != sourceImageBuild)
				pasteImageBuild(imageBuild);

			if ((++longOperation).wasCanceled())
				break;
		}

		grinder()->clipboardManager().restoreClipboard(clipboardContent);
	}
}

void ImageEditorController::clearImageBuild(ImageBuild* imageBuild)
{
	callControllerFunction("Clearing an image build", [this](ImageBuild* imageBuild) {
		// Unset any active layer from the current image build
		if (_activeImageBuild == imageBuild)
			switchLayer(nullptr);

		imageBuild->removeAllLayers();
		return true;
	}, imageBuild);
}

Layer* ImageEditorController::createDefaultLayer()
{
	auto layer = createLayer("Default layer").get();
	switchLayer(layer);
	return layer;
}

void ImageEditorController::validateLayerName(QString name, const Layer* layer) const
{
	auto imageBuild = layer ? layer->imageBuild() : _activeImageBuild;

	// Name must be non-empty and unique
	if (!name.isEmpty())
	{
		auto existingLayer = imageBuild->layers().selectByName(name).get();

		if (existingLayer && existingLayer != layer)
			throw ImageBuildException{imageBuild, _EXCPT(QString{"A layer with the name '%1' already exists"}.arg(name))};
	}
	else
		throw ImageBuildException{imageBuild, _EXCPT("The layer name may not be empty")};
}

void ImageEditorController::connectLayerSignals(ImageBuild* imageBuild, bool connectSignals) const
{
	// Listen for/stop listening for events from all layers
	for (auto& layer : imageBuild->layers())
		connectLayerSignals(layer.get(), connectSignals);
}

void ImageEditorController::connectLayerSignals(Layer* layer, bool connectSignals) const
{
	if (connectSignals)
	{
		// Get notified about layer and draft item events
		connect(layer, &Layer::draftItemCreated, this, &ImageEditorController::draftItemCreated);
		connect(layer, &Layer::draftItemRemoved, this, &ImageEditorController::draftItemRemoved);
		connect(layer, &Layer::layerShown, this, &ImageEditorController::layerVisibilityChanged);
		connect(layer, &Layer::layerHidden, this, &ImageEditorController::layerVisibilityChanged);
		connect(layer, &Layer::layerAlphaChanged, this, &ImageEditorController::layerAlphaChanged);
	}
	else
		disconnect(layer, nullptr, this, nullptr);
}

void ImageEditorController::refreshEditorImage()
{
	callControllerFunction("Refreshing editor image", [this]() {
		if (_activeScene)
			_activeScene->refreshImage(true);

		return true;
	});
}

void ImageEditorController::layerCreated(const std::shared_ptr<Layer>& layer) const
{
	// Create the layer pixels node for the new layer
	if (_activeScene)
		_activeScene->createLayerPixelsNode(layer);

	connectLayerSignals(layer.get());

	if (_layersList)
		_layersList->addObject(layer.get(), true, true);
}

void ImageEditorController::layerRemoved(const std::shared_ptr<Layer>& layer) const
{
	connectLayerSignals(layer.get(), false);

	// Remove all item nodes that belong to the removed layer
	if (_activeScene)
	{
		for (const auto& item : layer->draftItems())
			_activeScene->removeDraftItemNode(item);

		_activeScene->removeLayerPixelsNode(layer);
	}

	if (_layersList)
		_layersList->removeObject(layer.get());
}

void ImageEditorController::layerMoved(const std::shared_ptr<Layer>& layer, int indexOld, int indexNew) const
{
	Q_UNUSED(layer);

	// Reflect the new z-order in the scene
	if (_activeScene)
		_activeScene->updateLayerItemsOrder();

	if (_layersList)
		_layersList->swapLayers(indexOld, indexNew);
}

void ImageEditorController::layerVisibilityChanged() const
{
	// Update the visibility of all items in the layer
	if (_activeScene)
		_activeScene->updateLayerItemsVisibility(dynamic_cast<Layer*>(sender()));
}

void ImageEditorController::layerAlphaChanged(float alpha) const
{
	Q_UNUSED(alpha);

	// Update the alpha value of all items in the layer
	if (_activeScene)
		_activeScene->updateLayerItemsAlpha(dynamic_cast<Layer*>(sender()));
}

void ImageEditorController::draftItemCreated(const std::shared_ptr<DraftItem>& item) const
{
	if (_activeScene)
		_activeScene->createDraftItemNode(item);
}

void ImageEditorController::draftItemRemoved(const std::shared_ptr<DraftItem>& item) const
{
	if (_activeScene)
		_activeScene->removeDraftItemNode(item);
}
