/******************************************************************************
 * File: ProjectController.cpp
 * Date: 07.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ProjectController.h"
#include "PipelineController.h"
#include "core/GrinderApplication.h"
#include "project/Project.h"
#include "project/ProjectExceptions.h"
#include "project/ProjectSerializer.h"
#include "common/serialization/JsonSettingsCodec.h"
#include "ui/mainwnd/LabelsListWidget.h"
#include "ui/mainwnd/ImageReferencesListWidget.h"
#include "ui/LongOperationIndicator.h"
#include "util/ImageUtils.h"
#include "util/UIUtils.h"

ProjectController::ProjectController(Project* project) : GenericController("Project"),
	_project{project}
{
	if (!project)
		throw std::invalid_argument{_EXCPT("project may not be null")};

	// Listen for project events
	connect(_project, &Project::labelCreated, this, &ProjectController::labelCreated);
	connect(_project, &Project::labelRemoved, this, &ProjectController::labelRemoved);
	connect(_project, &Project::imageReferenceCreated, this, &ProjectController::imageReferenceCreated);
	connect(_project, &Project::imageReferenceRemoved, this, &ProjectController::imageReferenceRemoved);
}

void ProjectController::assignUiComponents(LabelsListWidget* labelsList, ImageReferencesListWidget* imageRefsList)
{
	if (!labelsList)
		throw std::invalid_argument{_EXCPT("labelsList may not be null")};

	if (!imageRefsList)
		throw std::invalid_argument{_EXCPT("imageRefsList may not be null")};

	_labelsList = labelsList;
	_imageReferencesList = imageRefsList;
}

void ProjectController::clearProject(bool createDefaultLabel)
{
	emit projectClearing();

	_suppressErrors = true;

	// Before clearing the project, make sure that no label or image reference is active
	switchLabel(nullptr);
	switchImageReference(nullptr);

	_project->clear();

	if (createDefaultLabel)
	{
		if (auto label = grinder()->projectController().createLabel("Default label"))
			switchLabel(label.get());
	}

	_suppressErrors = false;

	setCurrentProjectFile("");
	emit projectCleared();
}

void ProjectController::loadProject(QString fileName)
{
	emit projectLoading(fileName);

	clearProject();

	// Delay the actual loading till all internal signals have been processed in order to give Qt a chance to really clear the previous project
	QTimer::singleShot(0, [this, fileName] {
		bool result = callControllerFunction("Loading a project", [this](QString fileName) {
			LongOperationIndicator longOperation{QString{"Loading project '%1'..."}.arg(fileName)};

			JsonSettingsCodec codec;
			SettingsContainer projectSettings;
			codec.loadContainer(projectSettings, fileName);

			ProjectSerializer serializer{_project};
			serializer.deserializeProject(projectSettings, QFileInfo{fileName}.dir());

			return true;
		}, fileName);

		if (result)
		{
			setCurrentProjectFile(fileName);
			emit projectLoaded(fileName);

			// If no label is active, activate the first one after loading
			if (!_activeLabel && !_project->labels().empty())
				switchLabel(_project->labels().front().get());

			// If no image reference is active, activate the first one after loading
			if (!_activeImageReference && !_project->imageReferences().empty())
				switchImageReference(_project->imageReferences().front().get());
		}
		else
		{
			clearProject();
			throw ProjectException{_project, _EXCPT(QString{"Failed to load project '%1'"}.arg(fileName))};
		}
	});
}

void ProjectController::saveProject(QString fileName)
{
	emit projectSaving(fileName);

	bool result = callControllerFunction("Saving the current project", [this](QString fileName) {
		LongOperationIndicator longOperation{QString{"Saving project to '%1'..."}.arg(fileName)};

		ProjectSerializer serializer{_project};
		auto projectSettings = serializer.serializeProject();

		JsonSettingsCodec codec;
		codec.saveContainer(projectSettings, fileName);

		return true;
	}, fileName);

	if (result)
	{
		setCurrentProjectFile(fileName);
		emit projectSaved(fileName);
	}
	else
		throw ProjectException{_project, _EXCPT(QString{"Failed to save the current project to '%1'"}.arg(fileName))};
}

bool ProjectController::isProjectDirty() const
{
	// If there are no current settings, the application has just been started
	if (_currentProjectSettings.isEmpty())
		return false;

	// Compare the current project data to the saved data; if they differ, the project has been changed
	ProjectSerializer serializer{_project};
	auto projectSettings = serializer.serializeProject();

	return projectSettings != _currentProjectSettings;
}

void ProjectController::switchLabel(Label* label)
{
	if (label == _activeLabel)
		return;

	if (_activeLabel)
		emit labelSwitching(_activeLabel);

	_activeLabel = label;

	// Unset any currently active pipeline
	grinder()->pipelineController().hidePipeline();

	if (label)	// Show the pipeline associated with the label
		grinder()->pipelineController().showPipeline(label->pipeline());

	emit labelSwitched(_activeLabel);
}

void ProjectController::switchImageReference(ImageReference* imageRef)
{
	if (imageRef == _activeImageReference)
		return;

	if (_activeImageReference)
		emit imageReferenceSwitching(_activeImageReference);

	_activeImageReference = imageRef;
	emit imageReferenceSwitched(_activeImageReference);
}

void ProjectController::switchToNextImageReference()
{
	if (auto nextImageRef = _project->imageReferences().getNeighbor(_activeImageReference, ImageReferenceVector::Neighbor::Next))
		switchImageReference(nextImageRef.get());
}

void ProjectController::switchToPreviousImageReference()
{
	if (auto prevImageRef = _project->imageReferences().getNeighbor(_activeImageReference, ImageReferenceVector::Neighbor::Previous))
		switchImageReference(prevImageRef.get());
}

std::shared_ptr<Label> ProjectController::createLabel(QString name) const
{
	return callControllerFunction("Creating a new label", [this](QString name) {
		validateLabelName(name);
		return _project->createLabel(name);
	}, name);
}

void ProjectController::removeLabel(const Label* label)
{
	if (label)
	{
		// If removing the active label, hide it first
		if (_activeLabel == label)
			switchLabel(nullptr);

		callControllerFunction("Removing a label", [this](const Label* label) {
			_project->removeLabel(label);
			return true;
		}, label);
	}
}

void ProjectController::removeAllLabels()
{
	auto labels = _project->labels();

	for (const auto& label : labels)
		removeLabel(label.get());
}

void ProjectController::renameLabel(Label* label, QString newName) const
{
	callControllerFunction("Renaming a label", [this](Label* label, QString newName) {
		validateLabelName(newName, label);
		label->setName(newName);
		return true;
	}, label, newName);
}

std::shared_ptr<ImageReference> ProjectController::createImageReference(QString imagePath) const
{
	return callControllerFunction("Creating an image reference", [this](QString imagePath) {
		return _project->createImageReference(imagePath);
	}, imagePath);
}

std::vector<std::shared_ptr<ImageReference>> ProjectController::createImageReferences(QStringList imagePaths, bool ignoreUnknownFiletypes) const
{	
	static auto imageFormats = ImageUtils::getSupportedFormats();

	LongOperationIndicator longOperation{"Adding images...", static_cast<unsigned int>(imagePaths.size()), true};
	std::vector<std::shared_ptr<ImageReference>> images;

	for (auto imagePath : imagePaths)
	{
		if (ignoreUnknownFiletypes)
		{
			QFileInfo fileInfo{imagePath};

			if (!imageFormats.contains(fileInfo.suffix(), Qt::CaseInsensitive))
			{
				longOperation++;
				continue;
			}
		}

		longOperation.setText(QString{"Adding image '%1'..."}.arg(imagePath));
		images.push_back(createImageReference(imagePath));

		if ((++longOperation).wasCanceled())
			break;
	}

	return images;
}

void ProjectController::removeImageReference(const ImageReference* imageRef)
{
	if (imageRef)
	{
		// If removing the active image reference, unset it first
		if (_activeImageReference == imageRef)
			switchImageReference(nullptr);

		callControllerFunction("Removing an image reference", [this](const ImageReference* imageRef) {
			_project->removeImageReference(imageRef);
			return true;
		}, imageRef);
	}
}

void ProjectController::removeAllImageReferences()
{
	auto imageRefs = _project->imageReferences();

	for (const auto& imageRef : imageRefs)
		removeImageReference(imageRef.get());
}

void ProjectController::validateLabelName(QString name, const Label* label) const
{
	// Name must be non-empty and unique
	if (!name.isEmpty())
	{
		auto existingLabel = _project->labels().selectByName(name).get();

		if (existingLabel && existingLabel != label)
			throw ProjectException{_project, _EXCPT(QString{"A label with the name '%1' already exists"}.arg(name))};
	}
	else
		throw ProjectException{_project, _EXCPT("The label name may not be empty")};
}

void ProjectController::updateCurrentProjectData()
{
	ProjectSerializer serializer{_project};
	_currentProjectSettings = serializer.serializeProject();
}

void ProjectController::labelCreated(const std::shared_ptr<Label>& label) const
{
	if (_labelsList)
		_labelsList->addObject(label.get());
}

void ProjectController::labelRemoved(const std::shared_ptr<Label>& label) const
{
	if (_labelsList)
		_labelsList->removeObject(label.get());
}

void ProjectController::imageReferenceCreated(const std::shared_ptr<ImageReference>& imageRef) const
{
	if (_imageReferencesList)
		_imageReferencesList->addObject(imageRef.get());
}

void ProjectController::imageReferenceRemoved(const std::shared_ptr<ImageReference>& imageRef) const
{
	if (_imageReferencesList)
		_imageReferencesList->removeObject(imageRef.get());
}

void ProjectController::setCurrentProjectFile(QString fileName)
{
	_currentProjectFile = fileName;
	emit currentProjectFileChanged(fileName);

	// Whenever the project file has changed, the current project data needs to be updated, so do it here
	updateCurrentProjectData();
}
