/******************************************************************************
 * File: Pixel.h
 * Date: 19.6.2018
 *****************************************************************************/

#ifndef PIXEL_H
#define PIXEL_H

#include <QColor>
#include <QPoint>
#include <opencv2/core.hpp>

namespace grndr
{
	class ColorPixel
	{
	public:
		ColorPixel() { }
		ColorPixel(uint8_t r, uint8_t g, uint8_t b) : _b{b}, _g{g}, _r{r} { }
		ColorPixel(const QColor& color) { setColor(color); }
		ColorPixel(const ColorPixel& other) = default;
		ColorPixel(ColorPixel&& other) = default;

		ColorPixel& operator =(const QColor& color) { setColor(color); return *this; }

	public:
		QColor getColor() const;
		void setColor(const QColor& color);

		operator QColor() const { return getColor(); }

	public:
		bool operator ==(const ColorPixel& other) const;
		bool operator ==(const QColor& other) const { return ColorPixel{other} == *this; }

	private:
		uint8_t _b{0};
		uint8_t _g{0};
		uint8_t _r{0};
	};

	class GrayscalePixel
	{
	public:
		GrayscalePixel() { }
		GrayscalePixel(uint8_t p) : _p{p} { }
		GrayscalePixel(const QColor& color) { setColor(color); }
		GrayscalePixel(const GrayscalePixel& other) = default;
		GrayscalePixel(GrayscalePixel&& other) = default;

		GrayscalePixel& operator =(uint8_t p) { _p = p; return *this; }
		GrayscalePixel& operator =(const QColor& color) { setColor(color); return *this; }
		GrayscalePixel& operator =(const GrayscalePixel& other) = default;
		GrayscalePixel& operator =(GrayscalePixel&& other) = default;

	public:
		QColor getColor() const;
		void setColor(const QColor& color);

		operator uint8_t() const { return _p; }
		operator QColor() const { return getColor(); }

	public:
		bool operator ==(const GrayscalePixel& other) const { return _p == other._p; }
		bool operator ==(const QColor& other) const { return GrayscalePixel{other} == *this; }

	private:
		uint8_t _p{0};
	};

	class PixelRef
	{
	public:
		PixelRef(ColorPixel* pixel);
		PixelRef(GrayscalePixel* grayscalePixel);

	public:
		PixelRef& operator =(const QColor& color) { set(color); return *this; }

	public:
		QColor get() const { return _pixelGetter(this); }
		void set(const QColor& color) { _pixelSetter(this, color); }

		operator QColor() const { return get(); }

	private:
		QColor getPixel() const { return _pixel->getColor(); }
		void setPixel(const QColor& color) { _pixel->setColor(color); }
		QColor getGrayscalePixel() const { return _grayscalePixel->getColor(); }
		void setGrayscalePixel(const QColor& color) { _grayscalePixel->setColor(color); }

	private:
		ColorPixel* _pixel{nullptr};
		GrayscalePixel* _grayscalePixel{nullptr};

		std::function<QColor(const PixelRef*)> _pixelGetter{nullptr};
		std::function<void(PixelRef*, QColor)> _pixelSetter{nullptr};
	};

	class PixelAccessor
	{
	public:
		PixelAccessor(cv::Mat matrix) : _matrix{matrix} { verifyMatrix(); }

	public:
		PixelRef at(int x, int y);
		PixelRef at(QPoint pos) { return at(pos.x(), pos.y()); }
		void forEach(std::function<void(PixelRef, QPoint)> callback);

		PixelRef operator [](QPoint pos) { return at(pos); }

	private:
		void verifyMatrix();

	private:
		cv::Mat _matrix;
	};
}

namespace cv
{
	template<>
	class DataType<grndr::ColorPixel>
	{
	public:
		using value_type = grndr::ColorPixel;
		using work_type = typename cv::DataType<uint8_t>::work_type;
		using channel_type = uint8_t;

		enum
		{
			generic_type = 0,
			channels = 3,
			fmt = cv::traits::SafeFmt<channel_type>::fmt + ((channels - 1) << 8)
		};

		using vec_type = Vec<channel_type, channels>;
	};

	template<>
	class DataType<grndr::GrayscalePixel>
	{
	public:
		using value_type = grndr::GrayscalePixel;
		using work_type = typename cv::DataType<uint8_t>::work_type;
		using channel_type = uint8_t;

		enum
		{
			generic_type = 0,
			channels = 1,
			fmt = cv::traits::SafeFmt<channel_type>::fmt + ((channels - 1) << 8)
		};

		using vec_type = Vec<channel_type, channels>;
	};

	namespace traits
	{
		template<>
		struct Depth<grndr::ColorPixel> { enum { value = Depth<uint8_t>::value }; };
		template<>
		struct Type<grndr::ColorPixel> { enum { value = CV_MAKETYPE(Depth<uint8_t>::value, 3) }; };

		template<>
		struct Depth<grndr::GrayscalePixel> { enum { value = Depth<uint8_t>::value }; };
		template<>
		struct Type<grndr::GrayscalePixel> { enum { value = CV_MAKETYPE(Depth<uint8_t>::value, 3) }; };
	}
}

#endif
