/******************************************************************************
 * File: Engine.cpp
 * Date: 21.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "Engine.h"
#include "EngineExecutionContext.h"
#include "EngineExceptions.h"
#include "Processor.h"
#include "pipeline/Pipeline.h"
#include "project/Label.h"

void Engine::executeLabel(Label* label, const ImageReference* activeImageReference, ExecutionMode mode) const
{
	executeLabelEx(label, nullptr, activeImageReference, mode);
}

cv::Mat Engine::executeLabelEx(Label* label, const Port* requestPort, const ImageReference* activeImageReference, ExecutionMode mode) const
{
	cv::Mat portData;

	try {
		// We first need a block hierarchy to execute
		BlockHierarchy blockHierarchy{label->pipeline()};

		if (!blockHierarchy.empty())
		{
			EngineExecutionContext ctx{this, label, activeImageReference, mode};

			for (unsigned int currentLevel = 0; currentLevel < blockHierarchy.size(); ++currentLevel)
			{
				executeHierarchyLevel(blockHierarchy.at(currentLevel), ctx, requestPort ? requestPort->block() : nullptr);

				// Clean up unused data in the context
				ctx.purge(blockHierarchy, currentLevel);

				// If data at a specific port was requested, copy it to portData
				if (requestPort)
				{
					if (auto dataBlob = ctx.contextEntry(requestPort))
						portData = dataBlob->getMatrix();
				}

				if (ctx.wasAborted())
					break;
			}
		}
	} catch (std::exception& e) {
		// Forward any exceptions as an EngineException
		throw EngineException{this, _EXCPT(QString{"Executing '%2' failed: %1"}.arg(e.what()).arg(label->getName()))};
	}

	return portData;
}

void Engine::executeHierarchyLevel(const BlockHierarchy::HierarchyLevel& level, EngineExecutionContext& ctx, const Block* finalBlock) const
{
	for (const auto& block : level)
	{
		// Create a processor for the block and execute it
		if (auto processor = block->createProcessor())
			processor->execute(ctx);

		if (block == finalBlock)
		{
			ctx.abortProcessing();
			break;
		}
	}
}
