/******************************************************************************
 * File: ProcessorBase.cpp
 * Date: 21.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ProcessorBase.h"
#include "EngineExceptions.h"
#include "pipeline/Block.h"
#include "pipeline/PipelineExceptions.h"

ProcessorBase::ProcessorBase(const Block* block) :
	_block{block}
{
	if (!block)
		throw std::invalid_argument{_EXCPT("block may not be null")};
}

void ProcessorBase::execute(EngineExecutionContext& ctx)
{
	_engine = ctx.engine();

	if (!_engine)
		throw std::invalid_argument{_EXCPT("Invalid execution context")};
}

DataDescriptor ProcessorBase::getPortDataDescriptor(const Port* port, unsigned int index) const
{
	return port->dataDescriptors().at(index);
}

DataDescriptor ProcessorBase::getBestDataDescriptor(const DataDescriptor& dataDesc, const DataDescriptors& targetDescriptors) const
{
	std::vector<std::pair<DataDescriptor, unsigned int>> convertibleDescriptors;

	// Check all target descriptors for a perfect match or convertible descriptors
	for (const auto& targetDescriptor : targetDescriptors)
	{
		DataDescriptor targetDescAdjusted;

		if (dataDesc.match(targetDescriptor, &targetDescAdjusted))	// Found a perfect match <3
			return targetDescAdjusted;
		else if (dataDesc.canConvertTo(targetDescriptor))	// Not perfect, but convertible
			convertibleDescriptors.push_back(std::make_pair(targetDescriptor, dataDesc.getConversionCosts(targetDescriptor)));
	}

	if (convertibleDescriptors.size() > 0)
	{
		// Get the cheapest descriptor (in terms of needed conversion steps)
		std::sort(convertibleDescriptors.begin(), convertibleDescriptors.end(), [](const auto& desc1, const auto& desc2) { return desc1.second < desc2.second; });
		return convertibleDescriptors.front().first;
	}

	// No valid descriptor found, throw an exception
	throwProcessorException("The input data is not compatible with the expected data type");
	return DataDescriptor{};
}

DataBlob* ProcessorBase::portData(EngineExecutionContext& ctx, const Port* port, bool convert, bool required) const
{
	auto dataPort = resolveDataPort(port, required);
	DataBlob* data = dataPort ? ctx.contextEntry(dataPort) : nullptr;

	if (!data && required)
		throwProcessorException(QString{"No data could be retrieved for port '%1'"}.arg(port->getName()));

	if (convert)
	{
		// Get the best matching target data descriptor; if none can be found, an exception will be thrown
		auto targetDataDesc = getBestDataDescriptor(data->dataDescriptor(), port->dataDescriptors());
		data->convertTo(targetDataDesc);
	}

	return data;
}

void ProcessorBase::throwProcessorException(QString what) const
{
	throw ProcessorException{this, _EXCPT(QString{"%1: %2"}.arg(_block->getName()).arg(what))};
}

const Port* ProcessorBase::resolveDataPort(const Port* port, bool required) const
{
	try {
		auto dataPort = _block->dataPort(port);

		if (!dataPort && required)
			throwProcessorException(QString{"Port '%1' is not connected to a source"}.arg(port->getName()));

		return dataPort;
	} catch (BlockException& e) {
		// Re-throw the exception
		throwProcessorException(GetExceptionMessage(e.what()));
	}

	return nullptr;
}
