/******************************************************************************
 * File: DataBlob.cpp
 * Date: 19.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "DataBlob.h"
#include "DataExceptions.h"
#include "cv/ColorConv.h"

#include <opencv2/imgproc.hpp>

DataBlob::DataBlob(const DataDescriptor& dataDesc, ColorSpace colorSpace) :
	_dataDescriptor{dataDesc}, _colorSpace{colorSpace}
{
	if (dataDesc.isArbitrary())
		throw DataException{_EXCPT("dataDesc may not be arbitrary")};
}

void DataBlob::set(const cv::Mat& data)
{
	if (!data.empty())
		data.copyTo(_data);
	else
		clear();

	updateDataDescriptor();
}

void DataBlob::set(cv::Mat&& data)
{
	_data = std::move(data);
	updateDataDescriptor();
}

void DataBlob::convertTo(const DataDescriptor& dataDesc, bool normalize, double minValue, double maxValue)
{
	if (!_dataDescriptor.canConvertTo(dataDesc) || dataDesc.isDynamic())
		throw DataException{_EXCPT("Invalid conversion")};

	DataDescriptor dataDescNew = _dataDescriptor;

	int channels, depth;
	dataDesc.getCVMatrixType(&channels, &depth);

	try {
		// First, convert the value type if necessary
		if (dataDesc.getValueType() != DataDescriptor::ValueType::Any && depth != _data.depth())
		{
			if (normalize && dataDesc.getValueType() < _dataDescriptor.getValueType())	// Normalize only if the new type is smaller than the current one
				cv::normalize(_data, _data, minValue, maxValue, cv::NORM_MINMAX);

			_data.convertTo(_data, depth);

			// Update the new data descriptor to match the new value type
			dataDescNew = DataDescriptor{dataDescNew.getName(), dataDescNew.getStructureType(), dataDescNew.getFieldType(), dataDesc.getValueType()};
		}

		// Next, convert image colors count if necessary
		if (dataDesc.getFieldType() != DataDescriptor::FieldType::Any && channels != _data.channels())
		{
			// Check if a color <-> grayscale conversion can be done
			if (_dataDescriptor.canConvertToColor(dataDesc))
			{
				// Ensure that we're in RGB color space before going from grayscale to RGB
				setColorSpace(ColorSpace::RGB);

				cv::cvtColor(_data, _data, cv::COLOR_GRAY2BGR);

				// Update the new data descriptor to match the new field type
				dataDescNew = DataDescriptor{dataDescNew.getName(), dataDescNew.getStructureType(), DataDescriptor::FieldType::Color, dataDescNew.getValueType()};
			}
			else if (_dataDescriptor.canConvertToGrayscale(dataDesc))
			{
				// Ensure that we're in RGB color space before going to grayscale
				setColorSpace(ColorSpace::RGB);

				cv::cvtColor(_data, _data, cv::COLOR_BGR2GRAY);

				// Update the new data descriptor to match the new field type
				dataDescNew = DataDescriptor{dataDescNew.getName(), dataDescNew.getStructureType(), DataDescriptor::FieldType::Basic, dataDescNew.getValueType()};
			}
		}

		// Set the new data descriptor to match the converted type
		_dataDescriptor = dataDescNew;
	} catch (std::exception& e) {
		// Forward exceptions from OpenCV as a DataException
		throw DataException{_EXCPT(e.what())};
	}
}

void DataBlob::setColorSpace(ColorSpace colorSpace)
{
	if (colorSpace != _colorSpace)
	{
		try {
			ColorConv::convertColorSpace(_data, _colorSpace, colorSpace);
		} catch (std::exception& e) {
			// Forward exceptions from OpenCV as a DataException
			throw DataException{_EXCPT(e.what())};
		}

		_colorSpace = colorSpace;
	}
}

void DataBlob::updateDataDescriptor()
{
	_dataDescriptor.fromCVMatrixType(_data);
}
