/******************************************************************************
 * File: DataDescriptor.cpp
 * Date: 15.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "DataDescriptor.h"

#include <opencv2/core.hpp>

DataDescriptor DataDescriptor::arbitraryDescriptor(QString name)
{
	return DataDescriptor{name, StructureType::Any, FieldType::Any, ValueType::Any};
}

DataDescriptor DataDescriptor::imageDescriptor(bool colorImage, ValueType valueType)
{
	return DataDescriptor{colorImage ? "Color image" : "Grayscale image", StructureType::Matrix, colorImage ? FieldType::Color : FieldType::Basic, valueType};
}

DataDescriptor DataDescriptor::dynamicOutputDescriptor(QString name, StructureType structType)
{
	return DataDescriptor{name, structType, FieldType::Dynamic, ValueType::Dynamic};
}

DataDescriptor::DataDescriptor(QString name, StructureType structureType, FieldType fieldType, ValueType valueType) :
	_name{name}, _structureType{structureType}, _fieldType{fieldType}, _valueType{valueType}
{

}

bool DataDescriptor::canConvertTo(const DataDescriptor& dataDesc) const
{
	// All valid structure type conversions
	static const std::vector<std::pair<DataDescriptor::StructureType, DataDescriptor::StructureType>> validStructureTypeConversions = {
		{DataDescriptor::StructureType::Scalar, DataDescriptor::StructureType::Vector},
		{DataDescriptor::StructureType::Scalar, DataDescriptor::StructureType::Matrix},
		{DataDescriptor::StructureType::Vector, DataDescriptor::StructureType::Matrix},
	};

	// All valid field type conversions
	static const std::vector<std::pair<DataDescriptor::FieldType, DataDescriptor::FieldType>> validFieldTypeConversions = {
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Color},
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Point3D},
		{DataDescriptor::FieldType::Color, DataDescriptor::FieldType::Basic},
		{DataDescriptor::FieldType::Color, DataDescriptor::FieldType::Point3D},
		{DataDescriptor::FieldType::Point3D, DataDescriptor::FieldType::Basic},
		{DataDescriptor::FieldType::Point3D, DataDescriptor::FieldType::Color},
	};

	if (match(dataDesc))	// No conversion needed
		return true;

	auto structTypeFrom = _structureType;
	auto structTypeTo = dataDesc._structureType;

	if (structTypeFrom == DataDescriptor::StructureType::Dynamic || structTypeTo == DataDescriptor::StructureType::Any || canConvertTypes(validStructureTypeConversions, structTypeFrom, structTypeTo))
	{
		auto fieldTypeFrom = _fieldType;
		auto fieldTypeTo = dataDesc._fieldType;

		if (fieldTypeFrom == DataDescriptor::FieldType::Dynamic || fieldTypeTo == DataDescriptor::FieldType::Any || canConvertTypes(validFieldTypeConversions, fieldTypeFrom, fieldTypeTo))
		{
			// Value type: Everything can be converted to everything, so no further checks needed
			return true;
		}
	}

	// Conversion isn't possible
	return false;
}

bool DataDescriptor::canConvertToColor(const DataDescriptor& dataDesc) const
{
	// All grayscale to color conversions
	static const std::vector<std::pair<DataDescriptor::FieldType, DataDescriptor::FieldType>> grayscaleToColorConversions = {
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Color},
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Point3D},
	};

	return canConvertTypes(grayscaleToColorConversions, _fieldType, dataDesc._fieldType);
}

bool DataDescriptor::canConvertToGrayscale(const DataDescriptor& dataDesc) const
{
	// All color to grayscale conversions
	static const std::vector<std::pair<DataDescriptor::FieldType, DataDescriptor::FieldType>> colorToGrayscaleConversions = {
		{DataDescriptor::FieldType::Color, DataDescriptor::FieldType::Basic},
		{DataDescriptor::FieldType::Point3D, DataDescriptor::FieldType::Basic},
	};

	return canConvertTypes(colorToGrayscaleConversions, _fieldType, dataDesc._fieldType);
}

unsigned int DataDescriptor::getConversionCosts(const DataDescriptor& dataDesc) const
{
	// Count the steps necessary to convert to the given data descriptor
	unsigned int stepsNeeded = 0;

	if (_structureType != dataDesc._structureType && _structureType != StructureType::Dynamic)
		stepsNeeded += 1;

	if (_fieldType != dataDesc._fieldType && _fieldType != FieldType::Dynamic)
		stepsNeeded += 1;

	if (_valueType != dataDesc._valueType && _valueType != ValueType::Dynamic)
		stepsNeeded += 1;

	return stepsNeeded;
}

bool DataDescriptor::match(const DataDescriptor& dataDesc, DataDescriptor* adjustedDataDesc) const
{	
	DataDescriptor adjusted = *this;
	bool result = true;

	auto checkType = [&dataDesc](auto type, auto typeOther, auto dynamicVal, auto anyVal, auto dynamicAssigner) {
		if (type == dynamicVal)	// Handle dynamic type
		{
			if (typeOther != anyVal)
				dynamicAssigner(typeOther);
			else
				return false;
		}
		else if (type != typeOther && typeOther != anyVal)	// Types do not match
			return false;

		return true;
	};

	// Structure type
	if (!checkType(_structureType, dataDesc._structureType, StructureType::Dynamic, StructureType::Any, [&adjusted](auto newVal) { adjusted._structureType = newVal; }))
		result = false;

	// Field type
	if (!checkType(_fieldType, dataDesc._fieldType, FieldType::Dynamic, FieldType::Any, [&adjusted](auto newVal) { adjusted._fieldType = newVal; }))
		result = false;

	// Value type
	if (!checkType(_valueType, dataDesc._valueType, ValueType::Dynamic, ValueType::Any, [&adjusted](auto newVal) { adjusted._valueType = newVal; }))
		result = false;

	if (adjustedDataDesc)
		*adjustedDataDesc = adjusted;		

	return result;
}

int DataDescriptor::getCVMatrixType(int* chan, int* dep) const
{
	int channels = 0;
	int depth = 0;

	switch (_fieldType)
	{
	case FieldType::Basic:
		channels = 1;
		break;

	case FieldType::Point2D:
		channels = 2;
		break;

	case FieldType::Color:
	case FieldType::Point3D:
		channels = 3;
		break;

	default:
		break;
	}

	if (chan)
		*chan = channels;

	switch (_valueType)
	{
	case ValueType::Int8:
		depth = CV_8S;
		break;

	case ValueType::UInt8:
		depth = CV_8U;
		break;

	case ValueType::Int16:
		depth = CV_16S;
		break;

	case ValueType::UInt16:
		depth = CV_16U;
		break;

	case ValueType::Int32:
	case ValueType::UInt32:
		depth = CV_32S;
		break;

	case ValueType::Float:
		depth = CV_32F;
		break;

	case ValueType::Double:
		depth = CV_64F;
		break;

	default:
		break;
	}

	if (dep)
		*dep = depth;

	return CV_MAKETYPE(depth, channels);
}

void DataDescriptor::fromCVMatrixType(const cv::Mat& matrix, bool applyToDynamicOnly)
{
	if (_structureType == StructureType::Dynamic || !applyToDynamicOnly)
	{
		if (matrix.rows == 1 && matrix.cols == 1)
			_structureType = StructureType::Scalar;
		else if (matrix.rows > 1 && matrix.cols == 1)
			_structureType = StructureType::Vector;
		else
			_structureType = StructureType::Matrix;
	}

	if (_fieldType == FieldType::Dynamic || !applyToDynamicOnly)
	{
		switch (matrix.channels())
		{
		case 1:
			_fieldType = FieldType::Basic;
			break;

		case 2:
			_fieldType = FieldType::Point2D;
			break;

		case 3:
			_fieldType = FieldType::Color;
			break;

		default:
			break;
		}
	}

	if (_valueType == ValueType::Dynamic || !applyToDynamicOnly)
	{
		switch (matrix.type())
		{
		case CV_8S:
			_valueType = ValueType::Int8;
			break;

		case CV_8U:
			_valueType = ValueType::UInt8;
			break;

		case CV_16S:
			_valueType = ValueType::Int16;
			break;

		case CV_16U:
			_valueType = ValueType::UInt16;
			break;

		case CV_32S:
			_valueType = ValueType::Int32;
			break;

		case CV_32F:
			_valueType = ValueType::Float;
			break;

		case CV_64F:
			_valueType = ValueType::Double;
			break;

		default:
			break;
		}
	}
}

int DataDescriptor::getChannels() const
{
	int channels = 0;
	getCVMatrixType(&channels, nullptr);
	return channels;
}

bool DataDescriptor::isGrayscale() const
{
	return _structureType == StructureType::Matrix && getChannels() == 1;
}

bool DataDescriptor::isColor() const
{
	return _structureType == StructureType::Matrix && getChannels() == 3;
}

bool DataDescriptor::isArbitrary() const
{
	return _structureType == StructureType::Any || _fieldType == FieldType::Any || _valueType == ValueType::Any;
}

bool DataDescriptor::isDynamic() const
{
	return _structureType == StructureType::Dynamic || _fieldType == FieldType::Dynamic || _valueType == ValueType::Dynamic;
}
