/******************************************************************************
 * File: ImageBuildPool.cpp
 * Date: 12.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageBuildPool.h"
#include "ImageExceptions.h"
#include "project/Project.h"

const char* ImageBuildPool::Serialization_Group_Item = "Item";

const char* ImageBuildPool::Serialization_Value_Block = "Block";

ImageBuildPool::ImageBuildPool(Label* label)
{
	if (!label)
		throw std::invalid_argument{_EXCPT("label may not be null")};

	// Whenever a block or image reference is removed, remove the associated image builds
	connect(label->pipeline(), &Pipeline::blockRemoved, this, &ImageBuildPool::blockRemoved);
	connect(label->project(), &Project::imageReferenceRemoved, this, &ImageBuildPool::imageReferenceRemoved);
}

const ImageBuildVector* ImageBuildPool::imageBuilds(const Block* block) const
{
	if (_imageBuilds.find(block) != _imageBuilds.cend())
		return _imageBuilds.at(block).get();
	else
		return nullptr;
}

void ImageBuildPool::removeImageBuilds(const Block* block)
{
	if (block)
	{
		auto it = _imageBuilds.find(block);

		if (it != _imageBuilds.end())
		{
			// Send a removal notification for each image build in the build vector
			for (auto& build : *it->second)
				emit imageBuildRemoved(build);

			_imageBuilds.erase(it);
		}
		else
			throw ImageBuildException{nullptr, _EXCPT("Tried to remove the image builds of a block not belonging to this pool")};
	}
}

void ImageBuildPool::removeImageBuilds(const ImageReference* imageRef)
{
	// Remove all builds that contain the given image reference
	for (auto& entry : _imageBuilds)
		removeImageBuild(entry.first, imageRef);
}

std::shared_ptr<ImageBuild> ImageBuildPool::imageBuild(const Block* block, const std::vector<const ImageReference*>& imageReferences)
{
	if (auto buildVector = imageBuildVector(block))
		return imageBuild(buildVector, block, imageReferences);
	else
		throw ImageBuildException{nullptr, _EXCPT("Unable to create an image build")};
}

std::shared_ptr<ImageBuild> ImageBuildPool::imageBuild(const Block* block, const ImageReference* activeImageReference)
{
	// Get the image build for the given block and all its used image references
	auto imageReferences = block->assembleInputImageReferences(activeImageReference);
	return imageBuild(block, std::vector<const ImageReference*>{imageReferences.cbegin(), imageReferences.cend()});
}

void ImageBuildPool::removeImageBuild(const Block* block, const ImageReference* imageRef)
{
	if (block && imageRef)
	{
		auto blockIt = _imageBuilds.find(block);

		if (blockIt != _imageBuilds.end())
		{
			for (auto build : blockIt->second->selectByImageReference(imageRef))
			{
				auto it = blockIt->second->find(build);	// Should always be valid

				emit imageBuildRemoved(*it);
				blockIt->second->erase(it);
			}
		}
		else
			throw ImageBuildException{nullptr, _EXCPT("Tried to remove an image build from a block not belonging to this pool")};
	}
}

void ImageBuildPool::serialize(SerializationContext& ctx) const
{
	// Serialize all image builds
	for (const auto& build : _imageBuilds)
	{
		ctx.beginGroup(Serialization_Group_Item);
		ctx.settings()(Serialization_Value_Block) = ctx.getBlockIndex(build.first);

		ctx.beginGroup(ImageBuildVector::Serialization_Group, true);
		build.second->serialize(ImageBuildVector::Serialization_Element, ctx);
		ctx.endGroup();

		ctx.endGroup();
	}
}

void ImageBuildPool::deserialize(DeserializationContext& ctx)
{
	// Deserialize all image builds
	for (const auto elemSettings : ctx.settings().children(Serialization_Group_Item))
	{
		ctx.beginGroup(elemSettings);
		int blockIndex = ctx.settings()(Serialization_Value_Block, -1).toInt();

		if (auto block = ctx.getBlock(blockIndex))
		{
			auto buildVector = imageBuildVector(block);

			if (ctx.beginGroup(ImageBuildVector::Serialization_Group))
			{
				buildVector->deserialize(ImageBuildVector::Serialization_Element, ctx, [&ctx, &buildVector, block, this](const SettingsContainer& settings) -> std::shared_ptr<ImageBuild> {
					std::vector<const ImageReference*> imageReferences;

					for (auto imageRef : settings(ImageBuild::Serialization_Value_ImageReferences).toString().split(","))
					{
						auto imageRefIndex = imageRef.toInt();

						if (imageRefIndex != -1)
						{
							if (auto imageRef = ctx.getImageReference(imageRefIndex))
								imageReferences.push_back(imageRef);
						}
					}

					if (!imageReferences.empty())
						return imageBuild(buildVector, block, imageReferences);
					else
						return nullptr;
				});

				ctx.endGroup();
			}
		}

		ctx.endGroup();
	}
}

std::shared_ptr<ImageBuildVector> ImageBuildPool::imageBuildVector(const Block* block)
{
	// Create a new build vector if necessary
	if (_imageBuilds.find(block) == _imageBuilds.cend())
	{
		auto buildVector = std::make_shared<ImageBuildVector>();
		_imageBuilds.emplace(block, buildVector);

		return buildVector;
	}
	else
		return _imageBuilds.at(block);
}

std::shared_ptr<ImageBuild> ImageBuildPool::imageBuild(std::shared_ptr<ImageBuildVector>& builds, const Block* block, const std::vector<const ImageReference*>& imageReferences)
{
	// Create a new build if necessary
	if (!builds->selectByImageReferences(imageReferences))
	{
		auto build = std::make_shared<ImageBuild>(block, imageReferences);

		try {	// Propagate initialization errors to the caller
			build->initImageBuild();
		}
		catch (...) {
			throw;
		}

		builds->push_back(build);
		emit imageBuildCreated(build);

		return build;
	}
	else
	{
		auto build = builds->selectByImageReferences(imageReferences);
		return *builds->find(build);
	}
}

void ImageBuildPool::blockRemoved(const std::shared_ptr<Block>& block)
{
	// When a block is deleted, remove all associated image builds
	if (_imageBuilds.find(block.get()) != _imageBuilds.end())	// Only remove image builds of blocks currently in the pool
		removeImageBuilds(block.get());
}

void ImageBuildPool::imageReferenceRemoved(const std::shared_ptr<ImageReference>& imageRef)
{
	// When an image reference is deleted, remove all associated image builds
	removeImageBuilds(imageRef.get());
}
