/******************************************************************************
 * File: ImageTags.cpp
 * Date: 17.4.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageTags.h"
#include "image/ImageExceptions.h"

ImageTags& ImageTags::operator =(const ImageTags& imageTags)
{
	_tags.deepCopy(imageTags._tags);
	return *this;
}

void ImageTags::addTag(const std::shared_ptr<ImageTag>& tag)
{
	// Make sure that the name is unique
	if (_tags.selectByName(tag->getName()))
		throw ImageTagException{nullptr, _EXCPT(QString{"A tag with the name '%1' already exists"}.arg(tag->getName()))};

	_tags.push_back(tag);
	emit tagAdded(tag);
}

void ImageTags::updateTag(const ImageTag* tag)
{
	// Notify about the tag having been modified
	auto index = _tags.indexOf(tag);

	if (index != -1)
		emit tagModified(_tags.at(index));
}

void ImageTags::removeTag(const ImageTag* tag)
{
	if (tag)
	{
		auto it = _tags.find(tag);

		if (it != _tags.cend())
		{
			// Keep a copy of the shared_ptr holding the image tag to increase its use count;
			// otherwise, the image tag will be deleted before it has been removed from the vector, potentially causing a crash
			auto tag = *it;

			emit tagRemoved(*it);
			_tags.erase(it);
		}
		else
			throw ImageTagException{nullptr, _EXCPT("Tried to remove a tag not belonging to this list")};
	}
}

void ImageTags::removeAllTags()
{
	auto tags = _tags;

	for (const auto& tag : tags)
		removeTag(tag.get());
}

void ImageTags::serialize(SerializationContext& ctx) const
{
	// Serialize all image tags
	ctx.beginGroup(ImageTagVector::Serialization_Group, true);
	_tags.serialize(ImageTagVector::Serialization_Element, ctx);
	ctx.endGroup();
}

void ImageTags::deserialize(DeserializationContext& ctx)
{
	// Deserialize all image tags
	if (ctx.beginGroup(ImageTagVector::Serialization_Group))
	{
		_tags.deserialize(ImageTagVector::Serialization_Element, ctx, [this](const SettingsContainer& settings) {
			QString name = settings(ImageTag::Serialization_Value_Name).toString();

			// Create new tag
			auto tag = makeTag(name);
			addTag(tag);
			return tag;
		});

		ctx.endGroup();
	}
}
