/******************************************************************************
 * File: ImageTagsAllotment.cpp
 * Date: 03.5.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageTagsAllotment.h"
#include "ImageTags.h"
#include "image/ImageBuild.h"

const char* ImageTagsAllotment::Serialization_Value_AllottedTags = "AllottedTags";

ImageTagsAllotment& ImageTagsAllotment::operator =(const ImageTagsAllotment& imageTagsAllotment)
{
	unallotAllTags();

	// Only copy the alloted tags if both allotments use the same input image tags
	if (_inputImageTags == imageTagsAllotment._inputImageTags)
	{
		_allottedTags = imageTagsAllotment._allottedTags;
		emit allotmentChanged();
	}

	return *this;
}

void ImageTagsAllotment::setImageBuild(const ImageBuild* imageBuild)
{
	// Disconnect any previously connected signals from the image build
	if (_imageBuild)
		disconnect(_imageBuild, nullptr, this, nullptr);

	_imageBuild = imageBuild;

	setInputImageTags(_imageBuild->inputImageTags());

	// Listen for connection signals from the block's image tags in-port
	connect(_imageBuild, &ImageBuild::inputImageTagsChanged, this, &ImageTagsAllotment::inputImageTagsChanged);
}

void ImageTagsAllotment::allotTag(ImageTag* imageTag)
{
	if (_inputImageTags && _inputImageTags->tags().contains(imageTag))
	{
		_allottedTags.emplace(imageTag);
		emit allotmentChanged();
	}
}

void ImageTagsAllotment::unallotTag(ImageTag* imageTag)
{
	_allottedTags.erase(imageTag);
	emit allotmentChanged();
}

void ImageTagsAllotment::serialize(SerializationContext& ctx) const
{
	// Store the names of all alloted image tags; since tag names must be unique, this can be used to identify each tag
	QStringList allottedTags;

	for (const auto& imageTag : _allottedTags)
		allottedTags << imageTag->getName();

	ctx.settings()(Serialization_Value_AllottedTags) = allottedTags.join(",");
}

void ImageTagsAllotment::deserialize(DeserializationContext& ctx)
{
	unallotAllTags();

	if (_inputImageTags)
	{
		// Search for each tag name in the current input image tags
		for (auto tagName : ctx.settings()(Serialization_Value_AllottedTags).toString().split(","))
		{
			if (auto imageTag = _inputImageTags->tags().selectByName(tagName))
				allotTag(imageTag.get());
		}
	}
}

void ImageTagsAllotment::setInputImageTags(const ImageTags* imageTags)
{
	if (imageTags == _inputImageTags)
		return;

	if (_inputImageTags)
		disconnect(_inputImageTags, nullptr, this, nullptr);

	// The input image tags have changed, so clear any alloted tags
	unallotAllTags();

	_inputImageTags = imageTags;

	// If a tag has been removed, remove it from the allotment as well
	if (_inputImageTags)
	{
		connect(_inputImageTags, &ImageTags::tagModified, this, &ImageTagsAllotment::imageTagModified);
		connect(_inputImageTags, &ImageTags::tagRemoved, this, &ImageTagsAllotment::imageTagRemoved);
	}
}

void ImageTagsAllotment::imageTagRemoved(const std::shared_ptr<ImageTag>& imageTag)
{
	unallotTag(imageTag.get());
}
