/******************************************************************************
 * File: Layer.cpp
 * Date: 17.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "Layer.h"
#include "ImageBuild.h"
#include "ImageExceptions.h"
#include "DraftItemCatalog.h"

const char* Layer::Serialization_Value_Type = "Type";
const char* Layer::Serialization_Value_Visible = "Visible";
const char* Layer::Serialization_Value_Alpha = "Alpha";
const char* Layer::Serialization_Value_BackgroundImage = "BackgroundImage";

Layer::Layer(ImageBuild* imageBuild, QString name, Type type, const ImageReference* backgroundImage) : ImageBuildItem(imageBuild, name),
	_type{type}, _draftItems{this}, _layerPixels{this}, _backgroundImage{backgroundImage}
{
	if (type == Type::Standard)	// Standard layers don't have a background
		_backgroundImage = nullptr;
}

void Layer::initLayer()
{
	ImageBuildItem::initImageBuildItem();
}

std::shared_ptr<DraftItem> Layer::createDraftItem(DraftItemType type)
{
	if (type == DraftItemType::Undefined)
		throw std::invalid_argument(_EXCPT("type may not be DraftItemType::Undefined"));

	// Create new block using the block factory; cast the unique ptr to a shared one as well
	std::shared_ptr<DraftItem> item = DraftItemCatalog::createDraftItem(this, type);

	try {	// Propagate initialization errors to the caller
		item->initDraftItem();
	}
	catch (...) {
		throw;
	}

	_draftItems.push_back(item);
	emit draftItemCreated(item);

	return item;
}

void Layer::removeDraftItem(const DraftItem* item)
{
	if (item)
	{
		auto it = _draftItems.find(item);

		if (it != _draftItems.cend())
		{
			// Keep a copy of the shared_ptr holding the draft item to increase its use count;
			// otherwise, the draft item will be deleted before it has been removed from the vector, potentially causing a crash
			auto item = *it;

			emit draftItemRemoved(*it);
			_draftItems.erase(it);
		}
		else
			throw LayerException{this, _EXCPT("Tried to remove a draft item not belonging to this layer")};
	}
}

int Layer::getZOrder() const
{
	return _imageBuild->layers().indexOf(this);
}

void Layer::setVisible(bool visible)
{
	if (visible != _isVisible)
	{
		_isVisible = visible;

		if (_isVisible)
			emit layerShown();
		else
			emit layerHidden();
	}
}

void Layer::setAlpha(unsigned int alpha)
{
	if (alpha > 100)
		alpha = 100;

	if (alpha != _alpha)
	{
		_alpha = alpha;
		emit layerAlphaChanged(_alpha);
	}
}

void Layer::serialize(SerializationContext& ctx) const
{
	ImageBuildItem::serialize(ctx);

	// Serialize values
	ctx.settings()(Serialization_Value_Type) = static_cast<int>(_type);
	ctx.settings()(Serialization_Value_Visible) = _isVisible;
	ctx.settings()(Serialization_Value_Alpha) = _alpha;
	ctx.settings()(Serialization_Value_BackgroundImage) = _backgroundImage ? ctx.getImageReferenceIndex(_backgroundImage) : -1;

	// Serialize all draft items
	ctx.beginGroup(DraftItemVector::Serialization_Group, true);
	_draftItems.serialize(DraftItemVector::Serialization_Element, ctx);
	ctx.endGroup();

	// Serialize pixels
	ctx.beginGroup(LayerPixels::Serialization_Group);
	_layerPixels.serialize(ctx);
	ctx.endGroup();
}

void Layer::deserialize(DeserializationContext& ctx)
{
	ImageBuildItem::deserialize(ctx);

	// Deserialize values
	_type = static_cast<Type>(ctx.settings()(Serialization_Value_Type, static_cast<int>(Type::Standard)).toInt());
	_alpha = ctx.settings()(Serialization_Value_Alpha, 100).toUInt();
	_isVisible = ctx.settings()(Serialization_Value_Visible, true).toBool();

	int backgroundImageIndex = ctx.settings()(Serialization_Value_BackgroundImage, -1).toInt();

	if (backgroundImageIndex != -1)
		_backgroundImage = ctx.getImageReference(backgroundImageIndex);
	else
		_backgroundImage = nullptr;

	// Deserialize all draft items
	if (ctx.beginGroup(DraftItemVector::Serialization_Group))
	{
		_draftItems.deserialize(DraftItemVector::Serialization_Element, ctx, [this](const SettingsContainer& settings) {
			DraftItemType type = settings(DraftItem::Serialization_Value_Type, DraftItemType::Undefined).toString();

			return createDraftItem(type);
		});

		ctx.endGroup();
	}

	// Deserialize pixels
	if (ctx.beginGroup(LayerPixels::Serialization_Group))
	{
		_layerPixels.deserialize(ctx);
		ctx.endGroup();
	}
}
