/******************************************************************************
 * File: LayerPixels.cpp
 * Date: 18.5.2018
 *****************************************************************************/

#include "Grinder.h"
#include "LayerPixels.h"
#include "ImageBuild.h"
#include "renderers/LayerPixelsRenderer.h"

const char* LayerPixels::Serialization_Group = "Pixels";

LayerPixels::LayerPixels(Layer* layer) : LayerItem(layer)
{

}

void LayerPixels::renderDraftItems(const std::vector<DraftItem*>& draftItems, const ItemRenderer::RendererStyle& rendererStyle)
{
	// Create an image as a canvas we'll render the items on
	QSize imageSize{_layer->imageBuild()->imageData().cols, _layer->imageBuild()->imageData().rows};

	if (imageSize.isEmpty())	// No data, nothing to do
		return;

	QImage canvas{imageSize, QImage::Format_ARGB32};
	canvas.fill(0);	// Erase everything with an invisible color

	// Render using a painter
	QPainter painter{&canvas};
	painter.setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform, false);

	for (const auto& draftItem : draftItems)
	{
		// Use the draft item's renderer for rendering
		if (auto renderer = draftItem->createRenderer(rendererStyle))
			renderer->render(&painter, ItemRenderer::RenderModeFlag::RenderToImage, ItemRenderer::RenderFlag::NoFlag);
	}

	painter.end();

	// Items have been rendered, so copy the canvas pixels
	_pixelsData.beginPainting();

	for (int r = 0; r < canvas.size().height(); ++r)
	{
		auto rowData = reinterpret_cast<const QRgb*>(canvas.constScanLine(r));
		std::pair<int, int> stride{0, 0};
		QRgb strideColor = rowData[0];

		for (int c = 0; c < canvas.size().width(); ++c)
		{
			QRgb rgba = rowData[c];

			if (rgba != strideColor)	// Apply current stride and start a new one
			{
				if (strideColor != 0 && qAlpha(strideColor) > 0)
					_pixelsData.set(stride, r, strideColor);

				stride.first = stride.second = c;
				strideColor = rgba;
			}
			else	// Extend current stride
				stride.second = c;
		}

		// Apply the last pending stride as well
		if (strideColor != 0 && qAlpha(strideColor) > 0)
			_pixelsData.set(stride, r, strideColor);
	}

	_pixelsData.endPainting();
}

std::unique_ptr<ItemRenderer> LayerPixels::createRenderer(const ItemRenderer::RendererStyle& rendererStyle) const
{
	return std::make_unique<LayerPixelsRenderer>(&_pixelsData, rendererStyle);
}

void LayerPixels::serialize(SerializationContext& ctx) const
{
	// Serialize pixels data
	ctx.beginGroup(LayerPixelsData::Serialization_Group);
	_pixelsData.serialize(ctx);
	ctx.endGroup();
}

void LayerPixels::deserialize(DeserializationContext& ctx)
{
	// Deserialize pixels data
	if (ctx.beginGroup(LayerPixelsData::Serialization_Group))
	{
		_pixelsData.deserialize(ctx);
		ctx.endGroup();
	}
}
