/******************************************************************************
 * File: LayerPixelsData.cpp
 * Date: 17.5.2018
 *****************************************************************************/

#include "Grinder.h"
#include "LayerPixelsData.h"

const char* LayerPixelsData::Serialization_Group = "PixelsData";

QColor LayerPixelsData::get(int x, int y) const
{
	// Search all colors for the given point (x,y)
	for (const auto& colorEntry : _colorMap)
	{
		if (colorEntry.second.contains(y, x))	// Colors are mapped in a row->col fashion, so x and y are swapped here
			return colorEntry.first;
	}

	return QColor{};
}

void LayerPixelsData::set(const range_type& range, int y, QColor color)
{
	// Unset (range,y) from all colors
	clear(range, y);

	// Set (range,y) to the given color
	_colorMap[color].add(y, range);	// Colors are mapped in a row->col fashion, so range and y are swapped here

	emit dataModified();
}

void LayerPixelsData::clear(const range_type& range, int y)
{
	std::vector<QColor> emptyColors;

	// Clear the given points (range,y) from all colors
	for (auto it = _colorMap.begin(); it != _colorMap.end(); ++it)
	{
		auto& colorEntry = *it;
		colorEntry.second.remove(y, range);	// Colors are mapped in a row->col fashion, so range and y are swapped here

		// Remove empty colors
		if (colorEntry.second.empty())
			_colorMap.erase(it);
	}

	emit dataModified();
}

std::vector<QColor> LayerPixelsData::getColors() const
{
	std::vector<QColor> colors;

	for (const auto& colorEntry : _colorMap)
		colors.push_back(colorEntry.first);

	return colors;
}

bool LayerPixelsData::isSet(int x, int y) const
{
	auto color = get(x, y);	// If no pixel is set at (x,y), get returns an invalid color
	return color.isValid();
}

QRect LayerPixelsData::bounds() const
{
	QRect dataBounds;

	dataBounds.setLeft(std::numeric_limits<int>::max());
	dataBounds.setTop(std::numeric_limits<int>::max());
	dataBounds.setRight(std::numeric_limits<int>::min());
	dataBounds.setBottom(std::numeric_limits<int>::min());

	for (const auto& colorEntry : _colorMap)
	{
		auto rangeBounds = colorEntry.second.bounds();

		// Get bounds in x direction
		if (rangeBounds.first.first < dataBounds.left())
			dataBounds.setLeft(rangeBounds.first.first);

		if (rangeBounds.first.second > dataBounds.right())
			dataBounds.setRight(rangeBounds.first.second);

		// Get bounds in y direction
		if (rangeBounds.second.first < dataBounds.top())
			dataBounds.setTop(rangeBounds.second.first);

		if (rangeBounds.second.second > dataBounds.bottom())
			dataBounds.setBottom(rangeBounds.second.second);
	}

	return dataBounds;
}

void LayerPixelsData::serialize(SerializationContext& ctx) const
{
	// Serialize the entire color map
	for (const auto& colorEntry : _colorMap)
		ctx.settings()(colorEntry.first.name()) = colorEntry.second.toString();
}

void LayerPixelsData::deserialize(DeserializationContext& ctx)
{
	clear();

	// Deserialize all color map entries
	for (auto it = ctx.settings().values().begin(); it != ctx.settings().values().end(); ++it)
		_colorMap[QColor{it.key()}].fromString(it->toString());

	emit dataModified();
}
