/******************************************************************************
 * File: BlockCatalog.cpp
 * Date: 29.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "BlockCatalog.h"

#include "blocks/InputBlock.h"
#include "blocks/OutputBlock.h"
#include "blocks/ConvertToGrayscaleBlock.h"
#include "blocks/ConvertToColorBlock.h"
#include "blocks/NormalizeBlock.h"
#include "blocks/SplitChannelsBlock.h"
#include "blocks/MergeChannelsBlock.h"
#include "blocks/ReplaceColorBlock.h"
#include "blocks/AdaptiveThresholdBlock.h"
#include "blocks/BinaryThresholdBlock.h"
#include "blocks/AlphaBlendingBlock.h"
#include "blocks/ImageTagsBlock.h"

#define REGISTER_BLOCK_TYPE(cls, desc)	registerBlockType(cls::type_value, cls::category_value, [](Pipeline* pipeline, QString name) { return std::make_unique<cls>(pipeline, name); }, desc)

std::map<BlockType, BlockCatalog::BlockDescriptor> BlockCatalog::s_descriptors;

std::set<BlockType> BlockCatalog::getTypes(BlockCategory category)
{
	std::set<BlockType> types;

	for (const auto& desc : s_descriptors)
	{
		if (category == BlockCategory::Undefined || category == desc.second.category)
			types.insert(desc.second.type);
	}

	return types;
}

std::set<BlockCategory> BlockCatalog::getCategories()
{
	std::set<BlockCategory> cats;

	for (const auto& desc : s_descriptors)
	{
		if (desc.second.category != BlockCategory::Undefined)
			cats.insert(desc.second.category);
	}

	return cats;
}

BlockCategory BlockCatalog::getCategory(BlockType type)
{
	auto it = s_descriptors.find(type);

	if (it != s_descriptors.cend())
		return it->second.category;
	else
		return BlockCategory::Undefined;
}

QString BlockCatalog::getDescription(BlockType type)
{
	auto it = s_descriptors.find(type);

	if (it != s_descriptors.cend())
		return it->second.description;
	else
		return "";
}

void BlockCatalog::registerBlockType(BlockType type, BlockCategory category, BlockCatalog::block_creator_type creator, QString description)
{
	if (type == BlockType::Undefined)
		throw std::invalid_argument{_EXCPT("type may not be BlockType::Undefined")};

	if (category == BlockCategory::Undefined)
		throw std::invalid_argument{_EXCPT("category may not be BlockCategory::Undefined")};

	if (!creator)
		throw std::invalid_argument{_EXCPT("creator may not be null")};

	s_descriptors[type] = {type, category, creator, description};
}

std::unique_ptr<Block> BlockCatalog::createBlock(Pipeline* pipeline, BlockType type, QString name)
{
	if (type == BlockType::Undefined)
		throw std::invalid_argument{_EXCPT("type may not be BlockType::Undefined")};

	if (s_descriptors.find(type) != s_descriptors.end())
	{
		auto descriptor = s_descriptors.at(type);
		auto block = descriptor.creator(pipeline, name);

		if (!block)
			throw std::runtime_error{_EXCPT(QString{"Failed to create a block of type %1"}.arg(type))};

		return block;
	}
	else
		throw std::invalid_argument{_EXCPT("The given type has not been registered")};
}

void BlockCatalog::registerStandardBlocks()
{
	REGISTER_BLOCK_TYPE(InputBlock, "The beginning of every pipeline; provides an image of the image stack.");
	REGISTER_BLOCK_TYPE(OutputBlock, "The final stage of every pipeline; makes the image available for editing and saving.");
	REGISTER_BLOCK_TYPE(ConvertToGrayscaleBlock, "Converts its input to a grayscale image.");
	REGISTER_BLOCK_TYPE(ConvertToColorBlock, "Converts its input to a color image of various color spaces.");
	REGISTER_BLOCK_TYPE(NormalizeBlock, "Normalizes its input to a given range.");
	REGISTER_BLOCK_TYPE(SplitChannelsBlock, "Splits a color image into three separate channels.");
	REGISTER_BLOCK_TYPE(MergeChannelsBlock, "Merges three separate channels into one color image.");
	REGISTER_BLOCK_TYPE(ReplaceColorBlock, "Replaces a color by another one in an image.");
	REGISTER_BLOCK_TYPE(AdaptiveThresholdBlock, "Applies an adaptive threshold to its input.");
	REGISTER_BLOCK_TYPE(BinaryThresholdBlock, "Applies a binary threshold to its input.");
	REGISTER_BLOCK_TYPE(AlphaBlendingBlock, "Combines two images by alpha blending them together.");
	REGISTER_BLOCK_TYPE(ImageTagsBlock, "Provides a list of user-definable tags.");
}
