/******************************************************************************
 * File: BlockHierarchy.cpp
 * Date: 20.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "BlockHierarchy.h"
#include "PipelineExceptions.h"
#include "Pipeline.h"
#include "Block.h"

BlockHierarchy::BlockHierarchy(const Pipeline* pipeline)
{
	if (pipeline)
		createFromPipeline(pipeline);
}

void BlockHierarchy::createFromPipeline(const Pipeline* pipeline)
{
	if (!pipeline)
		throw std::invalid_argument{_EXCPT("pipeline may not be null")};

	_hierarchy.clear();

	// Level-0 blocks mark the beginning(s) of our pipeline
	auto level0 = findLevel0Blocks(pipeline);

	if (level0.empty())	// No level-0 blocks means that the pipeline has no beginnings (the pipeline starts with cycles)
		throw PipelineException{pipeline, _EXCPT("The pipeline has no starting blocks")};

	_hierarchy.push_back(std::move(level0));

	// Recursively create all following levels
	createNextLevel(pipeline);

	// All blocks must be part of the hierarchy; if this isn't the case, cyclic dependencies most likely exist
	unsigned int blocksInHierarchy = 0;

	for (const auto& level : _hierarchy)
		blocksInHierarchy += level.size();

	if (blocksInHierarchy != pipeline->blocks().size())
		throw PipelineException{pipeline, _EXCPT("Not all blocks could be added to the hierarchy; this is usually caused by cyclic dependencies in the graph")};
}

int BlockHierarchy::getBlockLevel(const Block* block) const
{
	for (unsigned int i = 0; i < _hierarchy.size(); ++i)
	{
		const auto& level = _hierarchy[i];

		for (const auto& blockInHierarchy : level)
		{
			if (blockInHierarchy == block)
				return i;
		}
	}

	return -1;
}

void BlockHierarchy::createNextLevel(const Pipeline* pipeline)
{
	const auto nextLevel = findNextLevelBlocks(pipeline, _hierarchy.back());

	// If a new level was created, add it and create the next level once again
	if (!nextLevel.empty())
	{
		_hierarchy.push_back(std::move(nextLevel));
		createNextLevel(pipeline);
	}
}

bool BlockHierarchy::isBlockInHierarchy(const Block* block) const
{
	return getBlockLevel(block) != -1;
}

BlockHierarchy::HierarchyLevel BlockHierarchy::findLevel0Blocks(const Pipeline* pipeline) const
{
	HierarchyLevel level0;

	// Find all blocks that have no inputs (level-0 blocks)
	for (const auto& block : pipeline->blocks())
	{
		bool isLevel0Block = true;

		for (const auto& port : block->ports().selectByDirection(Port::Direction::In))
		{
			if (port->isConnected())
			{
				isLevel0Block = false;
				break;
			}
		}

		if (isLevel0Block)
			level0.push_back(block.get());
	}

	return level0;
}

BlockHierarchy::HierarchyLevel BlockHierarchy::findNextLevelBlocks(const Pipeline* pipeline, const HierarchyLevel& level) const
{
	HierarchyLevel nextLevel;

	for (const auto& block : level)
	{
		auto nextLevelForBlock = findNextLevelBlocks(block);

		for (const auto& newBlock : nextLevelForBlock)
		{
			// If the found block is already in a lower level of the hierarchy, a cycle exists
			if (isBlockInHierarchy(newBlock))
				throw PipelineException{pipeline, _EXCPT("A cycle exists in the pipeline")};

			// Otherwise, add the found block to the level
			if (std::find(nextLevel.begin(), nextLevel.end(), newBlock) == nextLevel.end())
				nextLevel.push_back(newBlock);
		}
	}

	return nextLevel;
}

BlockHierarchy::HierarchyLevel BlockHierarchy::findNextLevelBlocks(const Block* block) const
{
	HierarchyLevel nextLevel;

	// Find all blocks this block is connected to
	for (const auto& port : block->ports().selectByDirection(Port::Direction::Out))
	{
		for (const auto& con : port->connections())
		{
			// Check if all inputs of the connected block are in the hierarchy; otherwise, skip it
			if (checkBlockInputs(con->dest(), block))
				nextLevel.push_back(con->dest());
		}
	}

	return nextLevel;
}

bool BlockHierarchy::checkBlockInputs(const Block* block, const Block* currentBlock) const
{
	for (const auto& port : block->ports().selectByDirection(Port::Direction::In))
	{
		for (const auto& con : port->getConnections(Port::Direction::In))
		{
			// Check if the source is already in the hierarchy; if not, this block must be postponed to a higher level
			if (con->source() != currentBlock && !isBlockInHierarchy(con->source()))
				return false;
		}
	}

	return true;
}
