/******************************************************************************
 * File: Pipeline.cpp
 * Date: 15.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "Pipeline.h"
#include "PipelineManager.h"
#include "Block.h"
#include "BlockCatalog.h"
#include "PipelineExceptions.h"
#include "core/GrinderApplication.h"
#include "util/SerializationUtils.h"

const char* Pipeline::Serialization_Value_Name = "Name";

Pipeline::Pipeline(QString name) :
	_name{name}, _blocks{this}
{

}

void Pipeline::initPipeline()
{

}

std::shared_ptr<Block> Pipeline::createBlock(BlockType type, QString name)
{
	if (type == BlockType::Undefined)
		throw std::invalid_argument(_EXCPT("type may not be BlockType::Undefined"));

	// Create new block using the block factory; cast the unique ptr to a shared one as well
	std::shared_ptr<Block> block = BlockCatalog::createBlock(this, type, name);

	try {	// Propagate initialization errors to the caller
		block->initBlock();
	}
	catch (...) {
		throw;
	}

	_blocks.push_back(block);
	emit blockCreated(block);

	return block;
}

void Pipeline::removeBlock(const Block* block)
{
	if (block)
	{
		auto it = _blocks.find(block);

		if (it != _blocks.cend())
		{
			// Keep a copy of the shared_ptr holding the block to increase its use count;
			// otherwise, the block will be deleted before it has been removed from the vector, potentially causing a crash
			auto block = *it;

			// Remove all connections to or from this block
			for (auto port : block->ports())
				port->disconnectFromAll();

			emit blockRemoved(*it);
			_blocks.erase(it);
		}
		else
			throw PipelineException{this, _EXCPT("Tried to remove a block not belonging to this pipeline")};
	}
}

void Pipeline::serialize(SerializationContext& ctx) const
{
	// Before serializing the pipeline, clear any previously saved connections
	ctx.clearConnections();

	// Serialize values
	ctx.settings()(Serialization_Value_Name) = _name;

	// Serialize all blocks
	ctx.beginGroup(BlockVector::Serialization_Group, true);
	_blocks.serialize(BlockVector::Serialization_Element, ctx);
	ctx.endGroup();

	// Serialize all connections
	ctx.beginGroup(ConnectionVector::Serialization_Group, true);
	SerializationUtils::serializeContainer(ctx.connections(), ConnectionVector::Serialization_Element, ctx);
	ctx.endGroup();
}

void Pipeline::deserialize(DeserializationContext& ctx)
{
	// Deserialize values
	_name = ctx.settings()(Serialization_Value_Name).toString();

	// Deserialize all blocks	
	if (ctx.beginGroup(BlockVector::Serialization_Group))
	{
		_blocks.deserialize(BlockVector::Serialization_Element, ctx, [this](const SettingsContainer& settings) {
			BlockType type = settings(Block::Serialization_Value_Type, BlockType::Undefined).toString();
			QString name = settings(Block::Serialization_Value_Name).toString();

			return createBlock(type, name);
		});

		ctx.endGroup();
	}

	// Deserialize all connections
	if (ctx.beginGroup(ConnectionVector::Serialization_Group))
	{
		SerializationUtils::deserializeContainer<ConnectionVector>(ConnectionVector::Serialization_Element, ctx, [&ctx](const SettingsContainer& settings) -> std::shared_ptr<Connection> {
			int sourceIndex = settings(Connection::Serialization_Value_Source, -1).toInt();
			int destIndex = settings(Connection::Serialization_Value_Dest, -1).toInt();
			auto sourcePort = ctx.getPort(sourceIndex);
			auto destPort = ctx.getPort(destIndex);

			if (sourcePort && destPort)
				return sourcePort->connectTo(destPort);
			else
				return nullptr;
		});

		ctx.endGroup();
	}
}
