/******************************************************************************
 * File: ImageReference.cpp
 * Date: 10.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageReference.h"
#include "project/ProjectExceptions.h"
#include "util/FileUtils.h"

#include <QCollator>
#include <opencv2/highgui.hpp>

const char* ImageReference::Serialization_Value_File = "File";
const char* ImageReference::Serialization_Value_Index = "Index";

ImageReference::ImageReference(Project* project, QString imageFilePath, const QStringList& imageSearchPaths) : ProjectItem(project),
	_imageFilePath{imageFilePath}
{
	// If image search paths are provided and the original file doesn't exist, search for it on those paths
	if (!QFileInfo::exists(_imageFilePath) && !imageSearchPaths.isEmpty())
		searchImage(imageSearchPaths);
}

cv::Mat ImageReference::loadImage() const
{
	if (_isValid)
		return cv::imread(_imageFilePath.toStdString(), cv::IMREAD_COLOR);
	else
		throw ImageReferenceException{this, _EXCPT("The image is invalid")};
}

void ImageReference::refreshImageInfo()
{
	_isValid = false;

	// Get general file info
	QFileInfo fileInfo{_imageFilePath};

	if (fileInfo.exists())
	{
		_imageInfo.fileSize = fileInfo.size();
		_imageInfo.fileDate = fileInfo.fileTime(QFile::FileModificationTime);

		// Try loading the image
		auto image = cv::imread(_imageFilePath.toStdString());

		if (!image.empty())
			_imageInfo.imageSize = QSize{image.cols, image.rows};
		else
			throw ImageReferenceException{this, _EXCPT("Invalid image format")};
	}
	else
		throw ImageReferenceException{this, _EXCPT("Invalid image file path")};

	_isValid = true;
}

bool ImageReference::operator <(QString image) const
{
	QCollator collator;
	collator.setNumericMode(true);

	return collator.compare(_imageFilePath, image) < 0;
}

bool ImageReference::operator ==(QString image) const
{
	return FileUtils::compareFileNames(_imageFilePath, image);
}

QString ImageReference::getImageFileName() const
{
	QFileInfo fileInfo{_imageFilePath};
	return fileInfo.fileName();
}

void ImageReference::serialize(SerializationContext& ctx) const
{
	// Serialize values
	ctx.settings()(Serialization_Value_File) = _imageFilePath;
	ctx.settings()(Serialization_Value_Index) = ctx.addImageReference(this);
}

void ImageReference::deserialize(DeserializationContext& ctx)
{
	ctx.addImageReference(ctx.settings()(Serialization_Value_Index).toInt(), this);
}

void ImageReference::searchImage(const QStringList& imageSearchPaths)
{
	QString fileName = QFileInfo{_imageFilePath}.fileName();

	for (auto searchPath : imageSearchPaths)
	{
		QString file = QFileInfo{QDir{searchPath}, fileName}.absoluteFilePath();

		// If the file exists in the current searchPath, use that one as the image file
		if (QFileInfo::exists(file))
		{
			_imageFilePath = file;
			break;
		}
	}
}
