/******************************************************************************
 * File: Project.cpp
 * Date: 07.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "Project.h"
#include "ProjectExceptions.h"
#include "core/GrinderApplication.h"
#include "pipeline/PipelineManager.h"

const char* Project::Serialization_Group = "Project";
const char* Project::Serialization_Value_Name = "Name";

std::shared_ptr<Label> Project::createLabel(QString name)
{
	if (name.isEmpty())
		throw std::invalid_argument{_EXCPT("name may not be empty")};

	if (_labels.selectByName(name))
		throw LabelException{nullptr, _EXCPT(QString{"A label with name '%1' already exists"}.arg(name))};

	// Create a pipeline for the label; if this fails, re-throw
	try {
		auto pipeline = grinder()->pipelineManager().createPipeline(name);

		// Pipeline creation succeeded, so create a label for it
		auto label = std::make_shared<Label>(this, pipeline);

		_labels.push_back(label);
		emit labelCreated(label);

		return label;
	} catch (std::exception& e) {
		throw LabelException{nullptr, _EXCPT(QString{"Failed to create label '%1': %2"}.arg(name).arg(e.what()))};
	}
}

void Project::removeLabel(QString name)
{
	if (!name.isEmpty())
	{
		auto label = _labels.selectByName(name);

		if (label)
			removeLabel(label.get());
		else
			throw LabelException{nullptr, _EXCPT(QString{"Tried to remove a non-existing label (Name: %1)"}.arg(name))};
	}
}

void Project::removeLabel(const Label* label)
{
	if (label)
	{
		auto it = _labels.find(label);

		if (it != _labels.cend())
		{
			// Remove the pipeline associated with the label from the pipeline manager
			grinder()->pipelineManager().removePipeline(it->get()->pipeline());

			// Keep a copy of the shared_ptr holding the label to increase its use count;
			// otherwise, the label will be deleted before it has been removed from the vector, potentially causing a crash
			auto label = *it;

			emit labelRemoved(*it);
			_labels.erase(it);
		}
		else
			throw LabelException{label, _EXCPT("Tried to remove a label not currently part of the project")};
	}
}

std::shared_ptr<ImageReference> Project::createImageReference(QString imagePath, const QStringList& imageSearchPaths)
{
	if (imagePath.isEmpty())
		throw std::invalid_argument{_EXCPT("imagePath may not be empty")};

	if (_imageReferences.selectByFilePath(imagePath))
		throw ImageReferenceException{nullptr, _EXCPT(QString{"The image '%1' has already been added to the project"}.arg(imagePath))};

	auto imageRef = std::make_shared<ImageReference>(this, imagePath, imageSearchPaths);

	try {	// Propagate image reference errors to the caller
		imageRef->refreshImageInfo();
	} catch (...) {
		throw;
	}

	_imageReferences.push_back(imageRef);
	emit imageReferenceCreated(imageRef);

	return imageRef;
}

void Project::removeImageReference(QString imagePath)
{
	if (!imagePath.isEmpty())
	{
		auto ref = _imageReferences.selectByFilePath(imagePath);

		if (ref)
			removeImageReference(ref.get());
		else
			throw ImageReferenceException{nullptr, _EXCPT(QString{"Tried to remove a non-existing image reference (Path: %1)"}.arg(imagePath))};
	}
}

void Project::removeImageReference(const ImageReference* imageRef)
{
	if (imageRef)
	{
		auto it = _imageReferences.find(imageRef);

		if (it != _imageReferences.cend())
		{
			// Keep a copy of the shared_ptr holding the image ref to increase its use count;
			// otherwise, the image ref will be deleted before it has been removed from the vector, potentially causing a crash
			auto imageRef = *it;

			emit imageReferenceRemoved(*it);
			_imageReferences.erase(it);
		}
		else
			throw ImageReferenceException{imageRef, _EXCPT("Tried to remove an image reference not currently part of the project")};
	}
}

void Project::clear()
{
	// Remove all items using the corresponding project functions so that they are removed in a proper manner
	while (!_labels.empty())
		removeLabel(_labels.back().get());

	while (!_imageReferences.empty())
		removeImageReference(_imageReferences.back().get());
}

void Project::serialize(SerializationContext& ctx) const
{
	// Serialize values
	ctx.settings()(Serialization_Value_Name) = _name;

	// Serialize all image references (has to be done before the label serialization, since labels reference them)
	ctx.beginGroup(ImageReferenceVector::Serialization_Group, true);
	_imageReferences.serialize(ImageReferenceVector::Serialization_Element, ctx);
	ctx.endGroup();

	// Serialize all labels
	ctx.beginGroup(LabelVector::Serialization_Group, true);
	_labels.serialize(LabelVector::Serialization_Element, ctx);
	ctx.endGroup();

	// Serialize image editors
	ctx.beginGroup(ImageEditorManager::Serialization_Group, true);
	grinder()->imageEditorManager().serialize(ctx);
	ctx.endGroup();
}

void Project::deserialize(DeserializationContext& ctx)
{
	// Deserialize values
	_name = ctx.settings()(Serialization_Value_Name).toString();

	// Deserialize all image references (has to be done before the label deserialization, since labels reference them)
	if (ctx.beginGroup(ImageReferenceVector::Serialization_Group))
	{
		_imageReferences.deserialize(ImageReferenceVector::Serialization_Element, ctx, [&ctx, this](const SettingsContainer& settings) {
			QString file = settings(ImageReference::Serialization_Value_File).toString();
			return createImageReference(file, ctx.imageSearchPaths());
		});

		ctx.endGroup();
	}

	// Deserialize all labels
	if (ctx.beginGroup(LabelVector::Serialization_Group))
	{
		_labels.deserialize(LabelVector::Serialization_Element, ctx, [this](const SettingsContainer& settings) {
			QString name = settings(Label::Serialization_Value_Name).toString();
			return createLabel(name);
		});

		ctx.endGroup();
	}

	// Deserialize image editors
	if (ctx.beginGroup(ImageEditorManager::Serialization_Group))
	{
		grinder()->imageEditorManager().deserialize(ctx);
		ctx.endGroup();
	}
}
