/******************************************************************************
 * File: GraphBlockList.cpp
 * Date: 03.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphBlockList.h"
#include "core/GrinderApplication.h"
#include "pipeline/PipelineManager.h"
#include "pipeline/BlockCatalog.h"
#include "ui/StyleSheet.h"
#include "ui/graph/GraphView.h"
#include "ui/graph/GraphBlockNode.h"
#include "res/Resources.h"

#define BLOCKTYPEBUTTON_TYPE_PROPERTY		"blockType"

GraphBlockList::GraphBlockList(QWidget* parent) : QFrame(parent),
	_layout{new QVBoxLayout{this}}, _dragInfo{this, GRAPHBLOCKLIST_DRAG_MIMETYPE, FILE_ICON_BLOCK}
{
	setObjectName("BlockStockpileCtrl");
	setStyleSheet(StyleSheet::loadStyleSheet(FILE_STYLESHEET_BLOCKLIST));

	_layout->setContentsMargins(2, 2, 2, 2);	

	createBlockList();

	// Listen for pipeline switching in order to update the button states
	connect(&grinder()->pipelineController(), &PipelineController::pipelineSwitched, this, &GraphBlockList::pipelineSwitched);
}

bool GraphBlockList::eventFilter(QObject* obj, QEvent* event)
{
	auto button = dynamic_cast<QToolButton*>(obj);

	if (button && button->isEnabled() && event->type() == QEvent::MouseButtonPress)	// Got a click on a type button
	{
		QMouseEvent* mouseEvent = static_cast<QMouseEvent*>(event);

		if (mouseEvent->buttons().testFlag(Qt::LeftButton))
			_dragInfo.initiateDrag(mouseEvent->pos(), button->property(BLOCKTYPEBUTTON_TYPE_PROPERTY));
	}

	return QFrame::eventFilter(obj, event);
}

void GraphBlockList::mouseMoveEvent(QMouseEvent* event)
{
	// If a drag has been initiated and the mouse has been moved far enough, start a drag
	if (_dragInfo.isDragInitiated())
	{
		if (event->buttons().testFlag(Qt::LeftButton))
			_dragInfo.processInitiatedDrag(event->pos());
		else
			_dragInfo.cancelInitiatedDrag();
	}

	QFrame::mouseMoveEvent(event);
}

void GraphBlockList::createBlockList()
{	
	// Add block categories and their associated types
	for (auto category : BlockCatalog::getCategories())
		createBlockCategoryEntry(category);

	_layout->addStretch();

	updateBlockTypeButtonStates(false);
}

void GraphBlockList::createBlockCategoryEntry(BlockCategory category)
{
	auto catLabel = new QLabel{category};
	catLabel->setAlignment(Qt::AlignHCenter);
	catLabel->setFont(GrinderApplication::boldFont(catLabel));

	// Take fill & border color from the graph style and apply it using CSS
	GraphStyle graphStyle;
	QColor borderColor = graphStyle.getBlockNodeStyle().getBlockCategoryColor(category);
	QColor fillColor = borderColor;

	fillColor.setHsl(fillColor.hslHue(), fillColor.hslSaturation(), 255 - (255 - fillColor.lightness()) * 0.4);

	QString css = QString{"QLabel { border: 1px solid %1; background-color: %2; }"}.arg(borderColor.name()).arg(fillColor.name());
	catLabel->setStyleSheet(css);

	_layout->addWidget(catLabel);

	// Create all types under this category
	for (auto type : BlockCatalog::getTypes(category))
		createBlockTypeEntry(type);
}

void GraphBlockList::createBlockTypeEntry(BlockType type)
{
	auto typeButton = new QToolButton{};

	// Get the block's description
	QString description = BlockCatalog::getDescription(type);

	if (description.isEmpty())
		description = "No description available";

	typeButton->setText(type);
	typeButton->setToolTip(description);
	typeButton->setStatusTip(QString{"Create a new %1 block"}.arg(type));
	typeButton->setIcon(QIcon{FILE_ICON_BLOCK});
	typeButton->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
	typeButton->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Preferred);
	typeButton->setAutoRaise(true);
	typeButton->setProperty(BLOCKTYPEBUTTON_TYPE_PROPERTY, type);

	typeButton->connect(typeButton, SIGNAL(clicked(bool)), this, SLOT(blockTypeButtonPressed()));
	typeButton->installEventFilter(this);	// Need to capture mouse events of the button

	_layout->addWidget(typeButton);
}

void GraphBlockList::blockTypeButtonPressed()
{
	BlockType type = sender()->property(BLOCKTYPEBUTTON_TYPE_PROPERTY).toString();

	if (type != BlockType::Undefined)
	{
		if (auto pipeline = grinder()->pipelineController().activePipeline())
		{
			QString newBlockName = StringUtils::generateUniqueItemName(pipeline->blocks(), type, &Block::getName);
			auto block = grinder()->pipelineController().createBlock(type, newBlockName);

			if (auto scene = grinder()->pipelineController().activeScene())
			{
				// Find the just-created block node and center on it
				auto blockNode = scene->findBlockNode(block.get());

				if (blockNode)
				{
					scene->view()->centerOn(blockNode);

					scene->clearSelection();
					blockNode->setSelected(true);
					blockNode->setFocus();
				}
			}
		}
	}
}

void GraphBlockList::updateBlockTypeButtonStates(bool enable) const
{
	for (auto button : findChildren<QToolButton*>())
		button->setEnabled(enable);
}
