/******************************************************************************
 * File: GraphBlockNode.cpp
 * Date: 22.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphBlockNode.h"
#include "GraphStyle.h"
#include "GraphScene.h"
#include "GraphPortNode.h"
#include "pipeline/Block.h"
#include "pipeline/BlockCatalog.h"
#include "core/GrinderApplication.h"
#include "controller/PipelineController.h"
#include "res/Resources.h"

#include <QFontMetricsF>

GraphBlockNode::GraphBlockNode(grndr::GraphScene* scene, const std::shared_ptr<Block>& block, QGraphicsItem* parent) : GraphNode(scene, parent),
	_block{block}, _blockStyle{_style.getBlockNodeStyle()}, _nameEdit{new QLineEdit{}}
{
	if (!block)
		throw std::invalid_argument{_EXCPT("block may not be null")};

	setFlag(ItemIsMovable);

	// Add all ports (in & out) as node children; must be created before updating the geometry
	createPorts();

	updateGeometry();	

	_nameEdit->setStyleSheet("QLineEdit { background: transparent; border: none; }");
	_nameEdit->setAlignment(Qt::AlignCenter);
	_nameEdit->setFont(_blockStyle.nameFont);
	_nameEdit->setContextMenuPolicy(Qt::NoContextMenu);
	_nameEdit->setVisible(false);

	_nameEditWidget = new QGraphicsProxyWidget{this};
	_nameEditWidget->setWidget(_nameEdit);	
	_nameEditWidget->setFlag(QGraphicsItem::ItemIsFocusable);

	// Create node actions
	_renameAction = createNodeAction("&Rename block", FILE_ICON_EDIT, SLOT(beginRenameBlock()), "Rename the selected block", "F2");
	_deleteAction->setText("&Delete block(s)");

	// Set the tooltip to the block's description
	QString description = BlockCatalog::getDescription(block->getType());
	setToolTip(!description.isEmpty() ? description : "No description available");
}

void GraphBlockNode::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	Q_UNUSED(option);
	Q_UNUSED(widget);

	painter->setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform);

	// Draw the node background (including outer glow if selected)
	if (isSelected())
	{
		bool inactive = !_scene->hasFocus();

		painter->save();
		painter->setPen(QPen{Qt::NoPen});
		painter->setBrush(inactive ? _blockStyle.selectionColorInactive : _blockStyle.selectionColor);
		painter->setOpacity(_blockStyle.selectionOpacity);
		painter->drawRoundedRect(_nodeRectSelected, _blockStyle.borderRadius, _blockStyle.borderRadius);
		painter->restore();
	}

	painter->fillRect(_nodeRect, _blockStyle.lightBackgroundColor);

	// Draw node contents
	if (auto block = _block.lock())	// Make sure that the underlying block still exists
		drawHeader(painter, block.get());

	// Draw node borders
	auto borderWidth = _blockStyle.borderWidth;
	auto borderWidthHalf = borderWidth * 0.5;

	QPen pen{_blockStyle.borderColor, borderWidth};
	pen.setCapStyle(Qt::RoundCap);
	pen.setJoinStyle(Qt::RoundJoin);

	painter->setPen(pen);
	painter->drawRoundedRect(_nodeRect, _blockStyle.borderRadius, _blockStyle.borderRadius);

	qreal y = _nodeRect.top() + _geometry.headerRect.height() - borderWidthHalf;
	painter->drawLine(_nodeRect.left() + borderWidthHalf, y, _nodeRect.right() - borderWidthHalf, y);
}

void GraphBlockNode::beginRenameBlock()
{
	if (_isEditingName)
		return;

	if (auto block = _block.lock())	// Make sure that the underlying block still exists
	{
		// Show in-place edit field
		connect(_nameEdit, &QLineEdit::editingFinished, this, &GraphBlockNode::endRenameBlock);

		auto headerRect = _nodeRect;
		headerRect.setHeight(_geometry.headerNameSize.height());

		_nameEditWidget->setGeometry(headerRect);
		_nameEditWidget->setVisible(true);

		_nameEdit->setText(block->getName());
		_nameEdit->selectAll();
		_nameEdit->setFocus();

		_isEditingName = true;

		update();
	}
}

void GraphBlockNode::endRenameBlock()
{
	if (!_isEditingName)
		return;

	disconnect(_nameEdit, nullptr, this, nullptr);

	if (auto block = _block.lock())	// Make sure that the underlying block still exists
	{
		// Hide in-place edit field and commit new name
		auto name = _nameEdit->text().trimmed();
		_nameEditWidget->setVisible(false);

		if (!name.isEmpty())
		{
			grinder()->pipelineController().renameBlock(block.get(), name);

			prepareGeometryChange();
			updateGeometry();
		}
	}

	_isEditingName = false;

	update();
	setFocus();
}

void GraphBlockNode::cancelRenameBlock()
{
	if (!_isEditingName)
		return;

	// Hide in-place edit field w/o committing new name
	disconnect(_nameEdit, nullptr, this, nullptr);

	_nameEditWidget->setVisible(false);

	_isEditingName = false;

	update();
	setFocus();
}

void GraphBlockNode::keyReleaseEvent(QKeyEvent* event)
{
	if (event->modifiers() == Qt::NoModifier)
	{
		if (event->key() == Qt::Key_F2 && !_isEditingName)	// Begin rename
		{
			beginRenameBlock();
			return;
		}
		else if (event->key() == Qt::Key_Escape)
		{
			if (_isEditingName)	// Cancel name edit
				cancelRenameBlock();
			else if (_connectionBlueprint)	// Cancel connection initiation
				endConnectionInitiation(false);

			return;
		}
	}

	GraphNode::keyReleaseEvent(event);
}

void GraphBlockNode::mousePressEvent(QGraphicsSceneMouseEvent* event)
{
	if (event->button() == Qt::LeftButton)
	{
		if (!_connectionBlueprint)	// Check if a new connection initiation must be started
		{
			// See if a port rect has been clicked; if so, initiate a new connection
			if (auto portNode = findPortByPosition(event->pos()))
				beginConnectionInitiation(portNode);

			updateConnectionInitiation(event->scenePos());
		}
	}

	if (!_connectionBlueprint)
		GraphNode::mousePressEvent(event);
}

void GraphBlockNode::mouseMoveEvent(QGraphicsSceneMouseEvent* event)
{
	if (_connectionBlueprint)
		updateConnectionInitiation(event->scenePos());
	else
		GraphNode::mouseMoveEvent(event);
}

void GraphBlockNode::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
	// Mouse-button has been released after a drag operation, so end any connection initiation
	if (_connectionBlueprint)
		endConnectionInitiation();

	GraphNode::mouseReleaseEvent(event);
}

void GraphBlockNode::mouseDoubleClickEvent(QGraphicsSceneMouseEvent* event)
{
	// If the header is double-click, initiate rename
	if (_geometry.headerRect.contains(event->pos()))
		beginRenameBlock();
}

void GraphBlockNode::updateGeometry()
{
	if (auto block = _block.lock())	// Make sure that the underlying block still exists
	{
		_geometry.headerNameSize = calcHeaderNameSize(block.get());
		_geometry.headerTypeSize = calcHeaderTypeSize(block.get());
		_geometry.inPortsSize = calcPortsSize(_inPortNodes);
		_geometry.outPortsSize = calcPortsSize(_outPortNodes);

		auto totalWidth = std::max({_geometry.headerNameSize.width(), _geometry.headerTypeSize.width(), _geometry.inPortsSize.width() + _geometry.outPortsSize.width()});
		auto totalHeight = _geometry.headerNameSize.height() + _geometry.headerTypeSize.height() + std::max(_geometry.inPortsSize.height(), _geometry.outPortsSize.height());

		auto minSize = _blockStyle.minimumSize;
		_nodeRect = QRectF{0, 0, std::max(totalWidth, minSize.width()), std::max(totalHeight, minSize.height())};
		_nodeRectSelected = _nodeRect + QMarginsF{_blockStyle.selectionMargin, _blockStyle.selectionMargin, _blockStyle.selectionMargin, _blockStyle.selectionMargin};

		_geometry.headerRect.setRect(_nodeRect.left(), _nodeRect.top(), _nodeRect.width(), _geometry.headerNameSize.height() + _geometry.headerTypeSize.height());
		_geometry.portsRect.setRect(_nodeRect.left(), _geometry.headerRect.bottom(), _nodeRect.width(), std::max(_geometry.inPortsSize.height(), _geometry.outPortsSize.height()));
	}

	// The geometry has been updated to make enough room for all ports, so bring them into layout
	layoutPorts(_inPortNodes, false);
	layoutPorts(_outPortNodes, true);

	GraphNode::updateGeometry();
}

std::vector<QAction*> GraphBlockNode::getNodeActions(QMenu& menu) const
{
	std::vector<QAction*> actions = GraphNode::getNodeActions(menu);
	actions.insert(actions.begin(), _renameAction);
	return actions;
}

QSizeF GraphBlockNode::calcHeaderNameSize(const Block* block) const
{
	auto marginsX = _blockStyle.textMargins.left() + _blockStyle.textMargins.right();
	auto marginsY = _blockStyle.textMargins.top() + _blockStyle.textMargins.bottom();

	QFontMetricsF fontNameMetrics{_blockStyle.nameFont};
	auto nameWidth = fontNameMetrics.width(block->getName()) + marginsX;
	auto nameHeight = fontNameMetrics.height() + marginsY;

	return QSizeF{nameWidth, nameHeight};
}

QSizeF GraphBlockNode::calcHeaderTypeSize(const Block* block) const
{
	auto marginsX = _blockStyle.textMargins.left() + _blockStyle.textMargins.right();
	auto marginsY = _blockStyle.textMargins.top() + _blockStyle.textMargins.bottom();

	QFontMetricsF fontTypeMetrics{_blockStyle.typeFont};
	auto typeWidth = fontTypeMetrics.width(block->getType()) + marginsX;
	auto typeHeight = fontTypeMetrics.height() + marginsY;

	return QSizeF{typeWidth, typeHeight};
}

QSizeF GraphBlockNode::calcPortsSize(const std::vector<GraphPortNode*>& ports) const
{
	QSizeF size{0.0, 0.0};

	for (auto portNode : ports)
	{
		auto rect = portNode->boundingRect() + _blockStyle.portMargins;

		size.setWidth(std::max(size.width(), rect.width()));
		size.setHeight(size.height() + rect.height());
	}

	return size;
}

void GraphBlockNode::drawHeader(QPainter* painter, const Block* block)
{
	painter->setPen(_blockStyle.textColor);

	auto rect = _geometry.headerRect;

	// Draw block name
	rect.setBottom(rect.top() + _geometry.headerNameSize.height());

	QLinearGradient grad{0.0, rect.top(), 0.0, rect.bottom()};
	grad.setColorAt(0.0, _blockStyle.getBlockCategoryColor(block->getCategory()));
	grad.setColorAt(1.0, _blockStyle.lightBackgroundColor);

	painter->fillRect(rect, QBrush{grad});

	if (!_isEditingName)
	{
		painter->setFont(_blockStyle.nameFont);
		painter->drawText(rect, Qt::AlignHCenter|Qt::AlignVCenter, block->getName());
	}

	// Draw block type
	rect.setTop(rect.bottom());
	rect.setBottom(rect.top() + _geometry.headerTypeSize.height());

	painter->setFont(_blockStyle.typeFont);
	painter->drawText(rect, Qt::AlignHCenter|Qt::AlignTop, block->getType());
}

void GraphBlockNode::createPorts()
{
	if (auto block = _block.lock())	// Make sure that the underlying block still exists
	{
		// Add incoming ports
		auto inPorts = block->ports().selectByDirection(Port::Direction::In);

		for (const auto& port : inPorts)
		{
			GraphPortNode* portNode = new GraphPortNode{this, _scene, port, this};
			_inPortNodes.push_back(portNode);
		}

		// Add outgoing ports
		auto outPorts = block->ports().selectByDirection(Port::Direction::Out);

		for (const auto& port : outPorts)
		{
			GraphPortNode* portNode = new GraphPortNode{this, _scene, port, this};
			_outPortNodes.push_back(portNode);
		}
	}
}

void GraphBlockNode::layoutPorts(const std::vector<grndr::GraphPortNode*>& ports, bool alignRight)
{
	QPointF pos{alignRight ? _geometry.portsRect.topRight() : _geometry.portsRect.topLeft()};

	// Move all ports into their proper place; in-ports on the left, out-ports on the right
	for (const auto& portNode : ports)
	{
		auto rect = portNode->boundingRect();
		auto x = alignRight ? pos.x() - rect.width() - _blockStyle.portMargins.right() : pos.x() + _blockStyle.portMargins.left();
		auto y = pos.y() + _blockStyle.portMargins.top();

		portNode->setPos(x, y);
		pos.setY(pos.y() + rect.height() + _blockStyle.portMargins.top() + _blockStyle.portMargins.bottom());
	}
}

GraphPortNode* GraphBlockNode::findPortByPosition(QPointF pos)
{
	// See if there is a port at position pos
	auto findPort = [pos](const auto& nodes) -> GraphPortNode* {
		for (const auto& node : nodes)
		{
			auto rect = node->mapRectToParent(node->getPortRect());

			if (rect.contains(pos))
				return node;
		}

		return nullptr;
	};

	auto port = findPort(_inPortNodes);

	if (!port)
		port = findPort(_outPortNodes);

	return port;
}

void GraphBlockNode::beginConnectionInitiation(GraphPortNode* portNode)
{
	if (!_connectionBlueprint)
	{
		_connectionBlueprint = std::make_unique<GraphConnectionBlueprint>(_scene, portNode);

		_scene->addItem(_connectionBlueprint.get());
		_scene->update();
	}
}

void GraphBlockNode::updateConnectionInitiation(QPointF mousePos)
{
	if (_connectionBlueprint)
	{
		_connectionBlueprint->setTargetPos(mousePos);

		// See if there is a block node under the cursor
		bool lockOnPort = false;

		for (auto item : _scene->items(mousePos))
		{
			if (auto blockNode = dynamic_cast<GraphBlockNode*>(item))
			{
				if (blockNode)
				{
					// We're above a block, so let's see if we are also above a port
					if (auto portNode = blockNode->findPortByPosition(blockNode->mapFromScene(mousePos)))
					{
						_connectionBlueprint->setTargetPortNode(portNode);

						lockOnPort = true;
						break;
					}
				}
			}
		}

		if (!lockOnPort)
			_connectionBlueprint->setTargetPortNode(nullptr);

		_scene->update();
	}
}

void GraphBlockNode::endConnectionInitiation(bool createConnection)
{
	if (_connectionBlueprint)
	{
		if (createConnection)
		{
			auto sourcePort = _connectionBlueprint->connectionPort(Port::Direction::Out);
			auto destPort = _connectionBlueprint->connectionPort(Port::Direction::In);

			try {
				grinder()->pipelineController().validateConnection(sourcePort, destPort);

				// We're fine, the connection is valid
				grinder()->pipelineController().createConnection(sourcePort, destPort);
			} catch (...) {
				// Ignore any exceptions (simply don't add the invalid connection)
			}
		}

		_scene->removeItem(_connectionBlueprint.get());
		_scene->update();

		_connectionBlueprint.reset(nullptr);
	}
}
