/******************************************************************************
 * File: GraphConnectionBlueprint.cpp
 * Date: 27.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphConnectionBlueprint.h"
#include "GraphBlockNode.h"
#include "GraphPortNode.h"
#include "GraphScene.h"
#include "core/GrinderApplication.h"
#include "controller/PipelineController.h"

GraphConnectionBlueprint::GraphConnectionBlueprint(GraphScene* scene, grndr::GraphPortNode* startPortNode, QGraphicsItem* parent) : GraphConnectionBase(scene, parent),
	_startPortNode{startPortNode}
{
	if (!startPortNode)
		throw std::invalid_argument{_EXCPT("startPortNode may not be null")};

	setFlag(ItemIsFocusable, false);
	setFlag(ItemIsSelectable, false);
	setFlag(ItemSendsGeometryChanges, false);

	// Make sure that the blueprint is always drawn on top
	setZValue(5.0f);

	// Create a hidden text node for showing the rejection reason
	_statusMessageNode = new GraphConnectionMessage{_scene, this};
	_statusMessageNode->setZValue(zValue() + 1.0);
	_statusMessageNode->setVisible(false);
}

void GraphConnectionBlueprint::setTargetPortNode(GraphPortNode* node)
{
	_targetPortNode = node;

	if (_targetPortNode && _targetPortNode != _startPortNode)	// A (potential) connection has been established
	{
		validateConnection();
	}
	else
	{
		_status = Status::Pending;
		_statusMessageNode->setVisible(false);
	}

	prepareGeometryChange();
	updateGeometry();
}

void GraphConnectionBlueprint::setTargetPos(QPointF pos)
{
	_targetPos = pos;

	prepareGeometryChange();
	updateGeometry();
}

Port* GraphConnectionBlueprint::_connectionPort(Port::Direction dir) const
{
	// Find the connection port that matches the given direction
	auto startPort = _startPortNode ? _startPortNode->port().lock() : nullptr;
	auto targetPort = _targetPortNode ? _targetPortNode->port().lock() : nullptr;

	if (startPort && startPort->getDirection() == dir)
		return startPort.get();
	else if (targetPort && targetPort->getDirection() == dir)
		return targetPort.get();
	else
		return nullptr;
}

void GraphConnectionBlueprint::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	Q_UNUSED(option);
	Q_UNUSED(widget);

	painter->setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform);
	painter->setOpacity(_connectionStyle.blueprintOpacity);

	QColor color;

	switch (_status)
	{
	case Status::Valid:
		color = _connectionStyle.blueprintValidColor;
		break;

	case Status::Invalid:
		color = _connectionStyle.blueprintInvalidColor;
		break;

	default:
		color = _connectionStyle.blueprintPendingColor;
	}

	drawConnection(painter, color, color);
}

QPointF GraphConnectionBlueprint::calcSourcePosition() const
{
	auto portNode = _targetPortNode;
	auto startPort = _startPortNode->port().lock();

	if (startPort && startPort->isOut())
		portNode = _startPortNode;

	if (portNode)
		return portNode->blockNode()->mapToScene(portNode->getCenterPos(false));
	else
		return _targetPos;
}

QPointF GraphConnectionBlueprint::calcDestPosition() const
{
	auto portNode = _targetPortNode;
	auto startPort = _startPortNode->port().lock();

	if (startPort && startPort->isIn())
		portNode = _startPortNode;

	if (portNode)
		return portNode->blockNode()->mapToScene(portNode->getCenterPos(false));
	else
		return _targetPos;
}

void GraphConnectionBlueprint::validateConnection()
{
	_status = Status::Invalid;
	_statusMessage = "Unknown rejection reason";

	auto sourcePort = connectionPort(Port::Direction::Out);
	auto destPort = connectionPort(Port::Direction::In);

	try {
		grinder()->pipelineController().validateConnection(sourcePort, destPort);

		// We're fine, the connection is valid
		_status = Status::Valid;
		_statusMessage = "";
	} catch (std::exception& e) {
		_statusMessage = GetExceptionMessage(e.what());
	}

	// Update the status message node
	_statusMessageNode->setVisible(_status == Status::Invalid);
	_statusMessageNode->setText(_statusMessage);
	_statusMessageNode->setPos(boundingRect().center() - _statusMessageNode->boundingRect().center());
}
