/******************************************************************************
 * File: GraphConnectionNode.cpp
 * Date: 25.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphConnectionNode.h"
#include "GraphBlockNode.h"
#include "GraphPortNode.h"
#include "GraphScene.h"
#include "pipeline/Block.h"
#include "pipeline/Port.h"
#include "pipeline/Connection.h"

GraphConnectionNode::GraphConnectionNode(GraphScene* scene, const std::shared_ptr<Connection>& con, QGraphicsItem* parent) : GraphConnectionBase(scene, parent),
	_connection{con}
{
	if (!con)
		throw std::invalid_argument{_EXCPT("con may not be null")};

	connectPortSignals();

	updateGeometry();

	// Create node actions
	_deleteAction->setText("&Remove connection(s)");

	// Set the tooltip showing information about this connection
	updateToolTip();
}

GraphPortNode* GraphConnectionNode::_sourcePortNode() const
{
	if (auto con = _connection.lock())	// Make sure that the underlying connection still exists
		return _scene->findPortNode(con->sourcePort());
	else
		return nullptr;
}

GraphPortNode* GraphConnectionNode::_destPortNode() const
{
	if (auto con = _connection.lock())	// Make sure that the underlying connection still exists
		return _scene->findPortNode(con->destPort());
	else
		return nullptr;
}

void GraphConnectionNode::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	Q_UNUSED(option);
	Q_UNUSED(widget);

	painter->setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing);

	if (auto con = _connection.lock())	// Make sure that the underlying connection still exists
	{		
		if (isSelected())	// Draw outer glow if selected
		{
			bool inactive = !_scene->hasFocus();
			QPen selectionPen{QBrush{inactive ? _connectionStyle.selectionColorInactive : _connectionStyle.selectionColor}, _connectionStyle.lineWidth + _connectionStyle.selectionMargin};
			selectionPen.setJoinStyle(Qt::RoundJoin);

			painter->save();
			painter->setPen(selectionPen);
			painter->setOpacity(_connectionStyle.selectionOpacity);
			painter->drawPath(_geometry.connectionPath);
			painter->restore();
		}

		drawConnection(painter, getPortColor(sourcePortNode()), getPortColor(destPortNode()));
	}
}

QPointF GraphConnectionNode::calcSourcePosition() const
{
	if (auto portNode = sourcePortNode())
		return portNode->blockNode()->mapToScene(portNode->getCenterPos());
	else
		return QPointF{};
}

QPointF GraphConnectionNode::calcDestPosition() const
{
	if (auto portNode = destPortNode())
		return portNode->blockNode()->mapToScene(portNode->getCenterPos());
	else
		return QPointF{};
}

void GraphConnectionNode::updateToolTip()
{
	if (auto con = _connection.lock())	// Make sure that the underlying connection still exists
	{
		QString toolTip = QString{"<nobr><b>From: </b>%1</nobr><br><nobr><b>To: </b>%2</nobr>"}
				.arg(con->sourcePort()->getFormattedName())
				.arg(con->destPort()->getFormattedName());

		setToolTip(toolTip);
	}
}

void GraphConnectionNode::connectPortSignals()
{
	// Remove any previous signals
	disconnect(nullptr, this, SLOT(connectedBlockMoved(VisualNode*)));

	// Listen for movement events of both involved blocks in order to update our own shape
	auto sourceNode = sourcePortNode();
	auto destNode = destPortNode();

	if (sourceNode)
	{
		connect(sourceNode->blockNode(), &GraphBlockNode::nodeMoved, this, &GraphConnectionNode::connectedBlockMoved);
		connect(sourceNode->blockNode(), &GraphBlockNode::nodeGeometryUpdated, this, &GraphConnectionNode::connectedBlockMoved);
	}

	if (destNode)
	{
		connect(destNode->blockNode(), &GraphBlockNode::nodeMoved, this, &GraphConnectionNode::connectedBlockMoved);
		connect(destNode->blockNode(), &GraphBlockNode::nodeGeometryUpdated, this, &GraphConnectionNode::connectedBlockMoved);
	}
}

QColor GraphConnectionNode::getPortColor(const GraphPortNode* portNode) const
{
	// Get the color of the block associated with the given port
	if (portNode)
	{
		if (auto port = portNode->port().lock())
			return _style.getBlockNodeStyle().getBlockCategoryColor(port->block()->getCategory());
	}

	return QColor{};
}

void GraphConnectionNode::connectedBlockMoved(VisualNode* node)
{
	Q_UNUSED(node);

	// One of the connected blocks has been moved, so we need to update our geometry
	prepareGeometryChange();
	updateGeometry();
}
