/******************************************************************************
 * File: GraphLayout.cpp
 * Date: 23.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphLayout.h"
#include "GraphBlockNode.h"
#include "GraphView.h"
#include "GraphScene.h"
#include "pipeline/Pipeline.h"
#include "util/StringConv.h"
#include "util/SerializationUtils.h"

const char* GraphLayout::Serialization_Group_Item = "Item";

const char* GraphLayout::Serialization_Value_Block = "Block";
const char* GraphLayout::Serialization_Value_Position = "Position";

void GraphLayout::applyLayout(GraphScene* scene, bool centerLayout) const
{
	qreal maxX = 0.0;
	qreal maxY = 0.0;

	if (centerLayout)
	{
		// Adjust layout to be centered
		for (const auto& layoutItem : _layout)
		{
			if (auto block = layoutItem->_block.lock())	// Make sure that the block still exists
			{
				if (auto blockNode = scene->findBlockNode(block.get()))
				{
					auto rect = blockNode->boundingRect(false);

					maxX = std::max(maxX, layoutItem->_position.x() + rect.width()) * 0.5;
					maxY = std::max(maxY, layoutItem->_position.y() + rect.height()) * 0.5;
				}
			}
		}
	}

	for (const auto& layoutItem : _layout)
	{
		if (auto block = layoutItem->_block.lock())	// Make sure that the block still exists
		{
			if (auto blockNode = scene->findBlockNode(block.get()))
				blockNode->setPos(QPointF{layoutItem->_position.x() - maxX, layoutItem->_position.y() - maxY});
		}
	}
}

void GraphLayout::fromScene(GraphScene* scene)
{
	if (!scene)
		throw std::invalid_argument{_EXCPT("scene may not be null")};

	_layout.clear();

	for (auto blockNode : scene->getNodes<GraphBlockNode>())
		_layout.emplace_back(std::make_shared<LayoutItem>(blockNode->_block, blockNode->pos()));
}

void GraphLayout::fromBlockHierarchy(GraphScene* scene, const BlockHierarchy& blockHierarchy)
{
	if (!scene)
		throw std::invalid_argument{_EXCPT("scene may not be null")};

	_layout.clear();

	QPointF currentPos{0.0, 0.0};
	auto layoutStyle = scene->view()->sceneStyle().getLayoutStyle();

	// Each level is layed out on a column
	for (const auto& level : blockHierarchy)
	{
		auto columnSize = addHierarchyLevel(scene, currentPos, level);

		currentPos.setX(currentPos.x() + columnSize.width() + layoutStyle.layoutMargins.width());
		currentPos.setY(0.0);
	}
}

void GraphLayout::serialize(SerializationContext& ctx) const
{
	SerializationUtils::serializeContainer(_layout, Serialization_Group_Item, ctx);
}

void GraphLayout::deserialize(const Pipeline* pipeline, DeserializationContext& ctx)
{
	SerializationUtils::deserializeContainer<Layout>(Serialization_Group_Item, ctx, [&ctx, pipeline, this](const SettingsContainer& settings) -> std::shared_ptr<LayoutItem> {
		int blockIndex = settings(Serialization_Value_Block, -1).toInt();
		QPointF position = StringConv::convertString<QPointF>(settings(Serialization_Value_Position).toString());
		auto block = ctx.getBlock(blockIndex);

		if (block)
		{
			// Get the shared_ptr belonging to the block
			auto it = pipeline->blocks().find(block);

			if (it != pipeline->blocks().cend())
				_layout.emplace_back(std::make_shared<LayoutItem>(*it, position));
		}

		// We don't need to explicitly deserialize the new layout item
		return nullptr;
	});
}

QSizeF GraphLayout::addHierarchyLevel(GraphScene* scene, QPointF startPos, const BlockHierarchy::HierarchyLevel& blocks)
{
	auto layoutStyle = scene->view()->sceneStyle().getLayoutStyle();
	auto currentPos = startPos;
	qreal width = 0.0;

	for (const auto& block : blocks)
	{
		auto blockNode = scene->findBlockNode(block);

		if (blockNode)
		{
			_layout.emplace_back(std::make_shared<LayoutItem>(blockNode->_block, currentPos));

			auto rect = blockNode->boundingRect(false);
			width = std::max(width, rect.width());
			currentPos.setY(currentPos.y() + rect.height() + layoutStyle.layoutMargins.height());
		}
	}

	return QSizeF{width, currentPos.y()};
}

void GraphLayout::LayoutItem::serialize(SerializationContext& ctx) const
{
	if (auto block = _block.lock())	// Make sure that the block still exists
	{
		ctx.settings()(Serialization_Value_Block) = ctx.getBlockIndex(block.get());
		ctx.settings()(Serialization_Value_Position) = StringConv::convertValue(_position);
	}
}

void GraphLayout::LayoutItem::deserialize(DeserializationContext& ctx)
{
	// Nothing to deserialize (all done by the superordinate layout)
	Q_UNUSED(ctx);
}
