/******************************************************************************
 * File: GraphPortNode.cpp
 * Date: 24.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphPortNode.h"
#include "GraphBlockNode.h"
#include "GraphConnectionBlueprint.h"
#include "GraphScene.h"
#include "core/GrinderApplication.h"
#include "pipeline/Port.h"
#include "controller/PipelineController.h"
#include "controller/EngineController.h"
#include "util/DataUtils.h"
#include "res/Resources.h"

#include <opencv2/highgui.hpp>

GraphPortNode::GraphPortNode(GraphBlockNode* blockNode, grndr::GraphScene* scene, const std::shared_ptr<grndr::Port>& port, QGraphicsItem* parent) : GraphNode(scene, parent),
	_blockNode{blockNode}, _port{port}, _portStyle{_style.getPortNodeStyle()}
{
	if (!blockNode)
		throw std::invalid_argument{_EXCPT("blockNode may not be null")};

	if (!port)
		throw std::invalid_argument{_EXCPT("port may not be null")};

	setFlag(ItemIsSelectable, false);
	setFlag(ItemIsFocusable, false);

	updateGeometry();

	// Create node actions
	_viewImageAction = createNodeAction("&View image", FILE_ICON_VIEWIMAGE, SLOT(viewImage()), "View the image at this port");
	_disconnectAction = createNodeAction("&Disconnect from all", FILE_ICON_DELETE, SLOT(disconnectFromAll()), "Remove all connections to/from this port");

	// Set the tooltip showing what this port is accepting/offering
	updateToolTip();
}

QPointF GraphPortNode::getCenterPos(bool shiftCenter) const
{
	auto centerPos = pos() + _geometry.portRect.center();

	if (shiftCenter)
	{
		if (auto port = _port.lock())	// Make sure that the underlying port still exists
		{
			// Slightly shift the center point right/left (makes creating connections easier)
			auto shift = _portStyle.innerSize + _style.getConnectionNodeStyle().lineWidth / 2.0;

			if (port->isOut())
				centerPos.setX(centerPos.x() + shift);
			else if (port->isIn())
				centerPos.setX(centerPos.x() - shift);
		}
	}

	return centerPos;
}

void GraphPortNode::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	Q_UNUSED(option);
	Q_UNUSED(widget);

	painter->setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform);

	// Draw the port connector
	QPen pen{_portStyle.borderColor, _portStyle.borderWidth};
	pen.setCapStyle(Qt::FlatCap);
	pen.setJoinStyle(Qt::MiterJoin);

	painter->setPen(pen);
	painter->fillRect(_geometry.portRect, _portStyle.darkBackgroundColor);
	painter->drawRect(_geometry.portRect);

	auto sizeDiff = (_portStyle.outerSize - _portStyle.innerSize) / 2.0;
	auto innerRect = _geometry.portRect - QMarginsF{sizeDiff, sizeDiff, sizeDiff, sizeDiff};

	if (auto port = _port.lock())	// Make sure that the underlying port still exists
	{
		if (!port->isConnected())
			painter->fillRect(innerRect, _portStyle.lightBackgroundColor);

		painter->drawRect(innerRect);

		// Draw the port name
		painter->setPen(QPen{_portStyle.textColor});
		painter->setFont(_portStyle.nameFont);
		painter->drawText(_geometry.portNameRect, (port->isIn() ? Qt::AlignLeft : Qt::AlignRight)|Qt::AlignVCenter, port->getName());
	}
}

void GraphPortNode::contextMenuEvent(QGraphicsSceneContextMenuEvent* event)
{
	// Select the block node exclusively if it hasn't been selected
	if (!_blockNode->isSelected())
	{
		_scene->clearSelection();
		_blockNode->setSelected(true);
	}

	showContextMenu(event->screenPos());

	event->accept();
}

void GraphPortNode::updateGeometry()
{
	if (auto port = _port.lock())	// Make sure that the underlying port still exists
	{
		_geometry.portRect = QRectF{0.0, 0.0, _portStyle.outerSize, _portStyle.outerSize};
		_geometry.portNameSize = calcPortNameSize(port.get());

		_nodeRect = QRectF{0.0, 0.0, _geometry.portRect.width() + _geometry.portNameSize.width(), std::max(_geometry.portRect.height(), _geometry.portNameSize.height())};
		_nodeRectSelected = _nodeRect;

		_geometry.portNameRect = _nodeRect;

		if (port->isOut())
		{
			// Update geometry for a right-aligned port
			_geometry.portRect.moveLeft(_nodeRect.right() - _geometry.portRect.width());
			_geometry.portNameRect.setRight(_geometry.portNameRect.right() - (_geometry.portRect.width() + _portStyle.textMargins.right()));
		}
		else
			_geometry.portNameRect.setLeft(_geometry.portNameRect.left() + _geometry.portRect.width() + _portStyle.textMargins.left());
	}

	GraphNode::updateGeometry();
}

QSizeF GraphPortNode::calcPortNameSize(const Port* port) const
{
	auto marginsX = _portStyle.textMargins.left() + _portStyle.textMargins.right();
	auto marginsY = _portStyle.textMargins.top() + _portStyle.textMargins.bottom();

	QFontMetricsF fontNameMetrics{_portStyle.nameFont};
	auto nameWidth = fontNameMetrics.width(port->getName()) + marginsX;
	auto nameHeight = fontNameMetrics.height() + marginsY;

	return QSizeF{nameWidth, nameHeight};
}

void GraphPortNode::viewImage()
{
	if (auto port = _port.lock())	// Make sure that the underlying port still exists
	{
		auto portImage = grinder()->engineController().executeLabelEx(nullptr, port.get(), nullptr, Engine::ExecutionMode::Execute);

		if (!portImage.empty())
			cv::imshow(port->getFormattedName().toStdString(), portImage);
	}
}

void GraphPortNode::disconnectFromAll()
{
	if (auto port = _port.lock())	// Make sure that the underlying port still exists
		grinder()->pipelineController().disconnectPortFromAll(port.get());
}

void GraphPortNode::updateToolTip()
{
	if (auto port = _port.lock())	// Make sure that the underlying port still exists
	{
		QString toolTip = QString{"<nobr><b>%1</b></nobr>"}.arg(port->getFormattedName());
		QStringList dataDescInfos;

		for (const auto& dataDesc : port->dataDescriptors())
		{
			dataDescInfos << QString{"<br><u>%1</u><br><em>&nbsp;&nbsp;Structure type: </em>%2<br><em>&nbsp;&nbsp;Field type: </em>%3<br><em>&nbsp;&nbsp;Value type: </em>%4"}
					 .arg(dataDesc.getName())
					 .arg(DataUtils::getDataDescriptorTypeName(dataDesc.getStructureType()))
					 .arg(DataUtils::getDataDescriptorTypeName(dataDesc.getFieldType()))
					 .arg(DataUtils::getDataDescriptorTypeName(dataDesc.getValueType()));
		}

		setToolTip(toolTip + dataDescInfos.join("<br>"));
	}
}

std::vector<QAction*> GraphPortNode::getNodeActions(QMenu& menu) const
{
	Q_UNUSED(menu);

	_viewImageAction->setEnabled(false);
	_disconnectAction->setEnabled(false);

	if (auto port = _port.lock())	// Make sure that the underlying port still exists
	{
		// Check if this port has at least one matrix type
		bool hasMatrixType = false;

		for (const auto& dataDesc : port->dataDescriptors())
		{
			if (dataDesc.getStructureType() == DataDescriptor::StructureType::Matrix)
			{
				hasMatrixType = true;
				break;
			}
		}

		if (hasMatrixType && (port->isOut() || (port->isIn() && port->isConnected())))	// In-ports must be connected to a source
			_viewImageAction->setEnabled(true);

		// Check if this port is connected to something
		if (!port->connections().empty() || !port->getConnections(Port::Direction::In).empty())
			_disconnectAction->setEnabled(true);
	}

	return {_viewImageAction, nullptr, _disconnectAction};
}
