/******************************************************************************
 * File: GraphScene.cpp
 * Date: 18.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphScene.h"
#include "GraphView.h"
#include "GraphStyle.h"
#include "GraphBlockNode.h"
#include "GraphPortNode.h"
#include "GraphConnectionNode.h"
#include "pipeline/Pipeline.h"

GraphScene::GraphScene(Pipeline* pipeline, GraphView* view, QSizeF initialSize) : VisualScene{view, QRectF{-(initialSize.width() / 2), -(initialSize.height() / 2), initialSize.width(), initialSize.height()}},
	_pipeline{pipeline}, _nodeFactory{this}
{
	if (!pipeline)
		throw std::invalid_argument{_EXCPT("pipeline may not be null")};
}

void GraphScene::buildScene()
{
	clear();

	// Add all elements of the pipeline of the label to the scene
	if (_pipeline)
	{
		// Add all blocks to the scene
		for (const auto& block : _pipeline->blocks())
			createBlockNode(block);

		// Add all connections to the scene
		for (const auto& block : _pipeline->blocks())
		{
			for (const auto& port : block->ports().selectByDirection(Port::Direction::Out))
			{
				for (const auto& con : port->connections())
					createConnectionNode(con);
			}
		}
	}
}

void GraphScene::createBlockNode(const std::shared_ptr<Block>& block)
{
	if (!block)
		throw std::invalid_argument{_EXCPT("block may not be null")};

	if (!findBlockNode(block.get()))
		addItem(_nodeFactory.createNode(block));
}

void GraphScene::removeBlockNode(GraphBlockNode* blockNode)
{
	if (blockNode)
		delete blockNode;
}

void GraphScene::createConnectionNode(const std::shared_ptr<Connection>& connection)
{
	if (!connection)
		throw std::invalid_argument{_EXCPT("connection may not be null")};

	if (!findConnectionNode(connection.get()))
		addItem(new GraphConnectionNode{this, connection});
}

void GraphScene::removeConnectionNode(GraphConnectionNode* connectionNode)
{
	if (connectionNode)
		delete connectionNode;
}

GraphBlockNode* GraphScene::_findBlockNode(const Block* block) const
{
	for (const auto& node : items())
	{
		auto blockNode = dynamic_cast<GraphBlockNode*>(node);

		if (blockNode && blockNode->block().lock().get() == block)
			return blockNode;
	}

	return nullptr;
}

GraphPortNode* GraphScene::_findPortNode(const Port* port) const
{
	auto blockNode = findBlockNode(port->block());

	if (blockNode)
	{
		auto searchPorts = [port](const std::vector<GraphPortNode*>& portNodes) -> GraphPortNode* {
			auto it = std::find_if(portNodes.cbegin(), portNodes.cend(), [port](auto portNode) { return portNode->port().lock().get() == port; });

			if (it != portNodes.cend())
				return *it;

			return nullptr;
		};

		if (auto inPort = searchPorts(blockNode->inPortNodes()))
			return inPort;

		if (auto outPort = searchPorts(blockNode->outPortNodes()))
			return outPort;
	}

	return nullptr;
}

GraphConnectionNode* GraphScene::_findConnectionNode(const Connection* con) const
{
	for (const auto& node : items())
	{
		auto conNode = dynamic_cast<GraphConnectionNode*>(node);

		if (conNode && conNode->connection().lock().get() == con)
			return conNode;
	}

	return nullptr;
}
