/******************************************************************************
 * File: DraftItemNode.cpp
 * Date: 21.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "DraftItemNode.h"
#include "ImageEditor.h"
#include "ImageTagsListWidget.h"
#include "image/DraftItem.h"
#include "image/ImageTags.h"
#include "util/UIUtils.h"
#include "util/MathUtils.h"
#include "res/Resources.h"

DraftItemNode::DraftItemNode(ImageEditorScene* scene, const std::shared_ptr<DraftItem>& item, QGraphicsItem* parent) : LayerItemNode(scene, item->layer(), parent),
	_draftItem{item}, _draftItemStyle{_style.getItemNodeStyle()}
{
	if (!item)
		throw std::invalid_argument{_EXCPT("item may not be null")};

	setFlag(ItemIsMovable);
	setAcceptDrops(true);

	if (auto draftItem = _draftItem.lock())	// Make sure that the underlying draft item still exists
	{
		// Get a renderer for the draft item
		_renderer = draftItem->createRenderer(_draftItemStyle);

		// Whenever a property of the draft item changes, update the node to reflect the new values
		for (auto& property : draftItem->properties())
			connect(property.get(), &PropertyBase::valueChanged, this, &DraftItemNode::updateNode);

		// Whenever image tags in the allotment have been modified, update the node as well
		connect(&draftItem->imageTagsAllotment()->object(), &ImageTagsAllotment::allotmentModified, this, &DraftItemNode::updateNode);
	}

	updateZOrder();
	updateVisibility();
	updateAlpha();

	// Create node actions
	_deleteAction = createNodeAction("&Delete item(s)", FILE_ICON_DELETE, SLOT(deleteNode()), "Remove the selected items", "Del");
}

void DraftItemNode::initDraftItemNode()
{
	_inPlaceEditor = createInPlaceEditor();

	if (_inPlaceEditor)
	{
		_inPlaceEditor->setParentItem(this);
		_inPlaceEditor->setVisible(false);
	}
}

void DraftItemNode::updateNode()
{
	if (auto draftItem = _draftItem.lock())	// Make sure that the underlying draft item still exists
	{
		// Reflect the draft item's position
		QPoint pos = *draftItem->position();
		setPos(QPointF{static_cast<qreal>(pos.x()), static_cast<qreal>(pos.y())});

		// Set the item's tool tip to show all assigned image tags
		QStringList imageTags;

		for (const auto& imageTag : draftItem->imageTagsAllotment()->object().allottedTags())
			imageTags << imageTag->getName();

		setToolTip(QString{"<b>Layer:</b> %2<br><b>Tags:</b> %1"}.arg(!imageTags.isEmpty() ? imageTags.join(", ") : "None").arg(draftItem->layer()->getName()));
	}

	LayerItemNode::updateNode();

	// Also update the in-place editor
	if (_inPlaceEditor)
		_inPlaceEditor->updateEditor();
}

void DraftItemNode::updateZOrder()
{
	if (isSelected())	// If selected, show on top of all other items (makes in-place editing easier)
	{
		setZValue(std::numeric_limits<qreal>::max());
	}
	else
	{
		if (auto draftItem = _draftItem.lock())	// Make sure that the underlying draft item still exists
			setZValue(draftItem->getZOrder());	// Use the z-order of the draft item
	}
}

QVariant DraftItemNode::itemChange(QGraphicsItem::GraphicsItemChange change, const QVariant& value)
{
	if (change == QGraphicsItem::ItemPositionHasChanged)
	{
		if (auto draftItem = _draftItem.lock())	// Make sure that the underlying draft item still exists
		{						
			// Update the draft item's position accordingly
			auto newPos = value.toPointF();
			draftItem->position()->setValue(newPos.toPoint());
		}
	}
	else if (change == QGraphicsItem::ItemSelectedHasChanged)
	{
		// Show the in-place editor if the item is currently selected
		if (_inPlaceEditor)
			_inPlaceEditor->setVisible(value.toBool());

		updateZOrder();
	}

	return ImageEditorNode::itemChange(change, value);
}

void DraftItemNode::dragEnterEvent(QGraphicsSceneDragDropEvent* event)
{
	if (event->mimeData()->hasFormat(IMAGETAGSLIST_DRAG_MIMETYPE))
		event->acceptProposedAction();
	else
		event->ignore();
}

void DraftItemNode::dragMoveEvent(QGraphicsSceneDragDropEvent* event)
{
	if (event->mimeData()->hasFormat(IMAGETAGSLIST_DRAG_MIMETYPE))
		event->acceptProposedAction();
	else
		event->ignore();
}

void DraftItemNode::dropEvent(QGraphicsSceneDragDropEvent* event)
{
	if (event->mimeData()->hasFormat(IMAGETAGSLIST_DRAG_MIMETYPE))
	{
		if (auto draftItem = _draftItem.lock())	// Make sure that the underlying draft item still exists
		{
			// Assign the dropped image tags to the image tags allotment of this draft item
			QString imageTags = event->mimeData()->data(IMAGETAGSLIST_DRAG_MIMETYPE).toStdString().data();
			_imageEditor->controller().assignImageTags(draftItem->imageTagsAllotment(), imageTags.split("\n"));

			// Select this node after assigning the tags
			if (_imageEditor->controller().activeScene())
				_imageEditor->controller().activeScene()->clearSelection();

			setSelected(true);

			event->acceptProposedAction();
		}
	}
	else
		event->ignore();
}

void DraftItemNode::updateGeometry()
{
	_nodeRect = getBoundingRect();
	_nodeRectSelected = _nodeRect + QMarginsF{_draftItemStyle.selectionMargin, _draftItemStyle.selectionMargin, _draftItemStyle.selectionMargin, _draftItemStyle.selectionMargin};

	ImageEditorNode::updateGeometry();
}

void DraftItemNode::deleteNode()
{
	_imageEditor->controller().removeSelectedNodes();
}
