/******************************************************************************
 * File: ImageEditorManager.cpp
 * Date: 13.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageEditorManager.h"
#include "ImageEditorWidget.h"
#include "core/GrinderApplication.h"
#include "controller/ProjectController.h"
#include "image/ImageExceptions.h"
#include "pipeline/Pipeline.h"

const char* ImageEditorManager::Serialization_Group = "ImageEditors";
const char* ImageEditorManager::Serialization_Element = "ImageEditor";

ImageEditorManager::ImageEditorManager(Project* project, ProjectController* projectController, PipelineManager* pipelineManager) : QObject(pipelineManager)
{
	// We need to close editors when an image reference or an entire pipeline is removed
	connect(pipelineManager, &PipelineManager::pipelineRemoved, this, &ImageEditorManager::pipelineRemoved);
	connect(project, &Project::imageReferenceRemoved, this, &ImageEditorManager::imageReferenceRemoved);

	// We need to store the settings of all open editors when saving the current project
	connect(projectController, &ProjectController::projectSaving, this, &ImageEditorManager::projectSaving);
}

void ImageEditorManager::showEditor(const Block* block, const std::shared_ptr<grndr::ImageBuild>& imageBuild)
{
	if (!block)
		throw std::invalid_argument{_EXCPT("block may not be null")};

	if (!imageBuild)
		throw std::invalid_argument{_EXCPT("imageBuild may not be null")};

	auto editor = findEditor(block);

	// Create an image editor if necessary
	if (!editor)
	{
		if (!(editor = createEditor(block)))
			throw ImageEditorException{nullptr, _EXCPT(QString{"Unable to create an image editor for block '%1'"}.arg(block->getName()))};
	}

	editor->setImageBuild(imageBuild);

	// Let whoever is interested in the newly created editor handle showing it
	emit editorShown(editor);
}

void ImageEditorManager::closeEditor(const Block* block)
{
	// If an image editor exists for the block, close and remove it
	if (auto editor = findEditor(block))
	{
		editor->dockWidget()->close();
		removeEditor(block);
	}
}

void ImageEditorManager::serialize(SerializationContext& ctx) const
{
	// Serialize all editor settings
	for (const auto& editorSettings : _editorSettings)
	{
		// Before storing the editor settings, set the index of the block they belong to (needed for deserialization)
		(*editorSettings.second)(ImageEditor::Serialization_Value_Block) = ctx.getBlockIndex(editorSettings.first);

		ctx.settings() << *editorSettings.second;
	}
}

void ImageEditorManager::deserialize(DeserializationContext& ctx)
{
	// Deserialize all editor settings
	for (const auto elemSettings : ctx.settings().children(Serialization_Element))
	{
		int blockIndex = (*elemSettings)(ImageEditor::Serialization_Value_Block, -1).toInt();

		if (auto block = ctx.getBlock(blockIndex))
			_editorSettings[block] = std::make_unique<SettingsContainer>(*elemSettings);
	}
}

ImageEditor* ImageEditorManager::createEditor(const Block* block)
{
	auto editor = std::make_unique<ImageEditor>();

	// When a dock is closed, remove its associated editor
	connect(editor->dockWidget(), &ImageEditorDockWidget::dockClosed, this, &ImageEditorManager::dockClosed);

	_editors.emplace(block, std::move(editor));

	// If a block is removed, close any open associated editors
	connect(block->pipeline(), &Pipeline::blockRemoved, this, &ImageEditorManager::blockRemoved);

	// Reflect pipeline and block renamings
	connect(block, &PipelineItem::itemRenamed, this, &ImageEditorManager::updateDockNames);
	connect(block->pipeline(), &Pipeline::pipelineRenamed, this, &ImageEditorManager::updateDockNames);

	// Load any saved editor settings
	loadEditorSettings(block, _editors[block].get());

	return _editors[block].get();
}

void ImageEditorManager::removeEditor(const Block* block)
{
	if (auto editor = findEditor(block))
		emit editorClosed(editor);

	// Disconnect any signals from the block
	disconnect(block, nullptr, this, nullptr);

	_editors.erase(block);

	// Disconnect any signals from the block's pipeline if no further blocks of the same pipeline are associated with an editor
	if (std::none_of(_editors.cbegin(), _editors.cend(), [block](const auto& editor) { return editor.first->pipeline() == block->pipeline(); }))
		disconnect(block->pipeline(), nullptr, this, nullptr);
}

ImageEditor* ImageEditorManager::findEditor(const Block* block) const
{
	auto it = _editors.find(block);

	if (it != _editors.end())
		return it->second.get();
	else
		return nullptr;
}

SettingsContainer* ImageEditorManager::findEditorSettings(const Block* block) const
{
	auto it = _editorSettings.find(block);

	if (it != _editorSettings.end())
		return it->second.get();
	else
		return nullptr;
}

void ImageEditorManager::loadEditorSettings(const Block* block, ImageEditor* editor) const
{
	if (auto settings = findEditorSettings(block))
	{
		DeserializationContext ctx{*settings, DeserializationContext::Mode::GenericSerialization};
		editor->deserialize(ctx);
	}
}

void ImageEditorManager::saveEditorSettings(const Block* block, const ImageEditor* editor)
{
	SerializationContext ctx{Serialization_Element, SerializationContext::Mode::GenericSerialization};
	editor->serialize(ctx);

	auto settings = std::make_unique<SettingsContainer>(std::move(ctx.settings()));
	_editorSettings[block] = std::move(settings);
}

void ImageEditorManager::dockClosed(ImageEditorDockWidget* dock)
{
	// Find the editor associated with the dock and remove it
	for (const auto& editor : _editors)
	{
		if (editor.second->dockWidget() == dock)
		{
			// Save all editor settings before closing
			saveEditorSettings(editor.first, editor.second.get());

			removeEditor(editor.first);
			break;
		}
	}
}

void ImageEditorManager::pipelineRemoved(const std::shared_ptr<Pipeline>& pipeline)
{
	// Close any editors associated with one of the pipeline's blocks
	for (const auto& block : pipeline->blocks())
		closeEditor(block.get());
}

void ImageEditorManager::blockRemoved(const std::shared_ptr<Block>& block)
{
	closeEditor(block.get());

	// Remove any saved settings for the deleted block
	_editorSettings.erase(block.get());
}

void ImageEditorManager::imageReferenceRemoved(const std::shared_ptr<ImageReference>& imageReference)
{
	// Remove all editors that are currently showing the removed image reference
	std::vector<const Block*> editorsToRemove;

	for (const auto& editor : _editors)
	{
		auto imageReferences = editor.second->controller().activeImageBuild()->imageReferences();

		if (std::find(imageReferences.cbegin(), imageReferences.cend(), imageReference.get()) != imageReferences.cend())
			editorsToRemove.push_back(editor.first);
	}

	for (const auto& editorBlock : editorsToRemove)
		closeEditor(editorBlock);
}

void ImageEditorManager::projectSaving(QString projectFile)
{
	Q_UNUSED(projectFile);

	// Save the settings of all opened editors when saving the current project
	for (const auto& editor : _editors)
		saveEditorSettings(editor.first, editor.second.get());
}

void ImageEditorManager::updateDockNames()
{
	// Update all dock names
	for (const auto& editor : _editors)
		editor.second->dockWidget()->updateDockName(editor.first);
}
