/******************************************************************************
 * File: ImageEditorPropertyWidget.cpp
 * Date: 23.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageEditorPropertyWidget.h"
#include "ImageEditorWidget.h"
#include "ImageEditorScene.h"
#include "core/GrinderApplication.h"
#include "controller/ImageEditorController.h"
#include "util/UIUtils.h"

ImageEditorPropertyWidget::ImageEditorPropertyWidget(QWidget* parent) : QWidget(parent),
	_layout{new QGridLayout{this}}
{
	_layout->setContentsMargins(6, 6, 6, 6);
	_layout->setHorizontalSpacing(6);
	_layout->setVerticalSpacing(4);
	setLayout(_layout);
}

void ImageEditorPropertyWidget::assignUiComponents(ImageEditor* imageEditor)
{
	if (!imageEditor)
		throw std::invalid_argument{_EXCPT("imageEditor may not be null")};

	_imageEditor = imageEditor;

	// Listen for various events in order to update the property list accordingly
	connect(&_imageEditor->editorTools(), &ImageEditorToolList::activeToolChanged, this, &ImageEditorPropertyWidget::imageEditorToolChanged);
	connect(&_imageEditor->controller(), &ImageEditorController::imageBuildSwitching, this, &ImageEditorPropertyWidget::imageBuildSwitching);
	connect(&_imageEditor->controller(), &ImageEditorController::imageBuildSwitched, this, &ImageEditorPropertyWidget::imageBuildSwitched);
}

void ImageEditorPropertyWidget::clear()
{
	// Remove all properties from the layout
	UIUtils::removeChildrenFromLayout(_layout);
}

void ImageEditorPropertyWidget::addProperties(const PropertyVector* properties)
{
	clear();

	bool propertiesEmpty = true;

	if (properties)
	{
		for (auto& property : properties->sorted())
		{
			if (!property->hasFlag(PropertyBase::Flag::ReadOnly) && !property->hasFlag(PropertyBase::Flag::Hidden))
			{
				addProperty(property);
				propertiesEmpty = false;
			}
		}
	}

	if (propertiesEmpty)
	{
		auto label = new QLabel{"No properties to display"};
		label->setEnabled(false);
		_layout->addWidget(label, 0, 0, 1, 2, Qt::AlignCenter);
	}

	// Add a spacer item to align all properties
	_layout->addItem(new QSpacerItem{0, 0, QSizePolicy::Expanding, QSizePolicy::Expanding}, _layout->rowCount(), 1);
}

void ImageEditorPropertyWidget::addProperty(const std::shared_ptr<PropertyBase>& property)
{
	auto editor = property->createEditor(nullptr);
	auto label = new QLabel{property->getName() + ":"};
	label->setFont(GrinderApplication::boldFont(label));

	if (!editor)
		editor = new QLabel{"No editor available"};

	editor->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Preferred);

	int row = _layout->rowCount();
	_layout->addWidget(label, row, 0);
	_layout->addWidget(editor, row, 1);
}

void ImageEditorPropertyWidget::populateProperties()
{
	const PropertyVector* properties = &_imageEditor->editorTools().activeTool()->properties();	// Use the active tool's properties by default

	// See if there are any items selected in the editor
	if (auto scene = _imageEditor->controller().activeScene())
	{
		auto selectedItems = scene->getNodes<DraftItemNode>(true);

		if (!selectedItems.empty())
		{
			// TODO: Multi-selection
			if (selectedItems.size() == 1)
			{
				if (auto draftItem = selectedItems.front()->draftItem().lock())	// Make sure that the underlying draft item still exists
					properties = &draftItem->properties();
			}
		}
	}

	addProperties(properties);
}

void ImageEditorPropertyWidget::imageEditorToolChanged(const ImageEditorTool* tool)
{
	Q_UNUSED(tool);
	populateProperties();
}

void ImageEditorPropertyWidget::imageBuildSwitching(ImageBuild* imageBuild)
{
	Q_UNUSED(imageBuild);

	clear();

	if (auto scene = _imageEditor->controller().activeScene())
		disconnect(scene, &ImageEditorScene::selectionChanged, this, &ImageEditorPropertyWidget::populateProperties);
}

void ImageEditorPropertyWidget::imageBuildSwitched(ImageBuild* imageBuild)
{
	Q_UNUSED(imageBuild);

	// Listen for selection changes in the current editor scene
	if (auto scene = _imageEditor->controller().activeScene())
		connect(scene, &ImageEditorScene::selectionChanged, this, &ImageEditorPropertyWidget::populateProperties);

	populateProperties();
}
