/******************************************************************************
 * File: ImageEditorScene.cpp
 * Date: 15.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageEditorScene.h"
#include "ImageEditor.h"
#include "DraftItemNode.h"
#include "LayerBackgroundNode.h"
#include "LayerPixelsNode.h"
#include "image/ImageExceptions.h"
#include "controller/ImageEditorController.h"
#include "util/CVUtils.h"
#include "res/Resources.h"

ImageEditorScene::ImageEditorScene(ImageEditor* imageEditor, ImageBuild* imageBuild, ImageEditorView* view) : VisualScene(view), ImageEditorComponent(imageEditor),
	_imageBuild{imageBuild}, _nodeFactory{this}
{
	if (!imageEditor)
		throw std::invalid_argument{_EXCPT("imageEditor may not be null")};

	if (!imageBuild)
		throw std::invalid_argument{_EXCPT("imageBuild may not be null")};

	_drawGrid = false;
}

void ImageEditorScene::buildScene()
{
	clear();

	createBackgroundNode();

	// Add all items of each layer to the scene
	if (_imageBuild)
	{
		for (const auto& layer : _imageBuild->layers())
		{
			for (const auto& item : layer->draftItems())
				createDraftItemNode(item);

			createLayerBackgroundNode(layer);
			createLayerPixelsNode(layer);
		}
	}
}

void ImageEditorScene::refreshImage(bool centerOnImage)
{
	// Try to get an image from the image build's data
	auto image = CVUtils::matrixToPixmap(_imageBuild->imageData());

	if (!image.isNull())
	{
		_imageNode->setPixmap(image);

		// Calculate the new scene rectangle -> add a certain border around the image which is always positioned at (0, 0)
		QRectF sceneRect{QPointF{0.0, 0.0}, image.size()};
		sceneRect += _view->sceneStyle().getEditorStyle().editorMargins;
		setSceneRect(sceneRect);

		if (centerOnImage)
			_view->centerOn(_imageNode);
	}
	else
	{
		_imageNode->setPixmap(QPixmap{FILE_ICON_EDITOR_NOIMAGE});	// If the image is invalid, show a question mark indicating an error
		throw ImageBuildException{_imageBuild, _EXCPT("Invalid or incompatible image data")};
	}

	update();
}

void ImageEditorScene::fitViewToImage()
{
	_view->fitToWindow(_imageNode);
}

void ImageEditorScene::createDraftItemNode(const std::shared_ptr<DraftItem>& item)
{
	if (!item)
		throw std::invalid_argument{_EXCPT("item may not be null")};

	if (!findDraftItemNode(item.get()))
	{
		auto node = _nodeFactory.createNode(item);

		try {	// Propagate initialization errors to the caller
			node->initDraftItemNode();
		}
		catch (...) {
			throw;
		}

		addItem(node);
	}
}

void ImageEditorScene::removeDraftItemNode(const std::shared_ptr<DraftItem>& item)
{
	removeNode(findDraftItemNode(item.get()));
}

void ImageEditorScene::createLayerBackgroundNode(const std::shared_ptr<Layer>& layer)
{
	if (!layer)
		throw std::invalid_argument{_EXCPT("layer may not be null")};

	if (auto backgroundImage = layer->backgroundImage())
	{
		if (!findLayerBackgroundNode(backgroundImage))
			addItem(new LayerBackgroundNode{this, layer.get()});
	}
}

void ImageEditorScene::removeLayerBackgroundNode(const std::shared_ptr<Layer>& layer)
{
	if (auto backgroundImage = layer->backgroundImage())
		removeNode(findLayerBackgroundNode(backgroundImage));
}

void ImageEditorScene::createLayerPixelsNode(const std::shared_ptr<Layer>& layer)
{
	if (!layer)
		throw std::invalid_argument{_EXCPT("layer may not be null")};

	if (!findLayerPixelsNode(&layer->layerPixels()))
		addItem(new LayerPixelsNode{this, layer.get()});
}

void ImageEditorScene::removeLayerPixelsNode(const std::shared_ptr<Layer>& layer)
{
	removeNode(findLayerPixelsNode(&layer->layerPixels()));
}

void ImageEditorScene::updateLayerItemsOrder()
{
	for (auto& layerItem : findLayerItemNodes())
		layerItem->updateZOrder();
}

void ImageEditorScene::updateLayerItemsVisibility(const Layer* layer)
{
	for (auto& layerItem : findLayerItemNodes(layer))
		layerItem->updateVisibility();
}

void ImageEditorScene::updateLayerItemsAlpha(const Layer* layer)
{
	for (auto& layerItem : findLayerItemNodes(layer))
		layerItem->updateAlpha();
}

void ImageEditorScene::mousePressEvent(QGraphicsSceneMouseEvent* mouseEvent)
{
	if (handleMouseEvent(mouseEvent, &ImageEditorTool::handleMousePressEvent) != ImageEditorTool::InputEventResult::Process)
		VisualScene::mousePressEvent(mouseEvent);	// The active tool didn't process the event, so call the default handler
}

void ImageEditorScene::mouseMoveEvent(QGraphicsSceneMouseEvent* mouseEvent)
{
	if (handleMouseEvent(mouseEvent, &ImageEditorTool::handleMouseMoveEvent) != ImageEditorTool::InputEventResult::Process)
		VisualScene::mouseMoveEvent(mouseEvent);	// The active tool didn't process the event, so call the default handler
}

void ImageEditorScene::mouseReleaseEvent(QGraphicsSceneMouseEvent* mouseEvent)
{
	if (handleMouseEvent(mouseEvent, &ImageEditorTool::handleMouseReleaseEvent) != ImageEditorTool::InputEventResult::Process)
		VisualScene::mouseReleaseEvent(mouseEvent);	// The active tool didn't process the event, so call the default handler
}

void ImageEditorScene::keyPressEvent(QKeyEvent* keyEvent)
{
	if (handleKeyEvent(keyEvent, &ImageEditorTool::handleKeyPressEvent) != ImageEditorTool::InputEventResult::Process)
		VisualScene::keyPressEvent(keyEvent);	// The active tool didn't process the event, so call the default handler
}

void ImageEditorScene::keyReleaseEvent(QKeyEvent* keyEvent)
{
	if (handleKeyEvent(keyEvent, &ImageEditorTool::handleKeyReleaseEvent) != ImageEditorTool::InputEventResult::Process)
		VisualScene::keyReleaseEvent(keyEvent);	// The active tool didn't process the event, so call the default handler
}

void ImageEditorScene::createBackgroundNode()
{
	// Set up the background node
	_imageNode = new QGraphicsPixmapItem{};
	_imageNode->setPos(QPointF{0.0, 0.0});
	_imageNode->setZValue(-1.0);	// Always the lowest item
	_imageNode->setShapeMode(QGraphicsPixmapItem::BoundingRectShape);
	_imageNode->setTransformationMode(Qt::SmoothTransformation);

	addItem(_imageNode);
}

ImageEditorTool::InputEventResult ImageEditorScene::handleMouseEvent(QGraphicsSceneMouseEvent* mouseEvent, std::function<ImageEditorTool::InputEventResult(ImageEditorTool*, QGraphicsSceneMouseEvent*)> handler)
{
	mouseEvent->ignore();

	// Let the active tool handle the mouse event
	if (auto activeTool = _imageEditor->editorTools().activeTool())
		return handler(activeTool, mouseEvent);

	return ImageEditorTool::InputEventResult::Ignore;
}

VisualSceneInputHandler::InputEventResult ImageEditorScene::handleKeyEvent(QKeyEvent* keyEvent, std::function<VisualSceneInputHandler::InputEventResult (ImageEditorTool*, QKeyEvent*)> handler)
{
	keyEvent->ignore();

	// Let the active tool handle the key event
	if (auto activeTool = _imageEditor->editorTools().activeTool())
		return handler(activeTool, keyEvent);

	return ImageEditorTool::InputEventResult::Ignore;
}

DraftItemNode* ImageEditorScene::_findDraftItemNode(const DraftItem* item) const
{
	for (const auto& node : items())
	{
		if (auto itemNode = dynamic_cast<DraftItemNode*>(node))
		{
			if (auto draftItem = itemNode->draftItem().lock())
			{
				if (draftItem.get() == item)
					return itemNode;
			}
		}
	}

	return nullptr;
}

LayerBackgroundNode* ImageEditorScene::_findLayerBackgroundNode(const ImageReference* imageRef) const
{
	for (const auto& node : items())
	{
		if (auto itemNode = dynamic_cast<LayerBackgroundNode*>(node))
		{
			if (itemNode->layer()->backgroundImage() == imageRef)
				return itemNode;
		}
	}

	return nullptr;
}

LayerPixelsNode* ImageEditorScene::_findLayerPixelsNode(const LayerPixels* layerPixels) const
{
	for (const auto& node : items())
	{
		if (auto itemNode = dynamic_cast<LayerPixelsNode*>(node))
		{
			if (&itemNode->layer()->layerPixels() == layerPixels)
				return itemNode;
		}
	}

	return nullptr;
}
