/******************************************************************************
 * File: ImageEditorView.cpp
 * Date: 15.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageEditorView.h"
#include "ImageEditorScene.h"
#include "ImageEditor.h"
#include "core/GrinderApplication.h"
#include "controller/ImageEditorController.h"
#include "util/UIUtils.h"
#include "res/Resources.h"

ImageEditorView::ImageEditorView(QWidget* parent) : VisualSceneView(parent)
{
	// Set the background of the view to a checkerboard pattern
	setBackgroundBrush(QPixmap{FILE_ICON_EDITOR_BACKGROUND});

	// Set widget border
	setStyleSheet(QString{"QGraphicsView { border: 1px solid %1; }"}.arg(QPalette{}.color(QPalette::Dark).name()));

	// Create view actions
	_showDirectionsAction = UIUtils::createAction(this, "&Show direction arrows", FILE_ICON_EDITOR_SHOWDIRECTIONS, SLOT(showDirectionArrows()), "Show direction arrows on items");
	_showDirectionsAction->setCheckable(true);
	_showDirectionsAction->setChecked(true);
	_showTagsAction = UIUtils::createAction(this, "&Show tags", FILE_ICON_EDITOR_SHOWTAGS, SLOT(showTags()), "Show tags on items");
	_showTagsAction->setCheckable(true);
	_showTagsAction->setChecked(true);	
	_zoomFitAction = UIUtils::createAction(this, "Zoom to &window", FILE_ICON_ZOOMFIT, SLOT(fitImageToWindow()), "Zoom the view to fit the image to its window", "Ctrl+Alt+A");

	_copyDraftItems = UIUtils::createAction(this, "&Copy item(s)", FILE_ICON_COPY, SLOT(copyDraftItems()), "Copy the selected items to the clipboard");
	_copyDraftItems->setShortcut(QKeySequence{QKeySequence::Copy});
	_pasteDraftItems = UIUtils::createAction(this, "&Paste item(s)", FILE_ICON_PASTE, SLOT(pasteDraftItems()), "Paste items from the clipboard");
	_pasteDraftItems->setShortcut(QKeySequence{QKeySequence::Paste});
	_cutDraftItems = UIUtils::createAction(this, "Cu&t item(s)", FILE_ICON_CUT, SLOT(cutDraftItems()), "Copy the selected items to the clipboard and remove them afterwards");
	_cutDraftItems->setShortcut(QKeySequence{QKeySequence::Cut});
	_convertDraftItemsAction = UIUtils::createAction(this, "&Convert to pixels", FILE_ICON_EDITOR_CONVERTTOPIXELS, SLOT(convertDraftItems()), "Converts the selected items to pixels", "Return");

	// Listen for clipboard changes to update our actions
	connect(&grinder()->clipboardManager(), &ClipboardManager::dataChanged, this, &ImageEditorView::updateActions);

	updateActions();
}

void ImageEditorView::setEditorScene(ImageEditorScene* scene)
{
	if (_editorScene)
	{
		disconnect(&_editorScene->imageEditor()->editorTools(), nullptr, this, nullptr);
		disconnect(&_editorScene->imageEditor()->environment(), nullptr, this, nullptr);
	}

	_editorScene = scene;
	VisualSceneView::setScene(scene);

	if (_editorScene)
	{
		// Update the view's cursor when the active image editor tool has changed
		connect(&_editorScene->imageEditor()->editorTools(), &ImageEditorToolList::activeToolChanged, this, &ImageEditorView::imageEditorToolChanged);

		// Update the view's actions when the image editor environment has changed
		connect(&_editorScene->imageEditor()->environment(), &ImageEditorEnvironment::showDirectionArrowsChanged, this, &ImageEditorView::showDirectionArrowsChanged);
		connect(&_editorScene->imageEditor()->environment(), &ImageEditorEnvironment::showTagsChanged, this, &ImageEditorView::showTagsChanged);
	}
}

void ImageEditorView::removeSelectedItems() const
{
	if (_editorScene)
		_editorScene->imageEditor()->controller().removeSelectedNodes();
}

void ImageEditorView::prepareNodeContextMenu(QMenu& menu) const
{
	QAction* firstAction = !menu.actions().isEmpty() ? menu.actions().first() : nullptr;

	menu.insertAction(firstAction, _cutDraftItems);
	menu.insertAction(firstAction, _copyDraftItems);
	menu.insertAction(firstAction, _pasteDraftItems);
	menu.insertSeparator(firstAction);
	menu.insertAction(firstAction, _convertDraftItemsAction);
}

void ImageEditorView::prepareNodesContextMenu(QMenu& menu) const
{
	prepareNodeContextMenu(menu);
}

std::vector<QAction*> ImageEditorView::getActions(AddActionsMode mode) const
{
	std::vector<QAction*> actions;

	if (mode != AddActionsMode::Toolbar)
	{
		actions.push_back(_cutDraftItems);
		actions.push_back(_copyDraftItems);
		actions.push_back(_pasteDraftItems);
		actions.push_back(nullptr);

		actions.push_back(_selectAllAction);
		actions.push_back(_deleteSelectedAction);
		actions.push_back(nullptr);
	}

	actions.push_back(_showDirectionsAction);
	actions.push_back(_showTagsAction);
	actions.push_back(nullptr);
	actions.push_back(_zoomInAction);
	actions.push_back(_zoomOutAction);
	actions.push_back(_zoomFullAction);
	actions.push_back(_zoomFitAction);

	return actions;
}
void ImageEditorView::drawBackground(QPainter* painter, const QRectF& rect)
{
	// Draw scale-invariant background
	auto scaleFactor = painter->transform().m11();
	auto bgRect = rect;

	bgRect.setTopLeft(bgRect.topLeft() * scaleFactor);
	bgRect.setBottomRight(bgRect.bottomRight() * scaleFactor);

	painter->scale(1.0 / scaleFactor, 1.0 / scaleFactor);
	painter->fillRect(bgRect, backgroundBrush());
	painter->scale(scaleFactor, scaleFactor);
}

void ImageEditorView::updateActions()
{
	VisualSceneView::updateActions();

	_zoomFitAction->setEnabled(_scene);

	_copyDraftItems->setEnabled(_scene && !_scene->selectedItems().isEmpty());
	_pasteDraftItems->setEnabled(_scene && grinder()->clipboardManager().hasData(DraftItemVector::Serialization_Element));
	_cutDraftItems->setEnabled(_scene && !_scene->selectedItems().isEmpty());
	_convertDraftItemsAction->setEnabled(_scene && !_scene->selectedItems().isEmpty());
}

void ImageEditorView::showDirectionArrows()
{
	if (_editorScene)
		_editorScene->imageEditor()->environment().setShowDirectionArrows(_showDirectionsAction->isChecked());
}

void ImageEditorView::showTags()
{
	if (_editorScene)
		_editorScene->imageEditor()->environment().setShowTags(_showTagsAction->isChecked());
}

void ImageEditorView::fitImageToWindow()
{
	if (_editorScene)
		_editorScene->fitViewToImage();
}

void ImageEditorView::imageEditorToolChanged(const ImageEditorTool* tool)
{
	setCursor(tool->getCursor());

	if (_editorScene)
		_editorScene->clearSelection();
}

void ImageEditorView::showDirectionArrowsChanged(bool show)
{
	_showDirectionsAction->setChecked(show);
	update();
}

void ImageEditorView::showTagsChanged(bool show)
{
	_showTagsAction->setChecked(show);
	update();
}

void ImageEditorView::copyDraftItems() const
{
	if (_editorScene)
		_editorScene->imageEditor()->controller().copySelectedNodes();
}

void ImageEditorView::pasteDraftItems() const
{
	if (_editorScene)
		_editorScene->imageEditor()->controller().pasteSelectedNodes();
}

void ImageEditorView::cutDraftItems() const
{
	if (_editorScene)
		_editorScene->imageEditor()->controller().cutSelectedNodes();
}

void ImageEditorView::convertDraftItems() const
{
	if (_editorScene)
		_editorScene->imageEditor()->controller().convertSelectedNodesToPixels();
}
