/******************************************************************************
 * File: ImageEditorWidget.cpp
 * Date: 13.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageEditorWidget.h"
#include "ui_ImageEditorWidget.h"
#include "tools/DefaultImageEditorTool.h"
#include "core/GrinderApplication.h"
#include "image/ImageBuild.h"
#include "image/ImageExceptions.h"
#include "image/ImageTags.h"
#include "ui/mainwnd/ImageReferencesListItem.h"
#include "ui/properties/editors/ImageTagsAllotmentPropertyEditor.h"
#include "ui/dlg/CheckListDialog.h"
#include "ui/StyleSheet.h"
#include "util/UIUtils.h"
#include "res/Resources.h"

Q_DECLARE_METATYPE(std::shared_ptr<ImageReference>)

const char* ImageEditorWidget::Serialization_Group = "Editor";

const char* ImageEditorWidget::Serialization_Value_ColorPresets = "ColorPresets";

ImageEditorWidget::ImageEditorWidget(ImageEditor* imageEditor, QWidget* parent) : QWidget(parent), ImageEditorComponent(imageEditor),
	ui{new Ui::ImageEditorWidget}
{
	if (!imageEditor)
		throw std::runtime_error{_EXCPT("imageEditor may not be null")};	

	// Needed to be able to queue imageReferenceRemoved signals
	qRegisterMetaType<std::shared_ptr<ImageReference>>();

	// Create editor actions
	_copyImageBuild = UIUtils::createAction(this, "&Copy image build", FILE_ICON_COPY, SLOT(copyImageBuild()), "Copy the current image build to the clipboard (Ctrl+Shift+C)", "Ctrl+Shift+C", Qt::WidgetWithChildrenShortcut);
	_pasteImageBuild = UIUtils::createAction(this, "&Paste image build", FILE_ICON_PASTE, SLOT(pasteImageBuild()), "Paste an image build from the clipboard (Ctrl+Shift+V)", "Ctrl+Shift+V", Qt::WidgetWithChildrenShortcut);
	_duplicateImageBuild = UIUtils::createAction(this, "&Copy image build to next image", FILE_ICON_EDITOR_COPYFROMPREVIOUS, SLOT(duplicateImageBuild()), "Copy the current image build to the next image (Ctrl+D)", "Ctrl+D", Qt::WidgetWithChildrenShortcut);
	_duplicateImageBuildTo = UIUtils::createAction(this, "Copy image build to &multiple images...", "", SLOT(duplicateImageBuildTo()), "Copy the current image build to multiple images (Ctrl+Shift+D)", "Ctrl+Shift+D", Qt::WidgetWithChildrenShortcut);

	_editImageBuildTags = UIUtils::createAction(this, "Edit the image build's tags...", FILE_ICON_TAGS, SLOT(editImageBuildTags()), "Edit the tags assigned to the current image build (Ctrl+Shift+T)", "Ctrl+Shift+T", Qt::WidgetWithChildrenShortcut);

	// Set up the duplicate image build menu
	_duplicateImageBuildMenu.addAction(_duplicateImageBuild);
	_duplicateImageBuildMenu.addSeparator();
	_duplicateImageBuildMenu.addAction(_duplicateImageBuildTo);
	_duplicateImageBuildMenu.setDefaultAction(_duplicateImageBuild);

	setupUi();	

	// Reflect primary color changes
	connect(&_imageEditor->environment(), &ImageEditorEnvironment::primaryColorChanged, this, &ImageEditorWidget::primaryColorChanged);

	// Listen for various events to update our actions
	connect(&grinder()->project(), SIGNAL(imageReferenceCreated(const std::shared_ptr<ImageReference>&)), this, SLOT(updateActions()));
	connect(&grinder()->project(), SIGNAL(imageReferenceRemoved(const std::shared_ptr<ImageReference>&)), this, SLOT(updateActions()), Qt::QueuedConnection);	// Delay this signal so that the image reference has been removed from the image references list
	connect(&grinder()->clipboardManager(), &ClipboardManager::dataChanged, this, &ImageEditorWidget::updateActions);
}

ImageEditorWidget::~ImageEditorWidget()
{
	delete ui;
}

void ImageEditorWidget::setImageBuild(const std::shared_ptr<ImageBuild>& imageBuild)
{
	if (auto scene = _imageEditor->controller().activeScene())
		disconnect(scene, nullptr, this, nullptr);

	if (auto imageBuild = _imageEditor->controller().activeImageBuild())
		disconnect(imageBuild, nullptr, this, nullptr);

	// Let the controller handle showing the new image build
	_imageEditor->controller().showImageBuild(imageBuild.get());

	// Listen for node selection changes in the new scene to update the primary color
	if (auto scene = _imageEditor->controller().activeScene())
		connect(scene, &ImageEditorScene::selectionChanged, this, &ImageEditorWidget::updatePrimaryColor);

	// Listen for various events to update our actions
	connect(imageBuild.get(), SIGNAL(layerCreated(const std::shared_ptr<Layer>&)), this, SLOT(updateActions()));
	connect(imageBuild.get(), SIGNAL(layerRemoved(const std::shared_ptr<Layer>&)), this, SLOT(updateActions()));
	connect(imageBuild.get(), SIGNAL(inputImageTagsChanged(const ImageTags*)), this, SLOT(updateActions()));

	updateActions();
}

void ImageEditorWidget::serialize(SerializationContext& ctx) const
{
	// Serialize color presets
	QStringList colorPresets;

	for (unsigned int i = 0; i < ui->colorPresetsWidget->getPresetsCount(); ++i)
		colorPresets << ui->colorPresetsWidget->getColor(i).name();

	ctx.settings()(Serialization_Value_ColorPresets) = colorPresets.join(";");
}

void ImageEditorWidget::deserialize(DeserializationContext& ctx)
{
	// Deserialize color presets
	QStringList colorPresets = ctx.settings()(Serialization_Value_ColorPresets).toString().split(";");

	for (int i = 0; i < colorPresets.size(); ++i)
		ui->colorPresetsWidget->setColor(i, QColor{colorPresets[i]});
}

void ImageEditorWidget::keyPressEvent(QKeyEvent* event)
{
	QWidget::keyPressEvent(event);

	// Handle the Esc key if it hasn't been consumed
	if (!event->isAccepted())
	{
		if (event->key() == Qt::Key_Escape)
		{
			if (_imageEditor->editorTools().activeTool()->getToolType() != DefaultImageEditorTool::tool_type)
			{
				_imageEditor->editorTools().activateDefaultTool();
				return;
			}
		}
	}
}

void ImageEditorWidget::setupUi()
{
	ui->setupUi(this);

	connect(ui->primaryColorWidget, &ColorWidget::colorChanged, this, &ImageEditorWidget::primaryColorSelected);
	connect(ui->colorPresetsWidget, &ColorPresetsWidget::colorSelected, this, &ImageEditorWidget::colorPresetSelected);

	// Create editor-global actions
	_newLayerAction = UIUtils::createAction(this, "&New layer", FILE_ICON_ADD, SLOT(newLayer()), "Add a new layer", "Ctrl+Shift+L", Qt::WidgetWithChildrenShortcut, ui->layersList);	

	// Assign the UI components to the controller
	_imageEditor->controller().assignUiComponents(ui->imageScene, ui->layersList);

	// Assign the UI components to other UI components
	ui->colorPresetsWidget->assignUiComponents(this);
	ui->imageEditorProperties->assignUiComponents(_imageEditor);
	ui->layersList->assignUiComponents(_imageEditor);
	ui->imageTagsList->assignUiComponents(_imageEditor);
	_imageEditor->editorTools().assignUiComponents(this);

	setupImageControlBar();

	// Set up the various UI components
	ui->layersList->setupUi(_newLayerAction, nullptr, ui->layersControlBar);

	ui->propertiesArea->setStyleSheet(StyleSheet::loadStyleSheet(FILE_STYLESHEET_IMAGEEDITORPROPERTIES));

	updateSceneZoomLevel(ui->imageScene->getZoom());
	updatePrimaryColorLabel();

	grinder()->settings().getSplitterState("ImageEditorMid", ui->splitterMid);
	grinder()->settings().getSplitterState("ImageEditorLeft", ui->splitterLeft);
	grinder()->settings().getSplitterState("ImageEditorLayersTags", ui->splitterLayersTags);

	updateActions();
}

void ImageEditorWidget::setupImageControlBar()
{
	ui->imageSceneControlBar->setDefaultAlignment(Qt::AlignRight);
	ui->imageSceneControlBar->setDefaultToolButtonStyle(Qt::ToolButtonIconOnly);

	// Add all image editor tool actions
	for (auto toolAction : _imageEditor->editorTools().getActions())
		ui->imageSceneControlBar->addAction(toolAction, Qt::ToolButtonFollowStyle, Qt::AlignLeft);

	// Add the copy&paste actions
	ui->imageSceneControlBar->addSeparator(Qt::AlignLeft);
	ui->imageSceneControlBar->addAction(_copyImageBuild, Qt::ToolButtonFollowStyle, Qt::AlignLeft);
	ui->imageSceneControlBar->addAction(_pasteImageBuild, Qt::ToolButtonFollowStyle, Qt::AlignLeft);
	ui->imageSceneControlBar->addAction(_duplicateImageBuild, Qt::ToolButtonFollowStyle, Qt::AlignLeft, &_duplicateImageBuildMenu);
	ui->imageSceneControlBar->addSeparator(Qt::AlignLeft);
	ui->imageSceneControlBar->addAction(_editImageBuildTags, Qt::ToolButtonFollowStyle, Qt::AlignLeft);

	ui->imageScene->setupUi(static_cast<QMenu*>(nullptr), ui->imageSceneControlBar);

	// Add the zoom level label
	_imageSceneZoomLabel = new QLabel{""};
	_imageSceneZoomLabel->setAlignment(Qt::AlignCenter);
	ui->imageSceneControlBar->addSeparator();
	ui->imageSceneControlBar->addWidget(_imageSceneZoomLabel);

	QFontMetrics fontMetrics{_imageSceneZoomLabel->font()};
	_imageSceneZoomLabel->setMinimumWidth(fontMetrics.width("9999%"));

	// Show the current zoom level in the image scene control bar
	connect(ui->imageScene, &VisualSceneView::zoomChanged, this, &ImageEditorWidget::updateSceneZoomLevel);
}

void ImageEditorWidget::updatePrimaryColorLabel()
{
	auto color = ui->primaryColorWidget->getColor();
	ui->lblPrimaryColor->setText(QString{"R: %1\nG: %2\nB: %3"}.arg(color.red()).arg(color.green()).arg(color.blue()));
}

void ImageEditorWidget::on_splitterMid_splitterMoved(int pos, int index)
{
	Q_UNUSED(pos);
	Q_UNUSED(index);

	grinder()->settings().setSplitterState("ImageEditorMid", ui->splitterMid);
}

void ImageEditorWidget::on_splitterLeft_splitterMoved(int pos, int index)
{
	Q_UNUSED(pos);
	Q_UNUSED(index);

	grinder()->settings().setSplitterState("ImageEditorLeft", ui->splitterLeft);
}

void ImageEditorWidget::on_splitterLayersTags_splitterMoved(int pos, int index)
{
	Q_UNUSED(pos);
	Q_UNUSED(index);

	grinder()->settings().setSplitterState("ImageEditorLayersTags", ui->splitterLayersTags);
}

void ImageEditorWidget::on_primaryColorWidget_customContextMenuRequested(const QPoint& pos)
{
	QMenu menu;
	auto presetsMenu = menu.addMenu("&Assign to preset");

	for (unsigned int i = 0; i < ui->colorPresetsWidget->getPresetsCount(); ++i)
	{
		auto action = presetsMenu->addAction(QString{"Preset &%1"}.arg(i + 1), this, SLOT(assignPrimaryColorToPreset()));
		action->setData(i);
	}

	menu.exec(ui->primaryColorWidget->mapToGlobal(pos));
}

void ImageEditorWidget::updateActions()
{
	auto imageBuild = _imageEditor->controller().activeImageBuild();

	_copyImageBuild->setEnabled(imageBuild);
	_duplicateImageBuild->setEnabled(imageBuild && grinder()->project().imageReferences().size() > 1);
	_pasteImageBuild->setEnabled(imageBuild && grinder()->clipboardManager().hasData(ImageBuildVector::Serialization_Element));

	_editImageBuildTags->setEnabled(_imageEditor->controller().activeImageBuild() && _imageEditor->controller().activeImageBuild()->inputImageTags());
}

void ImageEditorWidget::updateSceneZoomLevel(qreal zoomLevel)
{
	_imageSceneZoomLabel->setText(QString{"%1%"}.arg(static_cast<int>(zoomLevel * 100.0)));
}

void ImageEditorWidget::updatePrimaryColor()
{
	if (auto scene = _imageEditor->controller().activeScene())
	{
		auto selNodes = scene->getNodes<DraftItemNode>(true);

		if (!selNodes.empty())
		{
			if (auto item = selNodes.front()->draftItem().lock())	// Make sure that the underlying draft item still exists
				_imageEditor->environment().setPrimaryColor(item->primaryColor()->getValue());
		}
	}
}

void ImageEditorWidget::primaryColorSelected(QColor color, bool byUser)
{	
	if (byUser)
	{
		_imageEditor->environment().setPrimaryColor(color);

		if (_imageEditor->controller().activeScene())
		{
			auto nodes = _imageEditor->controller().activeScene()->getNodes<DraftItemNode>(true);
			_imageEditor->controller().setNodesPrimaryColor(color, nodes);
		}
	}

	updatePrimaryColorLabel();
}

void ImageEditorWidget::colorPresetSelected(QColor color)
{
	primaryColorSelected(color, true);
}

void ImageEditorWidget::assignPrimaryColorToPreset()
{
	if (auto action = dynamic_cast<QAction*>(sender()))
	{
		unsigned int index = action->data().toUInt();
		ui->colorPresetsWidget->setColor(index, ui->primaryColorWidget->getColor());
	}
}

void ImageEditorWidget::copyImageBuild() const
{
	_imageEditor->controller().copyImageBuild();
}

void ImageEditorWidget::pasteImageBuild() const
{
	_imageEditor->controller().pasteImageBuild();
}

void ImageEditorWidget::duplicateImageBuild() const
{
	_imageEditor->controller().duplicateImageBuild();
}

void ImageEditorWidget::duplicateImageBuildTo() const
{
	CheckListDialog dlg{"Select target images"};
	auto imageReferences = grinder()->project().imageReferences().sorted();
	imageReferences.erase(grinder()->projectController().activeImageReference());	// Do not include the current image reference in the available targets
	auto selectedImages = dlg.exec<ImageReference, ImageReferencesListItem>(imageReferences);

	if (!selectedImages.empty())
	{
		_imageEditor->controller().duplicateImageBuild(nullptr, selectedImages);
		QMessageBox::information(nullptr, "Duplicating image builds", "Finished duplicating the current image build.");
	}
}

void ImageEditorWidget::editImageBuildTags()
{
	if (auto imageBuild = _imageEditor->controller().activeImageBuild())
	{
		// Use the image tags allotment property editor to edit the image build's tags
		ImageTagsAllotmentPropertyEditor propertyEditor{imageBuild->imageTagsAllotment(), this};
		propertyEditor.invokeDialog();
	}
}

void ImageEditorWidget::primaryColorChanged(QColor color)
{
	ui->primaryColorWidget->setColor(color);
}
