/******************************************************************************
 * File: LayersListWidget.cpp
 * Date: 17.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "LayersListWidget.h"
#include "ImageEditor.h"
#include "image/ImageBuild.h"
#include "core/GrinderApplication.h"
#include "ui/widgets/ControlBar.h"
#include "ui/widgets/SliderValueWidget.h"
#include "util/UIUtils.h"
#include "util/StringUtils.h"
#include "res/Resources.h"

LayersListWidget::LayersListWidget(QWidget* parent) : MetaWidget(parent),
	_sliderValueWidget{new SliderValueWidget{"Opacity:", 0, 100, "%"}}
{
	// Create layers actions
	_renameLayerAction = UIUtils::createAction(this, "Rename &layer", FILE_ICON_EDIT, SLOT(renameLayer()), "Rename the selected layer", "F2");
	_moveUpAction = UIUtils::createAction(this, "Move &up", FILE_ICON_MOVEUP, SLOT(moveLayerUp()), "Move the selected layer up", "Ctrl+Shift+Up");
	_moveDownAction = UIUtils::createAction(this, "Move &down", FILE_ICON_MOVEDOWN, SLOT(moveLayerDown()), "Move the selected layer down", "Ctrl+Shift+Down");
	_convertPixelsToItemsAction = UIUtils::createAction(this, "Convert pixels to &items", FILE_ICON_EDITOR_CONVERTTOITEMS, SLOT(convertPixelsToItems()), "Convert all pixels on the selected layer to items");
	_copyLayer = UIUtils::createAction(this, "&Copy layer", FILE_ICON_COPY, SLOT(copyLayer()), "Copy the selected layer to the clipboard");
	_copyLayer->setShortcut(QKeySequence{QKeySequence::Copy});
	_pasteLayer = UIUtils::createAction(this, "&Paste layer", FILE_ICON_PASTE, SLOT(pasteLayer()), "Paste layer from the clipboard");
	_pasteLayer->setShortcut(QKeySequence{QKeySequence::Paste});
	_cutLayer = UIUtils::createAction(this, "Cu&t layer", FILE_ICON_CUT, SLOT(cutLayer()), "Copy the selected layer to the clipboard and remove it afterwards");
	_cutLayer->setShortcut(QKeySequence{QKeySequence::Cut});
	_removeLayerAction = UIUtils::createAction(this, "&Remove layer", FILE_ICON_DELETE, SLOT(removeLayer()), "Remove the selected layer", "Del");
	_removeAllLayersAction = UIUtils::createAction(this, "Remove all layers", "", SLOT(removeAllLayers()), "Remove all layers");

	// Create the alpha slider action
	auto sliderAction = new QWidgetAction{this};
	sliderAction->setDefaultWidget(_sliderValueWidget);
	_alphaAction = sliderAction;

	// Update the layer opacity (alpha) if the value has been changed
	connect(_sliderValueWidget, &SliderValueWidget::valueChanged, this, &LayersListWidget::layerOpacityChanged);

	_moveUpAction->setData(Qt::ToolButtonIconOnly);
	_moveDownAction->setData(Qt::ToolButtonIconOnly);
	_convertPixelsToItemsAction->setData(Qt::ToolButtonIconOnly);

	// Update the image build if the layer has been (un)checked
	connect(this, &LayersListWidget::itemChanged, this, &LayersListWidget::layerCheckChanged);

	// Get notified when a label name has been edited
	connect(itemDelegate(), &QAbstractItemDelegate::commitData, this, &LayersListWidget::layerRenamed, Qt::QueuedConnection);	// Must be queued to prevent issues if renaming fails

	// Listen for item selections in order to update the active layer and actions
	connect(this, &LayersListWidget::itemSelectionChanged, this, &LayersListWidget::selectedItemChanged);

	// Listen for clipboard changes to update our actions
	connect(&grinder()->clipboardManager(), &ClipboardManager::dataChanged, this, &LayersListWidget::updateActions);

	updateActions();
}

void LayersListWidget::assignUiComponents(ImageEditor* imageEditor)
{
	if (!imageEditor)
		throw std::invalid_argument{_EXCPT("imageEditor may not be null")};

	_imageEditor = imageEditor;

	// Listen for controller events in order to update the UI accordingly
	connect(&imageEditor->controller(), &ImageEditorController::imageBuildSwitched, this, &LayersListWidget::imageBuildSwitched);
	connect(&imageEditor->controller(), &ImageEditorController::layerSwitching, this, &LayersListWidget::layerSwitching);
	connect(&imageEditor->controller(), &ImageEditorController::layerSwitched, this, &LayersListWidget::layerSwitched);
}

void LayersListWidget::setupUi(QAction* newLayerAction, QMenu* menu, ControlBar* controlBar)
{
	_newLayerAction = newLayerAction;

	MetaWidget::setupUi(menu, controlBar);
}

void LayersListWidget::swapLayers(int indexOld, int indexNew)
{
	auto layerSelected = currentObjectItem();

	// Layers are ordered in a highest-to-lowest fashion in the list, so reverse their indices
	indexOld = count() - 1 - indexOld;
	indexNew = count() - 1 - indexNew;

	auto item = takeItem(indexOld);
	insertItem(indexNew, item);

	switchToObjectItem(layerSelected);
}

std::vector<QAction*> LayersListWidget::getActions(MetaWidget::AddActionsMode mode) const
{
	std::vector<QAction*> actions;

	if (mode == AddActionsMode::ContextMenu || mode == AddActionsMode::Toolbar)
	{
		if (mode == AddActionsMode::ContextMenu)
		{
			actions.push_back(_renameLayerAction);
			actions.push_back(nullptr);
			actions.push_back(_cutLayer);
			actions.push_back(_copyLayer);
			actions.push_back(_pasteLayer);
			actions.push_back(nullptr);
		}
	}

	actions.push_back(_newLayerAction);	
	actions.push_back(_removeLayerAction);
	actions.push_back(nullptr);	

	if (mode == AddActionsMode::ContextMenu)
	{
		actions.push_back(_alphaAction);
		actions.push_back(nullptr);
	}

	actions.push_back(_convertPixelsToItemsAction);
	actions.push_back(nullptr);

	actions.push_back(_moveUpAction);
	actions.push_back(_moveDownAction);

	if (mode == AddActionsMode::ContextMenu)
	{
		actions.push_back(nullptr);
		actions.push_back(_removeAllLayersAction);		
	}

	return actions;
}

void LayersListWidget::switchToObjectItem(LayersListItem* item, bool selectItem)
{
	if (item)
	{
		_imageEditor->controller().switchLayer(item->object());

		if (selectItem)
			setCurrentItem(item);
	}
	else
		_imageEditor->controller().switchLayer(nullptr);
}

void LayersListWidget::newLayer()
{
	QString newLayerName = StringUtils::generateUniqueItemName(_imageEditor->controller().activeImageBuild()->layers(), "New layer", &Layer::getName);
	bool ok = false;
	auto name = QInputDialog::getText(this, "Layer name", "Enter the name of the new layer:", QLineEdit::Normal, newLayerName, &ok);

	if (ok)
	{
		auto layer = _imageEditor->controller().createLayer(name.trimmed());

		if (layer)
		{
			// Find the just-created layer item and switch to it
			auto item = findObjectItem(layer.get());

			if (item)
				switchToObjectItem(item);
		}
	}
}

void LayersListWidget::moveLayerUp()
{
	if (auto layerSelected = currentObjectItem())
		_imageEditor->controller().moveLayer(layerSelected->object(), true);
}

void LayersListWidget::moveLayerDown()
{
	if (auto layerSelected = currentObjectItem())
		_imageEditor->controller().moveLayer(layerSelected->object(), false);
}

void LayersListWidget::convertPixelsToItems()
{
	if (auto layer = currentObject())
	{
		_imageEditor->controller().convertLayerPixelsToItems(layer);
		updateActions();
	}
}

void LayersListWidget::copyLayer()
{
	if (auto layerSelected = currentObjectItem())
		_imageEditor->controller().copyLayer(layerSelected->object());
}

void LayersListWidget::pasteLayer()
{
	_imageEditor->controller().pasteLayer();
}

void LayersListWidget::cutLayer()
{
	if (auto layerSelected = currentObjectItem())
		_imageEditor->controller().cutLayer(layerSelected->object());
}

void LayersListWidget::removeLayer()
{
	if (auto layer = currentObject())
	{
		_imageEditor->controller().removeLayer(layer);

		// Switch to the layer that is currently selected if no active one exists
		if (!_imageEditor->controller().activeLayer())
			switchToObjectItem(currentObjectItem());
	}
}

void LayersListWidget::removeAllLayers()
{
	_imageEditor->controller().removeAllLayers();
	updateActions();
}

void LayersListWidget::imageBuildSwitched(ImageBuild* imageBuild)
{
	// A new image build is shown in the editor, so populate its layers
	populateList(imageBuild->layers(), true, true);
}

void LayersListWidget::layerCheckChanged(QListWidgetItem* item) const
{
	// Update the visibility of the layer
	if (auto layerItem = dynamic_cast<LayersListItem*>(item))
		_imageEditor->controller().setLayerVisibility(layerItem->object(), item->checkState() == Qt::Checked);
}

void LayersListWidget::layerOpacityChanged(int value) const
{
	if (auto layer = currentObject())
		_imageEditor->controller().setLayerAlpha(layer, value);
}

void LayersListWidget::layerSwitching(Layer* layer)
{
	if (layer)
		disconnect(layer, nullptr, this, nullptr);
}

void LayersListWidget::layerSwitched(Layer* layer)
{
	objectSwitched(layer);
	updateActions();

	if (layer)
	{
		// Whenever pixels have been changed, update our actions
		connect(&layer->layerPixels().data(), &LayerPixelsData::dataModified, this, &LayersListWidget::updateActions);
	}
}

void LayersListWidget::layerRenamed(QWidget* editor) const
{
	if (auto layerItem = currentObjectItem())
	{
		viewport()->setUpdatesEnabled(false);

		auto newName = reinterpret_cast<QLineEdit*>(editor)->text();
		_imageEditor->controller().renameLayer(layerItem->object(), newName);

		layerItem->updateItem();
		viewport()->setUpdatesEnabled(true);
	}
}

void LayersListWidget::selectedItemChanged()
{
	// The active layer switches immediately when selecting a layer item
	switchLayer();
	updateActions();
}

void LayersListWidget::updateActions()
{
	auto currentLayer = currentObject();
	bool layerSelected = (currentLayer != nullptr);
	int removableLayersCount = 0;

	for (int i = 0; i < count(); ++i)
	{
		if (objectItem(i)->isRemovable())
			++removableLayersCount;
	}

	_renameLayerAction->setEnabled(layerSelected);
	_moveUpAction->setEnabled(layerSelected && currentRow() > 0);
	_moveDownAction->setEnabled(layerSelected && currentRow() < count() - 1);
	_convertPixelsToItemsAction->setEnabled(layerSelected && !currentLayer->layerPixels().data().empty());
	_copyLayer->setEnabled(layerSelected);
	_pasteLayer->setEnabled(grinder()->clipboardManager().hasData(LayerVector::Serialization_Element));
	_cutLayer->setEnabled(layerSelected && currentLayer->isRemovable());
	_removeLayerAction->setEnabled(layerSelected && currentLayer->isRemovable());
	_removeAllLayersAction->setEnabled(removableLayersCount > 0);

	_sliderValueWidget->setValue(currentLayer ? currentLayer->getAlpha() : 0);
}
