/******************************************************************************
 * File: PaintbrushTool.cpp
 * Date: 21.5.2018
 *****************************************************************************/

#include "Grinder.h"
#include "PaintbrushTool.h"
#include "ui/image/ImageEditor.h"
#include "util/ImageUtils.h"
#include "res/Resources.h"

const char* PaintbrushTool::tool_type = "PaintbrushTool";

PaintbrushTool::PaintbrushTool(ImageEditor* imageEditor) : PaintbrushTool(imageEditor, "Paintbrush", FILE_ICON_EDITOR_PAINTBRUSH, "P", QCursor{QPixmap{FILE_CURSOR_EDITOR_PAINTBRUSH}, 0, 23})
{

}

PaintbrushTool::PaintbrushTool(ImageEditor* imageEditor, QString name, QString icon, QString shortcut, QCursor cursor) : ImageEditorTool(imageEditor, name, icon, shortcut, cursor)
{
	_supportRightMouseButton = true;
}

void PaintbrushTool::createProperties()
{
	ImageEditorTool::createProperties();

	// Create specific properties
	_penWidth = createProperty<UIntProperty>(PropertyID::Width, "Brush width", 1);
	penWidth()->createConstraint<RangeConstraint>(1, 100);
	penWidth()->setDescription("The width of the pixels to paint.");
}

ImageEditorTool::InputEventResult PaintbrushTool::mousePressed(const QGraphicsSceneMouseEvent* event)
{
	auto pos = event->scenePos().toPoint();
	_lastPixelPos = pos;
	paintPixel(pos, _eraseByDefault);
	return InputEventResult::Process;
}

ImageEditorTool::InputEventResult PaintbrushTool::mouseMoved(const QGraphicsSceneMouseEvent* event)
{
	auto pos = event->scenePos().toPoint();
	paintPixel(pos, _eraseByDefault);
	_lastPixelPos = pos;
	return InputEventResult::Process;
}

VisualSceneInputHandler::InputEventResult PaintbrushTool::rightMousePressed(const QGraphicsSceneMouseEvent* event)
{
	auto pos = event->scenePos().toPoint();
	_lastPixelPos = pos;
	paintPixel(pos, !_eraseByDefault);
	return InputEventResult::Process;
}

VisualSceneInputHandler::InputEventResult PaintbrushTool::rightMouseMoved(const QGraphicsSceneMouseEvent* event)
{
	auto pos = event->scenePos().toPoint();
	paintPixel(pos, !_eraseByDefault);
	_lastPixelPos = pos;
	return InputEventResult::Process;
}

void PaintbrushTool::paintPixel(QPoint pos, bool erase) const
{
	// Draw a straight line from the current position to the last painted pixel
	std::vector<QPoint> linePoints;
	int width = *penWidth();

	if (_eraseByDefault)
	{
		auto adjustedWidth = static_cast<qreal>(width) / _imageEditor->controller().activeScene()->view()->getZoom();
		width = static_cast<int>(std::ceil(adjustedWidth));

		if (width <= 0)
			width = 1;
	}

	if (width > 1)
	{
		for (auto point : ImageUtils::getLinePoints(pos, _lastPixelPos))
		{
			int dFrom = -(width >> 2) + (width % 2 == 0 ? 1 : 0);
			int dTo = width >> 2;

			for (int dx = dFrom; dx <= dTo; ++dx)
			{
				for (int dy = dFrom; dy <= dTo; ++dy)
				{
					auto pointOffset = point + QPoint{dx, dy};

					if (std::find(linePoints.cbegin(), linePoints.cend(), pointOffset) == linePoints.cend())
						linePoints.push_back(point + QPoint{dx, dy});
				}
			}
		}
	}
	else
		linePoints = ImageUtils::getLinePoints(pos, _lastPixelPos);

	_imageEditor->controller().paintPixels(nullptr, !erase ? _imageEditor->environment().getPrimaryColor() : QColor{}, linePoints);
}
