/******************************************************************************
 * File: GrinderWindow.cpp
 * Date: 11.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GrinderWindow.h"
#include "ui_GrinderWindow.h"
#include "Version.h"
#include "common/AdjacentRange.h"
#include "ui/StyleSheet.h"
#include "ui/graph/GraphView.h"
#include "ui/dlg/OptionsDialog.h"
#include "core/GrinderApplication.h"
#include "pipeline/Pipeline.h"
#include "project/ProjectExceptions.h"
#include "util/UIUtils.h"
#include "res/Resources.h"

Q_DECLARE_METATYPE(std::shared_ptr<Block>)

GrinderWindow::GrinderWindow(QWidget* parent) : QMainWindow(parent),
	ui{new Ui::GrinderWindow()}
{
	// Needed to be able to queue blockRemoved signals
	qRegisterMetaType<std::shared_ptr<Block>>();

	setupUi();

	// Connect signals to update our UI
	connect(&grinder()->pipelineController(), &PipelineController::pipelineSwitched, this, &GrinderWindow::pipelineSwitched);
	connect(&grinder()->projectController(), &ProjectController::currentProjectFileChanged, this, &GrinderWindow::currentProjectFileChanged);
	connect(&grinder()->projectController(), SIGNAL(projectLoaded(QString)), this, SLOT(updateRecentProjects(QString)));
	connect(&grinder()->projectController(), SIGNAL(projectSaved(QString)), this, SLOT(updateRecentProjects(QString)));
	connect(&grinder()->project(), SIGNAL(imageReferenceCreated(const std::shared_ptr<ImageReference>&)), this, SLOT(updateActions()));
	connect(&grinder()->project(), SIGNAL(imageReferenceRemoved(const std::shared_ptr<ImageReference>&)), this, SLOT(updateActions()));

	// Connect image editor manager signals to show/hide the image editors
	connect(&grinder()->imageEditorManager(), &ImageEditorManager::editorShown, this, &GrinderWindow::imageEditorShown);
	connect(&grinder()->imageEditorManager(), &ImageEditorManager::editorClosed, this, &GrinderWindow::imageEditorClosed);

	updateActions();

	// If the corresponding option is enabled, load the most recent project; otherwise, create a new project
	bool createNewProject = true;

	if (grinder()->settings().startupOptions().loadLastProject)
		createNewProject = !loadProjectOnStartup();

	if (createNewProject)
		on_actNewProject_triggered();
}

GrinderWindow::~GrinderWindow()
{
	delete ui;
}

void GrinderWindow::closeEvent(QCloseEvent* event)
{
	if (!handleUnsavedProjectChanges())
	{
		event->ignore();
		return;
	}

	grinder()->settings().setWindowState("MainWindow", this);

	QMainWindow::closeEvent(event);
}

void GrinderWindow::on_actNewProject_triggered()
{
	if (!handleUnsavedProjectChanges())
		return;

	grinder()->projectController().clearProject(true);
	ui->statusBar->showMessage("Created a new project");
}

void GrinderWindow::on_actLoadProject_triggered()
{
	if (!handleUnsavedProjectChanges())
		return;

	QString fileName = UIUtils::askFileName(false, "ProjectFile", this, "Load project", "Grinder projects (*.grndr);;All files (*.*)");

	if (!fileName.isEmpty())
		loadProject(fileName);
}

void GrinderWindow::on_actSaveProject_triggered()
{
	QString fileName = grinder()->projectController().getCurrentProjectFile();

	if (!fileName.isEmpty())
		saveProject(fileName);
	else
		on_actSaveProjectAs_triggered();
}

void GrinderWindow::on_actSaveProjectAs_triggered()
{
	QString fileName = UIUtils::askFileName(true, "ProjectFile", this, "Save project as...", "Grinder projects (*.grndr)");

	if (!fileName.isEmpty())
		saveProject(fileName);
}

void GrinderWindow::on_actExecuteLabel_triggered()
{
	grinder()->engineController().executeLabel(nullptr, nullptr, Engine::ExecutionMode::View);
}

void GrinderWindow::on_actExecuteNextImage_triggered()
{
	grinder()->projectController().switchToNextImageReference();
	on_actExecuteLabel_triggered();
}

void GrinderWindow::on_actExecutePreviousImage_triggered()
{
	grinder()->projectController().switchToPreviousImageReference();
	on_actExecuteLabel_triggered();
}

void GrinderWindow::on_actOptions_triggered()
{
	OptionsDialog dlg;
	dlg.exec();
}

void GrinderWindow::on_actAbout_triggered()
{
	auto title = QString{"About %1..."}.arg(GRNDR_INFO_TITLE);
	auto text = QString{"<b>%1 %2</b><br>%3<br><br><a href='%4'>%4</a>"}.arg(GRNDR_INFO_TITLE).arg(GetVersionString(true)).arg(GRNDR_INFO_COPYRIGHT).arg(GRNDR_INFO_WEBSITE);

	QMessageBox::about(this, title, text);
}

void GrinderWindow::pipelineSwitched(Pipeline* pipeline)
{
	static std::vector<QMetaObject::Connection> pipelineSignals;

	// Remove old pipeline signals and connect new ones so that the actions can be updated when necessary
	for (auto& signal : pipelineSignals)
		disconnect(signal);

	pipelineSignals.clear();

	if (pipeline)
	{
		pipelineSignals.push_back(connect(pipeline, &Pipeline::blockCreated, this, &GrinderWindow::updateActions));
		pipelineSignals.push_back(connect(pipeline, &Pipeline::blockRemoved, this, &GrinderWindow::updateActions, Qt::QueuedConnection));	// Delay this signal so that the block has been removed from the blocks list
	}

	updateActions();
}

void GrinderWindow::currentProjectFileChanged(QString fileName)
{
	grinder()->settings().startupOptions().lastProjectFile = fileName;
	grinder()->settings().saveSettings(false);

	updateCaption();
}

void GrinderWindow::imageEditorShown(ImageEditor* editor)
{
	// Add the editor to the dock area if it hasn't been added yet
	if (!findChildren<QDockWidget*>().contains(editor->dockWidget()))
		addDockWidget(Qt::BottomDockWidgetArea, editor->dockWidget());

	if (grinder()->settings().imageEditorOptions().startUndocked)
	{
		// Show the editor in a floating window
		editor->dockWidget()->setFloating(true);
	}
	else
	{
		// Group all bottom image editor docks into tabs
		if (grinder()->settings().imageEditorOptions().groupIntoTabs)
			tabifyImageEditorDocks();
	}
}

void GrinderWindow::imageEditorClosed(ImageEditor* editor)
{
	// Remove the editor from the dock area
	removeDockWidget(editor->dockWidget());
}

void GrinderWindow::updateCaption()
{
	QString caption = QString{"%1 %2"}.arg(GRNDR_INFO_TITLE).arg(GetVersionString());

	if (!grinder()->projectController().getCurrentProjectFile().isEmpty())
		caption += QString{" - %1"}.arg(grinder()->projectController().getCurrentProjectFile());
	else
		caption += " - New project";

	setWindowTitle(caption);
}

void GrinderWindow::setupUi()
{
	ui->setupUi(this);

	// Assign the UI components to the controllers
	grinder()->pipelineController().assignUiComponents(ui->graphWidget->graphView());
	grinder()->projectController().assignUiComponents(ui->labelsList, ui->imageReferencesList);

	// Setup the various UI components
	ui->graphWidget->graphView()->setupUi(ui->menu_Graph, ui->graphToolBar);
	ui->labelsList->setupUi(ui->menu_Labels, ui->labelsControlBar);
	ui->imageReferencesList->setupUi(ui->menu_Images, ui->imageReferencesControlBar);
	ui->menu_Recent_Projects->setupUi(&grinder()->settings().recentProjects(), [this](QString fileName) { loadProject(fileName); });
	ui->propertiesTree->setupUi(ui->lblPropertyDesc);

	ui->frmPropertyDesc->setStyleSheet(StyleSheet::loadStyleSheet(FILE_STYLESHEET_PROPERTYDESC));

	grinder()->settings().getWindowState("MainWindow", this);

	updateCaption();
}

void GrinderWindow::updateActions()
{
	auto activePipeline = grinder()->pipelineController().activePipeline();
	bool enableExecute = activePipeline && !activePipeline->blocks().empty();

	ui->actExecuteLabel->setEnabled(enableExecute);
	ui->actExecuteNextImage->setEnabled(enableExecute && ui->imageReferencesList->count() > 1);
	ui->actExecutePreviousImage->setEnabled(enableExecute && ui->imageReferencesList->count() > 1);
}

bool GrinderWindow::loadProject(QString fileName)
{
	bool result = true;

	try {
		grinder()->projectController().loadProject(fileName);
		ui->statusBar->showMessage(QString{"Loaded project '%1'"}.arg(fileName));
	} catch (ProjectException& e) {
		ui->statusBar->showMessage(GetExceptionMessage(e.what()));
		updateRecentProjects(fileName, false);

		result = false;
	}

	return result;
}

bool GrinderWindow::saveProject(QString fileName)
{
	bool result = true;

	try {
		grinder()->projectController().saveProject(fileName);
		ui->statusBar->showMessage(QString{"Saved the current project to '%1'"}.arg(fileName));
	} catch (ProjectException& e) {
		ui->statusBar->showMessage(GetExceptionMessage(e.what()));

		result = false;
	}

	return result;
}

bool GrinderWindow::loadProjectOnStartup()
{
	QString lastProject = grinder()->settings().startupOptions().lastProjectFile;

	if (!lastProject.isEmpty() && QFile::exists(lastProject))
		return loadProject(lastProject);
	else
		return false;
}

bool GrinderWindow::handleUnsavedProjectChanges()
{
	if (grinder()->settings().generalOptions().askOnUnsavedChanges)
	{
		// Save the current graph layout so that any changes will appear in the isProjectDirty call
		ui->graphWidget->graphView()->updateCurrentGraphLayout();

		if (grinder()->projectController().isProjectDirty())
		{
			auto answer = QMessageBox::question(this, "Unsaved project changes", "There are unsaved project changes. Do you want to save now?", QMessageBox::Yes|QMessageBox::No|QMessageBox::Cancel, QMessageBox::Yes);

			if (answer == QMessageBox::Yes)
				on_actSaveProject_triggered();
			else if (answer == QMessageBox::Cancel)
				return false;
		}
	}

	return true;
}

void GrinderWindow::updateRecentProjects(QString fileName, bool addEntry)
{
	if (addEntry)
		grinder()->settings().recentProjects() += fileName;
	else
		grinder()->settings().recentProjects() -= fileName;

	ui->menu_Recent_Projects->updateUi();
}

void GrinderWindow::tabifyImageEditorDocks()
{
	// Gather all docks of the left, right, top and bottom areas
	std::vector<QDockWidget*> leftArea, rightArea, topArea, bottomArea;
	std::vector<QDockWidget*> activeDocks;

	for (auto dock : findChildren<QDockWidget*>())
	{
		if (dynamic_cast<ImageEditorDockWidget*>(dock) && !dock->isFloating())
		{
			switch (dockWidgetArea(dock))
			{
			case Qt::LeftDockWidgetArea:
				leftArea.push_back(dock);
				break;

			case Qt::RightDockWidgetArea:
				rightArea.push_back(dock);
				break;

			case Qt::TopDockWidgetArea:
				topArea.push_back(dock);
				break;

			case Qt::BottomDockWidgetArea:
				bottomArea.push_back(dock);
				break;

			default:
				break;
			}

			if (!dock->visibleRegion().isEmpty())
				activeDocks.push_back(dock);
		}
	}

	std::vector<std::vector<QDockWidget*>*> dockAreas{&leftArea, &rightArea, &topArea, &bottomArea};

	// Sort all docks of each area and tabify them
	for (auto& dockArea : dockAreas)
	{
		std::sort(dockArea->begin(), dockArea->end(), [](auto dock1, auto dock2) { return dock1->windowTitle() < dock2->windowTitle(); });

		for (const auto& dockPair : MakeAdjacentRange(*dockArea))
			tabifyDockWidget(dockPair.first, dockPair.second);
	}

	// Reactive all previously active tabs
	for (auto& dock : activeDocks)
		dock->raise();
}
