/******************************************************************************
 * File: ImageReferencesListWidget.cpp
 * Date: 10.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageReferencesListWidget.h"
#include "core/GrinderApplication.h"
#include "project/Project.h"
#include "ui/widgets/ControlBar.h"
#include "util/UIUtils.h"
#include "util/ImageUtils.h"
#include "util/FileUtils.h"
#include "res/Resources.h"

ImageReferencesListWidget::ImageReferencesListWidget(QWidget* parent) : MetaWidget(parent)
{
	setAcceptDrops(true);

	// Create labels actions
	_switchImageReferenceAction = UIUtils::createAction(this, "&Activate", FILE_ICON_ACTIVATE, SLOT(switchImageReference()), "Activate the selected image", "Return");
	_nextImageReferenceAction = UIUtils::createAction(this, "Switch to &next image", "", SLOT(nextImageReference()), "Activate the next image", "Ctrl+PgDown", Qt::WindowShortcut);
	_previousImageReferenceAction = UIUtils::createAction(this, "Switch to &previous image", "", SLOT(previousImageReference()), "Activate the previous image", "Ctrl+PgUp", Qt::WindowShortcut);
	_viewImageReferenceAction = UIUtils::createAction(this, "&View image", FILE_ICON_VIEWIMAGE, SLOT(viewImageReference()), "View the selected image");
	_addImageReferencesAction = UIUtils::createAction(this, "Add &images", FILE_ICON_ADD, SLOT(addImageReferences()), "Add one or more images to the project", "Ctrl+I", Qt::WindowShortcut);
	_removeImageReferenceAction = UIUtils::createAction(this, "&Remove image", FILE_ICON_DELETE, SLOT(removeImageReference()), "Remove the selected image", "Del");
	_removeAllImageReferencesAction = UIUtils::createAction(this, "Remove all images", "", SLOT(removeAllImageReferences()), "Remove all images");

	// Listen for item selections in order to update the actions
	connect(this, &ImageReferencesListWidget::itemSelectionChanged, this, &ImageReferencesListWidget::updateActions);

	// Listen for label switching in order to update the active state flags
	connect(&grinder()->projectController(), &ProjectController::imageReferenceSwitched, this, &ImageReferencesListWidget::imageReferenceSwitched);

	updateActions();
}

std::unique_ptr<QMenu> ImageReferencesListWidget::createContextMenu() const
{
	auto menu = widget_type::createContextMenu();
	menu->setDefaultAction(_switchImageReferenceAction);
	return menu;
}

std::vector<QAction*> ImageReferencesListWidget::getActions(AddActionsMode mode) const
{
	std::vector<QAction*> actions;

	if (mode == AddActionsMode::ContextMenu || mode == AddActionsMode::Toolbar)
	{
		actions.push_back(_switchImageReferenceAction);

		if (mode == AddActionsMode::ContextMenu)
			actions.push_back(_viewImageReferenceAction);

		actions.push_back(nullptr);
	}

	if (mode == AddActionsMode::MainMenu || mode == AddActionsMode::ContextMenu)
	{
		actions.push_back(_nextImageReferenceAction);
		actions.push_back(_previousImageReferenceAction);
		actions.push_back(nullptr);
	}

	actions.push_back(_addImageReferencesAction);
	actions.push_back(_removeImageReferenceAction);

	if (mode == AddActionsMode::ContextMenu)
	{
		actions.push_back(nullptr);
		actions.push_back(_removeAllImageReferencesAction);
	}

	return actions;
}

void ImageReferencesListWidget::switchToObjectItem(ImageReferencesListItem* item, bool selectItem)
{
	if (item)
	{
		grinder()->projectController().switchImageReference(item->object());

		if (selectItem)
			setCurrentItem(item);
	}
	else
		grinder()->projectController().switchImageReference(nullptr);
}

void ImageReferencesListWidget::dragEnterEvent(QDragEnterEvent* event)
{
	if (event->mimeData()->hasUrls())
	{
		event->setDropAction(Qt::CopyAction);
		event->accept();
	}
	else
		event->ignore();
}

void ImageReferencesListWidget::dragMoveEvent(QDragMoveEvent* event)
{
	if (event->mimeData()->hasUrls())
	{
		event->setDropAction(Qt::CopyAction);
		event->accept();
	}
	else
		event->ignore();
}

void ImageReferencesListWidget::dropEvent(QDropEvent* event)
{
	if (event->mimeData()->hasUrls())
	{
		auto urls = event->mimeData()->urls();
		QStringList files;

		for (auto url : urls)
			files << url.toLocalFile();

		addImageReferences(files);

		event->setDropAction(Qt::CopyAction);
		event->accept();
	}
	else
		event->ignore();
}

void ImageReferencesListWidget::nextImageReference() const
{
	grinder()->projectController().switchToNextImageReference();
}

void ImageReferencesListWidget::previousImageReference() const
{
	grinder()->projectController().switchToPreviousImageReference();
}

void ImageReferencesListWidget::viewImageReference()
{
	if (auto item = currentObjectItem())
		item->viewImage();
}

void ImageReferencesListWidget::addImageReferences()
{
	auto files = UIUtils::askFileNames("ImageReferences", this, "Add images", QString{"Images (%1);;All files (*.*)"}.arg(ImageUtils::getSupportedFormatFilters().join(" ")));
	addImageReferences(files);
}

void ImageReferencesListWidget::addImageReferences(QStringList files)
{
	// Expand the file list, recursively finding all image files, and create image references for them	
	files = FileUtils::expandFileList(files, ImageUtils::getSupportedFormatFilters());
	auto imageRefs = grinder()->projectController().createImageReferences(files);

	// If no active image exists, switch to the last added one
	if (!grinder()->projectController().activeImageReference())
	{
		if (!imageRefs.empty())
		{
			auto item = findObjectItem(imageRefs[imageRefs.size() - 1].get());

			if (item)
				switchToObjectItem(item);
		}
	}
}

void ImageReferencesListWidget::removeImageReference()
{
	if (auto imageRef = currentObject())
	{
		grinder()->projectController().removeImageReference(imageRef);

		// Switch to the image reference that is currently selected if no active one exists
		if (!grinder()->projectController().activeImageReference())
			switchToObjectItem(currentObjectItem());
	}
}

void ImageReferencesListWidget::removeAllImageReferences()
{
	grinder()->projectController().removeAllImageReferences();
	updateActions();
}

void ImageReferencesListWidget::imageReferenceSwitched(ImageReference* imageRef)
{
	objectSwitched(imageRef);
	updateActions();
}

void ImageReferencesListWidget::updateActions()
{
	bool imageSelected = (currentObjectItem() != nullptr);

	_switchImageReferenceAction->setEnabled(imageSelected && !currentObjectItem()->isActive());
	_nextImageReferenceAction->setEnabled(count() > 1);
	_previousImageReferenceAction->setEnabled(count() > 1);
	_viewImageReferenceAction->setEnabled(imageSelected);
	_removeImageReferenceAction->setEnabled(imageSelected);
	_removeAllImageReferencesAction->setEnabled(count() > 0);
}
