/******************************************************************************
 * File: LabelsListWidget.cpp
 * Date: 07.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "LabelsListWidget.h"
#include "core/GrinderApplication.h"
#include "project/Project.h"
#include "util/StringUtils.h"
#include "util/UIUtils.h"
#include "res/Resources.h"

LabelsListWidget::LabelsListWidget(QWidget* parent) : MetaWidget(parent)
{
	// Create labels actions
	_switchLabelAction = UIUtils::createAction(this, "&Activate", FILE_ICON_ACTIVATE, SLOT(switchLabel()), "Activate the selected label", "Return");
	_renameLabelAction = UIUtils::createAction(this, "Rename label", FILE_ICON_EDIT, SLOT(renameLabel()), "Rename the selected label", "F2");
	_newLabelAction = UIUtils::createAction(this, "&New label", FILE_ICON_ADD, SLOT(newLabel()), "Add a new label", "Ctrl+L", Qt::WindowShortcut);
	_removeLabelAction = UIUtils::createAction(this, "&Remove label", FILE_ICON_DELETE, SLOT(removeLabel()), "Remove the selected label", "Del");
	_removeAllLabelsAction = UIUtils::createAction(this, "Remove all labels", "", SLOT(removeAllLabels()), "Remove all labels");

	// Get notified when a label name has been edited
	connect(itemDelegate(), &QAbstractItemDelegate::commitData, this, &LabelsListWidget::labelRenamed, Qt::QueuedConnection);	// Must be queued to prevent issues if renaming fails

	// Listen for item selections in order to update the actions
	connect(this, &LabelsListWidget::itemSelectionChanged, this, &LabelsListWidget::updateActions);

	// Listen for label switching in order to update the active state flags
	connect(&grinder()->projectController(), &ProjectController::labelSwitched, this, &LabelsListWidget::labelSwitched);

	updateActions();
}

std::unique_ptr<QMenu> LabelsListWidget::createContextMenu() const
{
	auto menu = widget_type::createContextMenu();
	menu->setDefaultAction(_switchLabelAction);
	return menu;
}

std::vector<QAction*> LabelsListWidget::getActions(AddActionsMode mode) const
{
	std::vector<QAction*> actions;

	if (mode == AddActionsMode::ContextMenu || mode == AddActionsMode::Toolbar)
	{
		actions.push_back(_switchLabelAction);

		if (mode == AddActionsMode::ContextMenu)
			actions.push_back(_renameLabelAction);

		actions.push_back(nullptr);
	}

	actions.push_back(_newLabelAction);
	actions.push_back(_removeLabelAction);

	if (mode == AddActionsMode::ContextMenu)
	{
		actions.push_back(nullptr);
		actions.push_back(_removeAllLabelsAction);
	}

	return actions;
}

void LabelsListWidget::switchToObjectItem(LabelsListItem* item, bool selectItem)
{
	if (item)
	{
		grinder()->projectController().switchLabel(item->object());

		if (selectItem)
			setCurrentItem(item);
	}
	else
		grinder()->projectController().switchLabel(nullptr);
}

void LabelsListWidget::newLabel()
{
	QString newLabelName = StringUtils::generateUniqueItemName(grinder()->project().labels(), "New label", &Label::getName);
	bool ok = false;
	auto name = QInputDialog::getText(this, "Label name", "Enter the name of the new label:", QLineEdit::Normal, newLabelName, &ok);

	if (ok)
	{
		auto label = grinder()->projectController().createLabel(name.trimmed());

		if (label)
		{
			// Find the just-created label item and switch to it
			auto item = findObjectItem(label.get());

			if (item)
				switchToObjectItem(item);
		}
	}
}

void LabelsListWidget::removeLabel()
{
	if (auto label = currentObject())
	{
		grinder()->projectController().removeLabel(label);

		// Switch to the label that is currently selected if no active one exists
		if (!grinder()->projectController().activeLabel())
			switchToObjectItem(currentObjectItem());
	}
}

void LabelsListWidget::removeAllLabels()
{
	grinder()->projectController().removeAllLabels();
	updateActions();
}

void LabelsListWidget::labelSwitched(Label* label)
{
	objectSwitched(label);
	updateActions();
}

void LabelsListWidget::labelRenamed(QWidget* editor) const
{
	if (auto labelItem = currentObjectItem())
	{
		viewport()->setUpdatesEnabled(false);

		auto newName = reinterpret_cast<QLineEdit*>(editor)->text();
		grinder()->projectController().renameLabel(labelItem->object(), newName);

		labelItem->updateItem();
		viewport()->setUpdatesEnabled(true);
	}
}

void LabelsListWidget::updateActions()
{
	bool labelSelected = (currentObjectItem() != nullptr);

	_switchLabelAction->setEnabled(labelSelected && !currentObjectItem()->isActive());
	_renameLabelAction->setEnabled(labelSelected);
	_removeLabelAction->setEnabled(labelSelected);
	_removeAllLabelsAction->setEnabled(count() > 0);
}
