/******************************************************************************
 * File: VisualScene.impl.h
 * Date: 15.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "VisualScene.h"

template<typename ViewType>
VisualScene<ViewType>::VisualScene(view_type* view, QRectF sceneRect) : QGraphicsScene (sceneRect),
	_view{view}
{
	if (!view)
		throw std::invalid_argument{_EXCPT("view may not be null")};

	// Force redraw if the selection has changed
	connect(this, SIGNAL(selectionChanged()), this, SLOT(update()));
}

template<typename ViewType>
VisualScene<ViewType>::~VisualScene()
{
	// Prevent selection change events from firing when a scene is destroyed
	disconnect(this, &VisualScene::selectionChanged, nullptr, nullptr);
}

template<typename ViewType>
template<typename NodeType>
std::vector<const NodeType*> VisualScene<ViewType>::getNodes(bool selectedOnly) const
{
	auto selNodes = _getNodes<NodeType>(selectedOnly);
	std::vector<const NodeType*> nodes;

	for (auto node : selNodes)
		nodes.push_back(node);

	return nodes;
}

template<typename ViewType>
template<typename NodeType>
std::vector<NodeType*> VisualScene<ViewType>::_getNodes(bool selectedOnly) const
{
	std::vector<NodeType*> nodes;

	for (auto item : selectedOnly ? selectedItems() : items())
	{
		if (auto node = dynamic_cast<NodeType*>(item))
			nodes.push_back(node);
	}

	return nodes;
}

template<typename ViewType>
void VisualScene<ViewType>::drawBackground(QPainter* painter, const QRectF& rect)
{
	QGraphicsScene::drawBackground(painter, rect);

	if (_drawGrid)
	{
		auto gridStyle = _view->sceneStyle().getGridStyle();

		// Draw small grid (only if not zoomed in too much)
		if (_view->getZoom() < gridStyle.size * gridStyle.maxZoomFactor)
			drawGrid(painter, rect, gridStyle.size, gridStyle.colorLight);

		// Draw large grid
		drawGrid(painter, rect, gridStyle.size * gridStyle.darkGridInterval, gridStyle.colorDark);
	}
}

template<typename ViewType>
void VisualScene<ViewType>::drawGrid(QPainter* painter, const QRectF& rect, int interval, QColor color)
{
	QVector<QLineF> lines;

	qreal x = static_cast<int>(rect.left()) - (static_cast<int>(rect.left()) % interval);;

	while (x < rect.right())
	{
		lines << QLineF{x, rect.top(), x, rect.bottom()};
		x += interval;
	}

	qreal y = static_cast<int>(rect.top()) - (static_cast<int>(rect.top()) % interval);;

	while (y < rect.bottom())
	{
		lines << QLineF{rect.left(), y, rect.right(), y};
		y += interval;
	}

	painter->setPen(QPen{color});
	painter->drawLines(lines);
}
